/*****************************************************************************
 *------------------------ Utility  -----------------------------------------*
 *****************************************************************************/
#include <float.h>
#include <math.h>
#include "rsbtc.h"

/*****************************************************************************
 *- Callback Declarations ---------------------------------------------------*
 *****************************************************************************/

ViStatus rsbtc_CheckStatusCallback (ViSession io);

/*****************************************************************************
 *- Common Callback Declarations --------------------------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: rsbtc_InitAttributes
 * Purpose:  This function inits attributes to the desired values if needed.
 *****************************************************************************/
ViStatus rsbtc_InitAttributes (ViSession instrSession)
{
    ViStatus    error           = VI_SUCCESS;
    ViInt32     p2value         = 0;
    void        *p2void;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE],
                revision[RS_MAX_MESSAGE_BUF_SIZE],
                manufacturer[RS_MAX_MESSAGE_BUF_SIZE],
                model[RS_MAX_MESSAGE_BUF_SIZE],
                tmp_model[RS_MAX_MESSAGE_BUF_SIZE],
                idQ[RS_MAX_MESSAGE_BUF_SIZE],
                options_list[RS_MAX_MESSAGE_BUF_SIZE];
                

    buffer[RS_MAX_MESSAGE_BUF_SIZE -1] = 0;

    /*- Misc -*/

    if (!Rs_Simulating(instrSession))
        {
        /* Instrument identification */
        viCheckErr (viQueryf (instrSession, "*IDN?\n", "%[^\r\n]", idQ));
        idQ[RS_MAX_MESSAGE_BUF_SIZE - 1] = 0;
        checkErr (Rs_SetAttribute (instrSession, "", RSBTC_ATTR_ID_QUERY_RESPONSE, 0, idQ));
        
        /* Option(s) string */
        viCheckErr (viQueryf (instrSession, "*OPT?\n", "%[^\r\n]", options_list));
        options_list[RS_MAX_MESSAGE_BUF_SIZE - 1] = 0;
        
        }
    else /* Simulating */
        {
        /* Instrument identification */
        strncpy (idQ, RSBTC_SIMULATION_ID_QUERY, RS_MAX_MESSAGE_BUF_SIZE - 1);
        checkErr (Rs_SetAttribute (instrSession, "", RSBTC_ATTR_ID_QUERY_RESPONSE, 0, idQ));
        
        /* Option(s) string */
        strncpy (options_list, RSBTC_SIMULATION_OPT_QUERY, RS_MAX_MESSAGE_BUF_SIZE - 1);
        }

    /*- Inherent Instrument Specific Attributes ----------------------------*/

    /*- User Options -*/

    /*
    RS_ATTR_RANGE_CHECK             ... Rs_SpecificDriverNew
    RS_ATTR_QUERY_INSTRUMENT_STATUS ... Rs_SpecificDriverNew
    RS_ATTR_CACHE                   ... Rs_SpecificDriverNew
    RS_ATTR_SIMULATE                ... Rs_SpecificDriverNew
    RS_ATTR_RECORD_COERCIONS        ... not supported
    RS_ATTR_INTERCHANGE_CHECK       ... not supported
    RS_ATTR_SPY                     ... not supported
    RS_ATTR_USE_SPECIFIC_SIMULATION ... not supported
    */

    /*- Class Driver Identification -*/

    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_DESCRIPTION, 0, RSBTC_CLASS_DRIVER_DESCRIPTION));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_PREFIX, 0, RSBTC_CLASS_DRIVER_PREFIX));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_VENDOR, 0, RSBTC_CLASS_DRIVER_VENDOR));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_REVISION, 0, RSBTC_CLASS_DRIVER_REVISION));
    p2value = RSBTC_CLASS_SPEC_MAJOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION, 0, &p2value));
    p2value = RSBTC_CLASS_SPEC_MINOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION, 0, &p2value));

    /*- Driver Identification -*/
                                                        
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_DESCRIPTION, 0, RSBTC_SPECIFIC_DRIVER_DESCRIPTION));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_PREFIX, 0, RSBTC_SPECIFIC_DRIVER_PREFIX));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_LOCATOR, 0, RSBTC_SPECIFIC_DRIVER_LOCATOR));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_VENDOR, 0, RSBTC_SPECIFIC_DRIVER_VENDOR));

    sprintf (buffer, 
             "Driver: %s %d.%d.%d, Compiler: %s %3.2f, "
             "Components: RSEngine %.2f, VISA-Spec %.2f",
             RSBTC_SPECIFIC_DRIVER_PREFIX,
             RSBTC_MAJOR_VERSION, RSBTC_MINOR_VERSION, RSBTC_MINOR_MINOR_VERSION,
             RS_COMPILER_NAME, RS_COMPILER_VER_NUM, 
             RS_ENGINE_MAJOR_VERSION + RS_ENGINE_MINOR_VERSION/10.0, 
             Rs_ConvertVISAVer(VI_SPEC_VERSION));

    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_REVISION, 0, buffer));
    p2value = RSBTC_CLASS_SPEC_MAJOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION, 0, &p2value));
    p2value = RSBTC_CLASS_SPEC_MINOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION, 0, &p2value));

    /*- Driver Capabilities -*/
                                   
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SUPPORTED_INSTRUMENT_MODELS, 0, RSBTC_SUPPORTED_INSTRUMENT_MODELS));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_GROUP_CAPABILITIES, 0, RSBTC_GROUP_CAPABILITIES));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_FUNCTION_CAPABILITIES, 0, RSBTC_FUNCTION_CAPABILITIES));
    /* RS_ATTR_CHANNEL_COUNT        ... Rs_BuildRepCapTable */

    /*- Driver Setup -*/

    /* RS_ATTR_DRIVER_SETUP         ... Rs_SpecificDriverNew */

    /*- Instrument Identification -*/

    checkErr (Rs_GetAttribute (instrSession, "", RSBTC_ATTR_ID_QUERY_RESPONSE, 0, RSBTC_IO_BUFFER_SIZE, idQ));
    sscanf (idQ, "%[^,],%[^,],%*[^,],%256[^\n]", manufacturer, tmp_model, revision); // added 2 spaces in the parsing string
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_INSTRUMENT_MANUFACTURER, 0, manufacturer));
    sscanf (tmp_model,"%[^-0-9]",model); // added number elimination
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_INSTRUMENT_MODEL, 0, model));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_INSTRUMENT_FIRMWARE_REVISION, 0, revision));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_OPTIONS_LIST, 0, options_list));

    /*- Advanced Session Information -*/
                                  
    /*  
    RS_ATTR_IO_RESOURCE_DESCRIPTOR  ... rsbtc_InitWithOptions
    RS_ATTR_LOGICAL_NAME            ... not supported
    */

    /*- Inherent Vendor Specific Attributes --------------------------------*/

    /*- Error Info -*/

    /*
    RS_ATTR_PRIMARY_ERROR           ... Rs_SetErrorInfo
    RS_ATTR_SECONDARY_ERROR         ... Rs_SetErrorInfo
    RS_ATTR_ERROR_ELABORATION       ... Rs_SetErrorInfo
    */

    /*- Session I/O -*/

    /*
    RS_ATTR_VISA_RM_SESSION         ... rsbtc_RsInit
    RS_ATTR_IO_SESSION              ... Rs_UpdateIOSession
    RS_ATTR_IO_SESSION_TYPE         ... not supported
    */

    /*- Version Info -*/                     

    p2value = RSBTC_MAJOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MAJOR_VERSION, 0, &p2value));
    p2value = RSBTC_MINOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MINOR_VERSION, 0, &p2value));
    p2value = RSBTC_MINOR_MINOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MINOR_MINOR_VERSION, 0, &p2value));
    p2value = RSBTC_CLASS_SPEC_MAJOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_MAJOR_VERSION, 0, &p2value));
    p2value = RSBTC_CLASS_SPEC_MINOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CLASS_DRIVER_MINOR_VERSION, 0, &p2value));
    p2value = RS_ENGINE_MAJOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_ENGINE_MAJOR_VERSION, 0, &p2value));
    p2value = RS_ENGINE_MINOR_VERSION;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_ENGINE_MINOR_VERSION, 0, &p2value));
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_ENGINE_REVISION, 0, RS_ENGINE_REVISION));

    /*- Session Callbacks -*/                

    p2void = rsbtc_WaitForOPCCallback;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_OPC_CALLBACK, 0, &p2void));
    p2void = rsbtc_CheckStatusCallback;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_CHECK_STATUS_CALLBACK, 0, &p2void));
    p2value = RSBTC_OPC_TIMEOUT;
    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_OPC_TIMEOUT, 0, &p2value));

Error:
    return error;
}

/*****************************************************************************
 * Function: rsbtc_DefaultInstrSetup
 * Purpose:  This function sends a default setup to the instrument. The
 *           rsbtc_reset function calls this function. The
 *           rsbtc_RsInit function calls this function when the
 *           user passes VI_FALSE for the reset parameter. This function is
 *           useful for configuring settings that other instrument driver
 *           functions require.
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
ViStatus rsbtc_DefaultInstrSetup (ViSession instrSession)
{
    ViStatus    error = VI_SUCCESS;
	ViChar      msg[RS_MAX_MESSAGE_BUF_SIZE] = "";	

    /* Set all the attributes to the default state. Do not update inherent attributes! */
    checkErr (Rs_ApplyAttributeDefaults (instrSession, VI_FALSE));

    /* Init attributes */
    checkErr( rsbtc_InitAttributes (instrSession));

    if (!Rs_Simulating(instrSession))
        {

        viCheckErr( viPrintf (instrSession, "*CLS;*ESE %hu;*SRE 40\n", IEEE_488_2_ESE_MASK));
		viCheckErr (viQueryf (instrSession, ":SYST:ERR:ALL?\n", "%[^\r\n]", msg));
        }

Error:
    return error;
}

/*****************************************************************************
 * Function: rsbtc_CheckStatus
 * Purpose:  This function checks the status of the instrument to detect
 *           whether the instrument has encountered an error. This function
 *           is called at the end of most exported driver functions. If the
 *           instrument reports an error, this function returns
 *           RS_ERROR_INSTRUMENT_SPECIFIC. The user can set the
 *           rsbtc_ATTR_QUERY_INSTRUMENT_STATUS attribute to VI_FALSE to disable this
 *           check and increase execution speed.
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
ViStatus rsbtc_CheckStatus (ViSession instrSession)
{
    ViStatus    error = VI_SUCCESS;

    if (Rs_QueryInstrStatus (instrSession) && !Rs_Simulating (instrSession))
        {
        checkErr( rsbtc_CheckStatusCallback (instrSession));
        }

Error:
    return error;
}

/*****************************************************************************
 * Function: rsbtc_hardwareRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_hardwareRepCap (ViSession instrSession, ViInt32 board, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (board, 1, 8), 2, "Board");
    sprintf (rep_cap, "hw%d", board);
	
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_pathRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_pathRepCap (ViSession instrSession, ViInt32 path, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (path, 1, 2), 2, "Path");
    sprintf (rep_cap, "hw%d", path);
	
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_A3plpRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_A3plpRepCap (ViSession instrSession, ViInt32 plp, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (plp, 1, 64), 2, "PLP");
    sprintf (rep_cap, "A3PLP%d", plp);
	
Error:    
    return error;
}

 /*****************************************************************************
 * Function: rsbtc_A3SubfRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_A3SubfRepCap (ViSession instrSession, ViInt32 subf, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (subf, 1, 256), 2, "Subframe");
    sprintf (rep_cap, "SF%d", subf);
	
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_plpRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_plpRepCap (ViSession instrSession, ViInt32 plp, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (plp, 1, 20), 2, "PLP");
    sprintf (rep_cap, "PLP%d", plp);
	
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_groupPLPRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_groupPLPRepCap (ViSession instrSession, ViInt32 groupPLP, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    				
    viCheckParm (rsbtc_invalidViInt32Range (groupPLP, 1, 40), 2, "PLP Group");
    sprintf (rep_cap, "PLPG%d", groupPLP);
	
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_dSliceRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_dSliceRepCap (ViSession instrSession, ViInt32 dSlice, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (dSlice, 1, 4), 2, "DSL");
    sprintf (rep_cap, "DSL%d", dSlice - 1);
	
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_notchRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_notchRepCap (ViSession instrSession, ViInt32 notch, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (notch, 1, 4), 2, "Notch");
    sprintf (rep_cap, "N%d", notch - 1);
	
Error:    
    return error;
}
/*****************************************************************************
 * Function: rsbtc_layerRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_layerRepCap (ViSession instrSession, ViInt32 layer, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    ViString    layerArr[] = {"A","B","C"};
    
    viCheckParm (rsbtc_invalidViInt32Range (layer, 0, 2), 2, "Layer");
    
    sprintf (rep_cap, "%s", layerArr[layer]);
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_SegLayerRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_segLayerRepCap (ViSession instrSession, ViInt32 segLayer, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    ViString    segLayerArr[] = {"A1A","A1B","A1C","A2A","A2B","A2C","B1","B2","B3","B4","B5","B6","B7"};
    
    viCheckParm (rsbtc_invalidViInt32Range (segLayer, 0, 12), 2, "SegLayer");
    
    sprintf (rep_cap, "%s", segLayerArr[segLayer]);
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_profileRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_profileRepCap (ViSession instrSession, ViInt32 profile, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
	viCheckParm (rsbtc_invalidViInt32Range (profile, 1, 2), 2, "Profile");
    
    sprintf (rep_cap, "PR%ld", profile);
Error:    
    return error;
}

/*****************************************************************************
 * Function: rsbtc_arbRepCap
 * Purpose:  This function checks instance repeated capability and 
 *           creates proper string.
 *****************************************************************************/
ViStatus rsbtc_arbRepCap (ViSession instrSession, ViInt32 arb, ViChar rep_cap[])
{
    ViStatus    error   = VI_SUCCESS;
    
    viCheckParm (rsbtc_invalidViInt32Range (arb, 1, 8), 2, "ARB");
    sprintf (rep_cap, "A%d", arb);
	
Error:    
    return error;
}
/*===========================================================================*/
/* Function: Long Signed Integer Value Out Of Range - ViInt32                */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViStatus rsbtc_invalidViInt32Range (ViInt32 val, ViInt32 min, ViInt32 max)
{
    return ((val < min || val > max) ? RS_ERROR_INVALID_PARAMETER : VI_SUCCESS);
}
/*===========================================================================*/
/* Function: Long Unsigned Integer Value Out Of Range - ViUInt32             */
/* Purpose:  This function checks a long unsigned integer value to see if it */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViStatus rsbtc_invalidViUInt32Range (ViUInt32 val, ViUInt32 min, ViUInt32 max)
{
    return ((val < min || val > max) ? RS_ERROR_INVALID_PARAMETER : VI_SUCCESS);
}
/*===========================================================================*/
/* Function: Real (Double) Value Out Of Range - ViReal64                     */
/* Purpose:  This function checks a real (double) value to see if it lies    */
/*           between a minimum and maximum value.  If the value is out of    */
/*           range, the return value is VI_TRUE, otherwise the return value  */
/*           is VI_FALSE.                                                    */
/*===========================================================================*/
ViStatus rsbtc_invalidViReal64Range (ViReal64 val, ViReal64 min, ViReal64 max)
{
    return ((val < min || val > max) ? RS_ERROR_INVALID_PARAMETER : VI_SUCCESS);
}
/******************************************************************************
 * Function: Read Data With Unknown length
 * Description: This function allocate buffer and read data from the instrument
 * WARNING:  You must deallocate buffer, if any pointer has been returned.
 *           This function does not work correctly, when you read bin data via
 *           serial interface
 *****************************************************************************/
ViStatus rsbtc_ReadDataUnknownLength(ViSession instrSession,
                ViChar **outBuffer, ViUInt32 *count)
{
    ViStatus error = VI_SUCCESS;
    ViUInt32 actSize, retCnt = 0;
    ViChar   *buffer = NULL, *p2buffer;

    // *outBuffer = NULL;
    if (count) *count = 0;

    /* Allocate Buffer */
    actSize = (ViUInt32) RSBTC_IO_BUFFER_SIZE;
    buffer = (ViChar *)calloc((size_t)actSize, sizeof(ViChar));
    if (buffer == NULL) return VI_ERROR_ALLOC;

    p2buffer = buffer;
    do
        {
        if ((error = viRead (instrSession, (ViBuf) p2buffer, (ViUInt32) RSBTC_IO_BUFFER_SIZE, &retCnt)) < 0)
            {
            free(buffer);
            return error;
            }
        if (retCnt == (ViUInt32) RSBTC_IO_BUFFER_SIZE) /* Buffer was small, reallocate it */
        {
            actSize += RSBTC_IO_BUFFER_SIZE;
            buffer = (ViChar *)realloc(buffer, (size_t) actSize);
            if (buffer == NULL) return VI_ERROR_ALLOC;
            p2buffer = buffer + (actSize - RSBTC_IO_BUFFER_SIZE); /* Set pointer to end of data in rallocated buffer */
        }
        else
        {
            *(p2buffer+retCnt) = '\0'; /* For sure set end of string on the end of data */
        }
       } while (error == 0x3FFF0006); /* if buffer was small, continue with next read */

    *outBuffer = buffer;
    if (count) *count = (actSize - RSBTC_IO_BUFFER_SIZE) + retCnt;

    return error;
}


/*****************************************************************************
 * Function: rsbtc_WaitForOPC
 * Purpose:  This function waits for the instrument to complete the
 *           execution of all pending operations. This function is a
 *           wrapper for the WaitForOPCCallback. It can be called from
 *           other instrument driver functions.
 *
 *           The maxTime parameter specifies the maximum time to wait for
 *           operation complete in milliseconds.
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
ViStatus rsbtc_WaitForOPC (ViSession instrSession, ViInt32 maxTime)
{
    ViStatus    error = VI_SUCCESS;

    if (Rs_Simulating(instrSession) == 0)
        {
        ViInt32  oldOPCTimeout = 0;

        checkErr( Rs_GetAttribute (instrSession, NULL, RSBTC_ATTR_OPC_TIMEOUT, 0,
                                   (ViInt32) sizeof (oldOPCTimeout), &oldOPCTimeout));

        (void)Rs_SetAttribute (instrSession, NULL, RSBTC_ATTR_OPC_TIMEOUT, 0, &maxTime);

        error = rsbtc_WaitForOPCCallback (instrSession);

        (void)Rs_SetAttribute (instrSession, NULL, RSBTC_ATTR_OPC_TIMEOUT, 0, &oldOPCTimeout);

        viCheckErr( error);
        }
Error:
    return error;
}

/*****************************************************************************
 * Function: rsbtc_dashInt_ReadCallback
 * Purpose:  This function overrides standard calback funtion. Returns NAN if
 *           instrument returns "---" or "n/a".
 *****************************************************************************/
ViStatus rsbtc_dashInt_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
	return rsbtc_dashString_ReadCallback(instrSession, repCapName, attributeId, bufSize, value);
}

/*****************************************************************************
 * Function: rsbtc_dashReal_ReadCallback
 * Purpose:  This function overrides standard calback funtion. Returns NAN if
 *           instrument returns "---" or "n/a".
 *****************************************************************************/
ViStatus rsbtc_dashReal_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
	return rsbtc_dashString_ReadCallback(instrSession, repCapName, attributeId, bufSize, value);
}

/*****************************************************************************
 * Function: rsbtc_dashString_ReadCallback
 * Purpose:  This function overrides standard calback funtion. Reads dashes 
 *           returned
 *****************************************************************************/
 ViStatus rsbtc_dashString_ReadCallback(ViSession instrSession, ViConstString repCapName, ViAttr attributeId, ViInt32 bufSize, void *value)
{
    RsSessionPropertiesPtr  sessionProperties = Rs_ViSession (instrSession);
    ViStatus    error           = VI_SUCCESS;
    ViUInt32    count           = bufSize;
    ViInt32     idx             = 0;
    ViChar      *cmd            = NULL,
                *buffer         = NULL;
	ViUInt32    local_count		= 0;
	
    viCheckErr (Rs_GetAttributeIndex (instrSession, attributeId, &idx));

    /* If no command available, exit the function */
    if (sessionProperties -> attr[idx].command == NULL)
        return VI_SUCCESS;

    /* Check if attribute uses repeated capabilities */
    if (sessionProperties -> attr[idx].usercaps == VI_TRUE)
        {
        if (!repCapName)
            viCheckErr (RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);

        if (strlen (repCapName) == 0)
            viCheckErr (RS_ERROR_INVALID_REPEATED_CAPABILITY_NAME);
        }

    /* Build the command string */
    viCheckErr (Rs_BuildCommandString (instrSession, attributeId, idx, repCapName, &cmd));  

    /* --- If the range table for the attribute exists, revert 'command string'
       back to the 'discreteOrMinValue' and return it as attribute value. --- */

    if (sessionProperties -> attr[idx].rangeTable)
        if (sessionProperties -> attr[idx].rangeTable -> type == RS_VAL_DISCRETE)
        {
            ViUInt16    i = 0;
            ViChar      *p2string;
            ViInt32     found;
			
            viCheckAlloc (buffer = (ViChar*) malloc (RS_MAX_MESSAGE_BUF_SIZE));
            count = RS_MAX_MESSAGE_BUF_SIZE;

            if (sessionProperties -> attr[idx].access == RS_VAL_READ_WRITE)
			{
				viCheckErr (viPrintf(instrSession, "%s?\n", cmd));
			}
            else
			{
				viCheckErr (viPrintf(instrSession, "%s\n", cmd));
			}
            	
			viCheckErr (Rs_ReadDataUnknownLength(instrSession, &buffer, &local_count));
			
			if ((strncmp (buffer, "---", 3)) == 0 || (strncmp (buffer, "n/a", 3)) == 0)
			{
				switch (sessionProperties -> attr[idx].dataType)
                {
	                case RS_VAL_INT32:
	                    *(ViInt32 *)value = RSBTC_NAN_VI_INT32;
	                    break;
	                case RS_VAL_REAL64:
	                    *(ViReal64 *)value = RSBTC_NAN;
	                    break;
	                case RS_VAL_BOOLEAN:
						*(ViBoolean *)value = VI_FALSE;
	                    break;
	                case RS_VAL_STRING:
	                case RS_VAL_ADDR:
	                case RS_VAL_SESSION:
	                case RS_VAL_EVENT:
	                case RS_VAL_UNKNOWN_TYPE:
	                default:
	                    viCheckErr (RS_ERROR_INVALID_TYPE);
                }

            	goto Error;
            }
			    
            /* Go over range table values and check if the command string is present */
            do
                {
                p2string = sessionProperties -> attr[idx].rangeTable -> rangeValues[i].cmdString;

                if (sessionProperties -> attr[idx].rangeTable -> rangeValues[i].cmdString == RS_RANGE_TABLE_END_STRING)
                    return RS_ERROR_INVALID_VALUE;

				found = strncmp (buffer, p2string, local_count - 1);

                if (found == 0)
                    break;

                i++;

                } while (i);

            /* Return value from range table */
            switch (sessionProperties -> attr[idx].dataType)
                {
                case RS_VAL_INT32:
                    *(ViInt32 *)value = (ViInt32) sessionProperties -> attr[idx].rangeTable -> rangeValues[i].discreteOrMinValue;
                    break;
                case RS_VAL_REAL64:
                    *(ViReal64 *)value = sessionProperties -> attr[idx].rangeTable -> rangeValues[i].discreteOrMinValue;
                    break;
                case RS_VAL_STRING:
                case RS_VAL_BOOLEAN:
                case RS_VAL_ADDR:
                case RS_VAL_SESSION:
                case RS_VAL_EVENT:
                case RS_VAL_UNKNOWN_TYPE:
                default:
                    viCheckErr (RS_ERROR_INVALID_TYPE);
                }
            
            goto Error;
		}

        /* --- Attribute data type is used for VISA I/O --- */
        switch (sessionProperties -> attr[idx].dataType)
            {
            case RS_VAL_INT32:
            case RS_VAL_REAL64:
			case RS_VAL_BOOLEAN:
            	if (sessionProperties -> attr[idx].access == RS_VAL_READ_WRITE)
				{
	                viCheckErr (viPrintf(instrSession, "%s?\n", cmd));
	            }
				 else
				{
					viCheckErr (viPrintf(instrSession, "%s\n", cmd));
	            }
				   
				viCheckErr (Rs_ReadDataUnknownLength(instrSession, &buffer, &local_count));
				
				if ((strncmp (buffer, "---", 3)) == 0 || (strncmp (buffer, "n/a", 3)) == 0)
				{
					switch (sessionProperties -> attr[idx].dataType)
	                {
		                case RS_VAL_INT32:
		                    *(ViInt32 *)value = RSBTC_NAN_VI_INT32;
		                    break;
		                case RS_VAL_REAL64:
		                    *(ViReal64 *)value = RSBTC_NAN;
		                    break;
		                case RS_VAL_BOOLEAN:
							*(ViBoolean *)value = VI_FALSE;
		                    break;
		                case RS_VAL_STRING:
		                case RS_VAL_ADDR:
		                case RS_VAL_SESSION:
		                case RS_VAL_EVENT:
		                case RS_VAL_UNKNOWN_TYPE:
		                default:
		                    viCheckErr (RS_ERROR_INVALID_TYPE);
	                }
    
	            	goto Error;
	            }
				switch (sessionProperties -> attr[idx].dataType)
	            {
	                case RS_VAL_INT32:
	                    *(ViInt32 *)value =  atoi (buffer);
	                    break;
	                case RS_VAL_REAL64:
	                    *(ViReal64 *)value = atof (buffer);
	                    break;
	                case RS_VAL_BOOLEAN:
						*(ViBoolean *)value =  (ViBoolean) atoi (buffer);
	                    break;
	                case RS_VAL_STRING:
	                case RS_VAL_ADDR:
	                case RS_VAL_SESSION:
	                case RS_VAL_EVENT:
	                case RS_VAL_UNKNOWN_TYPE:
	                default:
	                    viCheckErr (RS_ERROR_INVALID_TYPE);
	            }
			    break;
            case RS_VAL_STRING:
			case RS_VAL_ADDR:
            case RS_VAL_SESSION:
            case RS_VAL_EVENT:
            case RS_VAL_UNKNOWN_TYPE:
            default:
                viCheckErr (RS_ERROR_INVALID_TYPE);
            }

Error:

    if (buffer)
        free (buffer);

    if (cmd)
        free (cmd);

    return error;
}
