/*****************************************************************************
 *  Rohde&Schwarz BTC Broadcast Test Center instrument driver    
 *  VXIpnp Style Instrument Driver
 *  Original Release: 12/2013
 *  By: 
 *
 *  Should you have any technical questions please contact the hotline of
 *  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 *  e-mail: CustomerSupport@rohde-schwarz.com
 *
 *****************************************************************************/

#ifndef __RSBTC_HEADER
#define __RSBTC_HEADER

//#include "rsidr_core.h"
#include "rsbtc_utility.h"
#include "rsbtc_attributes.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#if defined(__cplusplus) || defined(__cplusplus__)
extern "C" {
#endif

/****************************************************************************
 *----------------- Instrument Driver Revision Information -----------------*
 ****************************************************************************/
#define RSBTC_SIMULATION_ID_QUERY      "Rohde&Schwarz,R&S BTC,2114.3000k02/900000,2.30.0.0"  /* Default definition of ID Query Response for simulation */
#define RSBTC_VALID_ID_RESPONSE_STRING "Rohde&Schwarz"      /* Valid response for identification query */
#define RSBTC_SIMULATION_OPT_QUERY     ""      /* Simulated response for *OPT? command */

#define RSBTC_IO_BUFFER_SIZE           1024L   /* I/O buffer size */
#define RSBTC_OPC_TIMEOUT              5000L   /* Maximum time to wait for OPC in milliseconds */
#define IEEE_488_2_ERROR_MASK          0x4    /* 488.2 STB Bits */   
#define IEEE_488_2_ESE_MASK            0x3D    /* 488.2 Event Status Register (ESE) Bits */
#define RSBTC_NAN                      9.91e+37
#define RSBTC_NAN_VI_INT32             0x80000000
#define RSBTC_NAN_VI_INT64             0x8000000000000000 
#define RSBTC_TOKEN_BUFFER_SIZE        50      /* Size for rep capability values */
#define MAX_OPC_TIME                   60000L 

/* Class Driver Identification */

#define RSBTC_CLASS_DRIVER_DESCRIPTION           ""
#define RSBTC_CLASS_DRIVER_PREFIX                ""
#define RSBTC_CLASS_DRIVER_VENDOR                ""
#define RSBTC_CLASS_DRIVER_REVISION              ""
#define RSBTC_CLASS_SPEC_MAJOR_VERSION           1L    /* Class specification major version */
#define RSBTC_CLASS_SPEC_MINOR_VERSION           0L    /* Class specification minor version */

/* Driver Identification */

#define RSBTC_SPECIFIC_DRIVER_DESCRIPTION        ""
#define RSBTC_SPECIFIC_DRIVER_PREFIX             "rsbtc"
#define RSBTC_SPECIFIC_DRIVER_LOCATOR            ""
#define RSBTC_SPECIFIC_DRIVER_VENDOR             ""
#define RSBTC_MAJOR_VERSION                      2L    /* Instrument driver major version          */
#define RSBTC_MINOR_VERSION                      32L    /* Instrument driver minor version          */
#define RSBTC_MINOR_MINOR_VERSION                0L    /* Instrument driver minor minor version    */

/* Driver Capabilities */          

#define RSBTC_SUPPORTED_INSTRUMENT_MODELS        ""    /* Instrument driver supported model(s)     */
#define RSBTC_GROUP_CAPABILITIES                 ""
#define RSBTC_FUNCTION_CAPABILITIES              ""

/****************************************************************************
 *---------------------------- Attribute Defines ---------------------------*
 ****************************************************************************/

    /*- Inherent Instrument Specific Attributes ----------------------------*/

        /* User Options */

#define RSBTC_ATTR_RANGE_CHECK                                           RS_ATTR_RANGE_CHECK                                     /* ViBoolean */
#define RSBTC_ATTR_QUERY_INSTRUMENT_STATUS                               RS_ATTR_QUERY_INSTRUMENT_STATUS                         /* ViBoolean */
#define RSBTC_ATTR_CACHE                                                 RS_ATTR_CACHE                                           /* ViBoolean */
#define RSBTC_ATTR_SIMULATE                                              RS_ATTR_SIMULATE                                        /* ViBoolean */
#define RSBTC_ATTR_RECORD_COERCIONS                                      RS_ATTR_RECORD_COERCIONS                                /* ViBoolean */
#define RSBTC_ATTR_INTERCHANGE_CHECK                                     RS_ATTR_INTERCHANGE_CHECK                               /* ViBoolean */
#define RSBTC_ATTR_SPY                                                   RS_ATTR_SPY                                             /* ViBoolean */
#define RSBTC_ATTR_USE_SPECIFIC_SIMULATION                               RS_ATTR_USE_SPECIFIC_SIMULATION                         /* ViBoolean */

        /* Class Driver Identification */

#define RSBTC_ATTR_CLASS_DRIVER_DESCRIPTION                              RS_ATTR_CLASS_DRIVER_DESCRIPTION                        /* ViString, read-only */
#define RSBTC_ATTR_CLASS_DRIVER_PREFIX                                   RS_ATTR_CLASS_DRIVER_PREFIX                             /* ViString, read-only */
#define RSBTC_ATTR_CLASS_DRIVER_VENDOR                                   RS_ATTR_CLASS_DRIVER_VENDOR                             /* ViString, read-only */
#define RSBTC_ATTR_CLASS_DRIVER_REVISION                                 RS_ATTR_CLASS_DRIVER_REVISION                           /* ViString, read-only */
#define RSBTC_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION                 RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MAJOR_VERSION           /* ViInt32,  read-only */
#define RSBTC_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION                 RS_ATTR_CLASS_DRIVER_CLASS_SPEC_MINOR_VERSION           /* ViInt32,  read-only */

        /* Driver Identification */

#define RSBTC_ATTR_SPECIFIC_DRIVER_DESCRIPTION                           RS_ATTR_SPECIFIC_DRIVER_DESCRIPTION                     /* ViString, read-only */
#define RSBTC_ATTR_SPECIFIC_DRIVER_PREFIX                                RS_ATTR_SPECIFIC_DRIVER_PREFIX                          /* ViString, read-only */
#define RSBTC_ATTR_SPECIFIC_DRIVER_LOCATOR                               RS_ATTR_SPECIFIC_DRIVER_LOCATOR                         /* ViString, read-only */
#define RSBTC_ATTR_SPECIFIC_DRIVER_VENDOR                                RS_ATTR_SPECIFIC_DRIVER_VENDOR                          /* ViString, read-only */
#define RSBTC_ATTR_SPECIFIC_DRIVER_REVISION                              RS_ATTR_SPECIFIC_DRIVER_REVISION                        /* ViString, read-only */
#define RSBTC_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION              RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION        /* ViInt32,  read-only */
#define RSBTC_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION              RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION        /* ViInt32,  read-only */

        /* Driver Capabilities */

#define RSBTC_ATTR_SUPPORTED_INSTRUMENT_MODELS                           RS_ATTR_SUPPORTED_INSTRUMENT_MODELS                     /* ViString, read-only */
#define RSBTC_ATTR_GROUP_CAPABILITIES                                    RS_ATTR_GROUP_CAPABILITIES                              /* ViString, read-only */
#define RSBTC_ATTR_FUNCTION_CAPABILITIES                                 RS_ATTR_FUNCTION_CAPABILITIES                           /* ViString, read-only */
#define RSBTC_ATTR_CHANNEL_COUNT                                         RS_ATTR_CHANNEL_COUNT                                   /* ViInt32,  read-only */

        /* Driver Setup */

#define RSBTC_ATTR_DRIVER_SETUP                                          RS_ATTR_DRIVER_SETUP                                    /* ViString            */

        /* Instrument Identification */

#define RSBTC_ATTR_INSTRUMENT_MANUFACTURER                               RS_ATTR_INSTRUMENT_MANUFACTURER                         /* ViString, read-only */
#define RSBTC_ATTR_INSTRUMENT_MODEL                                      RS_ATTR_INSTRUMENT_MODEL                                /* ViString, read-only */
#define RSBTC_ATTR_INSTRUMENT_FIRMWARE_REVISION                          RS_ATTR_INSTRUMENT_FIRMWARE_REVISION                    /* ViString, read-only */

        /* Advanced Session Information */

#define RSBTC_ATTR_IO_RESOURCE_DESCRIPTOR                                RS_ATTR_IO_RESOURCE_DESCRIPTOR                          /* ViString, read-only */
#define RSBTC_ATTR_LOGICAL_NAME                                          RS_ATTR_LOGICAL_NAME                                    /* ViString, read-only */

	
	/*- Instrument-Specific Attributes -------------------------------------*/
/* BEGIN GENERATE */
#define RSBTC_ATTR_BASEBAND_SIGNAL_TYPE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 2
#define RSBTC_ATTR_BASEBAND_SIGNAL_TRANSMISSION_DTV_STANDARD                RS_SPECIFIC_PUBLIC_ATTR_BASE + 3
#define RSBTC_ATTR_BASEBAND_SIGNAL_TRANSMISSION_ATV_STANDARD                RS_SPECIFIC_PUBLIC_ATTR_BASE + 4
#define RSBTC_ATTR_BASEBAND_SIGNAL_TRANSMISSION_AUDIO_BC_STANDARD           RS_SPECIFIC_PUBLIC_ATTR_BASE + 5
#define RSBTC_ATTR_BASEBAND_SIGNAL_FREQUENCY_OFFSET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 6
#define RSBTC_ATTR_BASEBAND_SIGNAL_POLARITY                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1299
#define RSBTC_ATTR_BASEBAND_SIGNAL_AWGN                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1190
#define RSBTC_ATTR_BASEBAND_SIGNAL_AWGN_CN                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1191
#define RSBTC_ATTR_BASEBAND_SIGNAL_AWGN_EB_N0_RATIO                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1192
#define RSBTC_ATTR_BASEBAND_SIGNAL_AWGN_SETTINGS_BANDWIDTH_COUPLING         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1193
#define RSBTC_ATTR_BASEBAND_SIGNAL_AWGN_SETTINGS_RECEIVER_BANDWIDTH         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1194
#define RSBTC_ATTR_BASEBAND_SIGNAL_AWGN_SETTINGS_GENERATED_NOISE_BW         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1195
#define RSBTC_ATTR_IP_INPUT                                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1225
#define RSBTC_ATTR_IP_INPUT_ALIAS                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1226
#define RSBTC_ATTR_IP_INPUT_TYPE                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1227
#define RSBTC_ATTR_IP_INPUT_MULTICAST_ADDRESS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1228
#define RSBTC_ATTR_IP_INPUT_PORT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1229
#define RSBTC_ATTR_IP_INPUT_IGMPV3_SOURCE_ADDRESS                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1230
#define RSBTC_ATTR_IP_INPUT_PING_SOURCE_ADDRESS                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1231
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_INTERFACE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 7
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_SOURCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 8
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_INPUT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 9
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_IP_TS_CHANNEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 10
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_INPUT_FORMAT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 11
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PID                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 12
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_SID                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 13
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_NUMBER_OF_PLP                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 14
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_PLP_INPUT_FORMAT                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 15
#define RSBTC_ATTR_READ_DVBT2_INPUT_SIGNAL_PLP_PACKET_LENGTH                RS_SPECIFIC_PUBLIC_ATTR_BASE + 16
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_PLP_STUFFING                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 17
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_PLP_TEST_SIGNAL                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 18
#define RSBTC_ATTR_READ_DVBT2_INPUT_SIGNAL_PLP_MAX_USEFUL_DATA_RATE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 19
#define RSBTC_ATTR_READ_DVBT2_INPUT_SIGNAL_PLP_MEASURED_DATA_RATE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 20
#define RSBTC_ATTR_READ_DVBT2_INPUT_SIGNAL_PLP_USEFUL_DATA_RATE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 21
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_ANALYZER                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 22
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_ANALYZER_RESET_LOG_FILE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 23
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_MEAS_MODE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 24
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMIN1       RS_SPECIFIC_PUBLIC_ATTR_BASE + 25
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMIN2       RS_SPECIFIC_PUBLIC_ATTR_BASE + 26
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMIN3       RS_SPECIFIC_PUBLIC_ATTR_BASE + 27
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMAX1       RS_SPECIFIC_PUBLIC_ATTR_BASE + 28
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMAX2       RS_SPECIFIC_PUBLIC_ATTR_BASE + 29
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMAX3       RS_SPECIFIC_PUBLIC_ATTR_BASE + 30
#define RSBTC_ATTR_DVBT2_INPUT_SIGNAL_T2_MI_PACKET_TIMING_T2_MI_TMAX4       RS_SPECIFIC_PUBLIC_ATTR_BASE + 31
#define RSBTC_ATTR_DVBT2_CONFIGURING_TEST_TS_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 32
#define RSBTC_ATTR_DVBT2_CONFIGURING_PID_TEST_PACKET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 33
#define RSBTC_ATTR_DVBT2_CONFIGURING_PID                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 34
#define RSBTC_ATTR_DVBT2_CONFIGURING_PAYLOAD_TEST_PACKET                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 35
#define RSBTC_ATTR_DVBT2_CONFIGURING_PRBS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 36
#define RSBTC_ATTR_DVBT2_MODE_STREAM_PLP_ID                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 37
#define RSBTC_ATTR_DVBT2_MODE_STREAM_PLP_GROUP_ID                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 38
#define RSBTC_ATTR_DVBT2_MODE_STREAM_PLP_TYPE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 39
#define RSBTC_ATTR_DVBT2_MODE_STREAM_BB_MODE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 40
#define RSBTC_ATTR_DVBT2_MODE_STREAM_ISSY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 41
#define RSBTC_ATTR_DVBT2_MODE_STREAM_NULL_PACKET_DELETION                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 42
#define RSBTC_ATTR_DVBT2_MODE_STREAM_IN_BAND_SIGNALING                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 43
#define RSBTC_ATTR_DVBT2_MODE_STREAM_IN_BAND_SIGNALING_A                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 44
#define RSBTC_ATTR_DVBT2_MODE_STREAM_IN_BAND_SIGNALING_B                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 45
#define RSBTC_ATTR_DVBT2_MODE_STREAM_OTHER_IN_BAND_PLPS                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 46
#define RSBTC_ATTR_DVBT2_MODE_STREAM_CM_TYPE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 47
#define RSBTC_ATTR_DVBT2_MODE_STREAM_STATIC_FLAG                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 48
#define RSBTC_ATTR_DVBT2_MODE_STREAM_STATIC_PADDING_FLAG                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 49
#define RSBTC_ATTR_DVBT2_BICM_FEC_FRAME                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 50
#define RSBTC_ATTR_DVBT2_BICM_CODE_RATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 51
#define RSBTC_ATTR_DVBT2_BICM_CONSTELLATION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 52
#define RSBTC_ATTR_DVBT2_BICM_CONSTELLATION_ROTATION                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 53
#define RSBTC_ATTR_DVBT2_BICM_TIME_INTERLEAVER_TYPE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 54
#define RSBTC_ATTR_DVBT2_BICM_FRAME_INTERVAL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 55
#define RSBTC_ATTR_DVBT2_BICM_TIME_INTERLEAVER_LENGTH                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 56
#define RSBTC_ATTR_DVBT2_BICM_FIRST_FRAME_INDEX                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 57
#define RSBTC_ATTR_DVBT2_BICM_FEC_BLOCKS_PER_IF                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 58
#define RSBTC_ATTR_DVBT2_BICM_FEC_MAX_NUMBER_OF_BLOCKS                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 59
#define RSBTC_ATTR_DVBT2_FRAMING_CHANNEL_BANDWIDTH                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 60
#define RSBTC_ATTR_DVBT2_FRAMING_USED_BANDWIDTH                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 61
#define RSBTC_ATTR_DVBT2_FRAMING_BANDWIDTH_VARIATION                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 62
#define RSBTC_ATTR_DVBT2_FRAMING_FFT_SIZE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 63
#define RSBTC_ATTR_DVBT2_FRAMING_GUARD_INTERVAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 64
#define RSBTC_ATTR_DVBT2_FRAMING_PILOT_PATTERN                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 65
#define RSBTC_ATTR_DVBT2_FRAMING_T2_FRAMES_PER_SUPER_FRAME                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 66
#define RSBTC_ATTR_DVBT2_FRAMING_OFDM_SYMBOLS_PER_T2_FRAME                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 67
#define RSBTC_ATTR_DVBT2_FRAMING_DATA_SYMBOLS_PER_T2_FRAME                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 68
#define RSBTC_ATTR_DVBT2_FRAMING_SUBSLICES_PER_T2_FRAME                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 69
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_NETWORK_MODE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 70
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_PROFILE_MODE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 71
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_TRANSMISSION_SYSTEM                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 72
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_MISO_GROUP_MODE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 73
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_MISO_GROUP                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 74
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_PEAK_TO_AVRAGE_POWER_RATIO               RS_SPECIFIC_PUBLIC_ATTR_BASE + 75
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_FUTURE_EXTENSION_FRAMES                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 76
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_FEF_TYPE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 77
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_FEF_LENGTH                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 78
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_FEF_INTERVAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 79
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_FEF_PAYLOAD                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 80
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_TIME_FREQUENCY_SLICING                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 81
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_T2_VERSION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 82
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_POST_MODULATION                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 83
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_REPETITION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 84
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_POST_EXTENSION                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 85
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_POST_SCRAMBLED                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 86
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_T2_BASE_LITE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 87
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_NUM_AUX_STREAMS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 88
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_RF_SIGNALLING                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 89
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_L1_SIGNALLING_FREQUENCY                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 90
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_CELL_ID                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 91
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_NETWORK_ID                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 92
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_ID                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 93
#define RSBTC_ATTR_DVBT2_T2_SYSTEM_TX_ID_AVAILABILITY                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 94
#define RSBTC_ATTR_DVBT2_SFN_PROCESS_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 95
#define RSBTC_ATTR_DVBT2_SFN_STATIC_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 96
#define RSBTC_ATTR_DVBT2_SFN_DYNAMIC_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 97
#define RSBTC_ATTR_DVBT2_SFN_TOTAL_DELAY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 98
#define RSBTC_ATTR_DVBT2_SFN_MAX_DEVIATION_TIME                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 99
#define RSBTC_ATTR_DVBT2_SFN_MUTE_FIRST_SUPER_FRAME_P1                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 100
#define RSBTC_ATTR_DVBT2_SFN_1PPS_ROUTING                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1087
#define RSBTC_ATTR_DVBT2_SFN_TIMESTAMP_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 101
#define RSBTC_ATTR_DVBT2_SFN_UPDATE_TIMESTAMP                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 102
#define RSBTC_ATTR_DVBT2_SFN_TIMESTAMP_MODE_UTC_DATE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 103
#define RSBTC_ATTR_DVBT2_SFN_TIMESTAMP_MODE_UTC_TIME                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 104
#define RSBTC_ATTR_DVBT2_SFN_TIMESTAMP_MODE_OFFSET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 105
#define RSBTC_ATTR_DVBT2_SFN_TIMESTAMP_MODE_SECONDS_SINCE_2000              RS_SPECIFIC_PUBLIC_ATTR_BASE + 106
#define RSBTC_ATTR_DVBT2_FRAME_INFO_SUPER_FRAME_DURATION                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 107
#define RSBTC_ATTR_DVBT2_FRAME_INFO_T2_FRAME_DURATION                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 108
#define RSBTC_ATTR_DVBT2_FRAME_INFO_FUTURE_EXTENSION_FRAME_DURATION         RS_SPECIFIC_PUBLIC_ATTR_BASE + 109
#define RSBTC_ATTR_DVBT2_FRAME_INFO_P1_SYMBOL_DURATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 110
#define RSBTC_ATTR_DVBT2_FRAME_INFO_OFDM_SYMBOL_DURATION_TP2                RS_SPECIFIC_PUBLIC_ATTR_BASE + 111
#define RSBTC_ATTR_DVBT2_FRAME_INFO_OFDM_SYMBOL_DURATION_TS                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 112
#define RSBTC_ATTR_DVBT2_FRAME_INFO_L1_PRE_SIGNALING_BITS                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 113
#define RSBTC_ATTR_DVBT2_FRAME_INFO_L1_PRE_SIGNALING_CELLS                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 114
#define RSBTC_ATTR_DVBT2_FRAME_INFO_L1_POST_SIGNALING_BITS                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 115
#define RSBTC_ATTR_DVBT2_FRAME_INFO_L1_POST_SIGNALING_CELLS                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 116
#define RSBTC_ATTR_DVBT2_FRAME_INFO_PLP_DATA_CELLS_PER_T2_FRAME             RS_SPECIFIC_PUBLIC_ATTR_BASE + 117
#define RSBTC_ATTR_DVBT2_FRAME_INFO_PLP_DATA_CELLS_IN_CURRENT_T2_FRAME      RS_SPECIFIC_PUBLIC_ATTR_BASE + 118
#define RSBTC_ATTR_DVBC2_INPUT_SIGNAL_NUMBER_OF_PLP                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 119
#define RSBTC_ATTR_DVBC2_INPUT_SIGNAL_INPUT_FORMAT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 120
#define RSBTC_ATTR_DVBC2_INPUT_SIGNAL_SOURCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 121
#define RSBTC_ATTR_DVBC2_INPUT_SIGNAL_INPUT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 122
#define RSBTC_ATTR_DVBC2_INPUT_SIGNAL_IP_TS_CHANNEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 123
#define RSBTC_ATTR_READ_DVBC2_INPUT_SIGNAL_PACKET_LENGTH                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 124
#define RSBTC_ATTR_READ_DVBC2_INPUT_SIGNAL_MAX_DATA_RATE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 125
#define RSBTC_ATTR_READ_DVBC2_INPUT_SIGNAL_MEASURED_DATA_RATE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 126
#define RSBTC_ATTR_READ_DVBC2_INPUT_SIGNAL_USEFUL_DATA_RATE                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 127
#define RSBTC_ATTR_DVBC2_INPUT_SIGNAL_TEST_SIGNAL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 128
#define RSBTC_ATTR_DVBC2_CONFIGURING_TEST_TS_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 129
#define RSBTC_ATTR_DVBC2_CONFIGURING_PID_TEST_PACKET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 130
#define RSBTC_ATTR_DVBC2_CONFIGURING_PID                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 131
#define RSBTC_ATTR_DVBC2_CONFIGURING_PAYLOAD_TEST_TS_PACKET                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 132
#define RSBTC_ATTR_DVBC2_CONFIGURING_PRBS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 133
#define RSBTC_ATTR_DVBC2_MODE_STREAM_ADAPTATION_PLP_ID                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 134
#define RSBTC_ATTR_DVBC2_MODE_STREAM_ADAPTATION_PLP_BUNDLING                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1232
#define RSBTC_ATTR_DVBC2_MODE_STREAM_ADAPTATION_PLP_TYPE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 135
#define RSBTC_ATTR_DVBC2_MODE_STREAM_ADAPTATION_BB_MODE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 136
#define RSBTC_ATTR_DVBC2_MODE_STREAM_ADAPTATION_ISSY                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 137
#define RSBTC_ATTR_DVBC2_MODE_STREAM_ADAPTATION_NULL_PACKET_DELETION        RS_SPECIFIC_PUBLIC_ATTR_BASE + 138
#define RSBTC_ATTR_DVBC2_BICM_FEC_FRAME                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 139
#define RSBTC_ATTR_DVBC2_BICM_CODE_RATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 140
#define RSBTC_ATTR_DVBC2_BICM_CONSTELLATION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 141
#define RSBTC_ATTR_DVBC2_BICM_XFEC_FRAME_LENGTH                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 143
#define RSBTC_ATTR_DVBC2_BICM_MAPPED_TO_DATA_SLICE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 144
#define RSBTC_ATTR_DVBC2_DATA_SLICE_MIN_REL                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 145
#define RSBTC_ATTR_DVBC2_DATA_SLICE_MAX_REL                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 146
#define RSBTC_ATTR_DVBC2_DATA_SLICE_MIN_ABS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 147
#define RSBTC_ATTR_DVBC2_DATA_SLICE_MAX_ABS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 148
#define RSBTC_ATTR_DVBC2_DATA_SLICE_NUMBER_OF_DATA_SLICES                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 149
#define RSBTC_ATTR_DVBC2_DATA_SLICE_ID                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 150
#define RSBTC_ATTR_DVBC2_DATA_SLICE_TI_DEPTH                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 151
#define RSBTC_ATTR_DVBC2_DATA_SLICE_TYPE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 152
#define RSBTC_ATTR_DVBC2_DATA_SLICE_FEC_HEADER_TYPE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 153
#define RSBTC_ATTR_DVBC2_DATA_SLICE_NUMBER_OF_MAPPED_PLPS                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 154
#define RSBTC_ATTR_DVBC2_DATA_SLICE_OFFSET_LEFT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 155
#define RSBTC_ATTR_DVBC2_DATA_SLICE_TUNE_POSITION_REL                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 156
#define RSBTC_ATTR_DVBC2_DATA_SLICE_OFFSET_RIGHT                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 157
#define RSBTC_ATTR_DVBC2_DATA_SLICE_START_REL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 158
#define RSBTC_ATTR_DVBC2_DATA_SLICE_STOP_REL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 159
#define RSBTC_ATTR_DVBC2_DATA_SLICE_WIDTH                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 160
#define RSBTC_ATTR_DVBC2_DATA_SLICE_START_ABS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 161
#define RSBTC_ATTR_DVBC2_DATA_SLICE_TUNE_POSITION_ABS                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 162
#define RSBTC_ATTR_DVBC2_DATA_SLICE_STOP_ABS                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 163
#define RSBTC_ATTR_DVBC2_DATA_SLICE_START_FREQUENCY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 164
#define RSBTC_ATTR_DVBC2_DATA_SLICE_TUNE_FREQUENCY                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 165
#define RSBTC_ATTR_DVBC2_DATA_SLICE_STOP_FREQUENCY                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 166
#define RSBTC_ATTR_DVBC2_BICM_NUMBER_OF_XFEC_FRAMES                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 142
#define RSBTC_ATTR_DVBC2_NOTCH_NUMBER_OF_NOTCHES                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 167
#define RSBTC_ATTR_DVBC2_NOTCH_START_REL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 168
#define RSBTC_ATTR_DVBC2_NOTCH_STOP_REL                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 169
#define RSBTC_ATTR_DVBC2_NOTCH_WIDTH                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 170
#define RSBTC_ATTR_DVBC2_NOTCH_TYPE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 171
#define RSBTC_ATTR_DVBC2_NOTCH_START_ABS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 172
#define RSBTC_ATTR_DVBC2_NOTCH_STOP_ABS                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 173
#define RSBTC_ATTR_DVBC2_NOTCH_START_FREQUENCY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 174
#define RSBTC_ATTR_DVBC2_NOTCH_STOP_FREQUENCY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 175
#define RSBTC_ATTR_DVBC2_FRAMING_CHANNEL_BANDWIDTH                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 176
#define RSBTC_ATTR_DVBC2_FRAMING_C2_BANDWIDTH                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 177
#define RSBTC_ATTR_DVBC2_FRAMING_USED_BANDWIDTH                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 178
#define RSBTC_ATTR_DVBC2_FRAMING_GUARD_INTERVAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 179
#define RSBTC_ATTR_DVBC2_FRAMING_BASEBAND_FILTER                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 180
#define RSBTC_ATTR_DVBC2_FRAMING_DX                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 181
#define RSBTC_ATTR_DVBC2_FRAMING_START_FREQUENCY                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 182
#define RSBTC_ATTR_DVBC2_FRAMING_CALC_CENTER_FREQUENCY                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 183
#define RSBTC_ATTR_DVBC2_FRAMING_ADJUST_RF_FREQUENCY                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 184
#define RSBTC_ATTR_DVBC2_FRAMING_CARRIER_SPACING                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 185
#define RSBTC_ATTR_DVBC2_FRAMING_NUMBER_OF_PREAMBLE_SYMBOLS                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 186
#define RSBTC_ATTR_DVBC2_C2_SYSTEM_NETWORK_ID                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 187
#define RSBTC_ATTR_DVBC2_C2_SYSTEM_ID                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 188
#define RSBTC_ATTR_DVBC2_C2_SYSTEM_L1_TIME_INTERLEAVER                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 189
#define RSBTC_ATTR_DVBC2_C2_SYSTEM_RESERVED_TONES                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 190
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 191
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_SOURCE_LP                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 192
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 193
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_INPUT_LP                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 194
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 195
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 196
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 197
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_STUFFING_LP                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 198
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_PACKET_LENGTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 199
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_PACKET_LENGTH_LP                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 200
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 201
#define RSBTC_ATTR_DVBT_INPUT_SIGNAL_TEST_SIGNAL_LP                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 202
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 203
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE_LP           RS_SPECIFIC_PUBLIC_ATTR_BASE + 204
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_MEASURED_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 205
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_MEASURED_DATA_RATE_LP             RS_SPECIFIC_PUBLIC_ATTR_BASE + 206
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_USEFUL_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 207
#define RSBTC_ATTR_READ_DVBT_INPUT_SIGNAL_USEFUL_DATA_RATE_LP               RS_SPECIFIC_PUBLIC_ATTR_BASE + 208
#define RSBTC_ATTR_DVBT_CONFIGURING_TEST_TS_PACKET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 209
#define RSBTC_ATTR_DVBT_CONFIGURING_PID_TEST_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 210
#define RSBTC_ATTR_DVBT_CONFIGURING_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 211
#define RSBTC_ATTR_DVBT_CONFIGURING_PAYLOAD_TEST_STUFF                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 212
#define RSBTC_ATTR_DVBT_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 213
#define RSBTC_ATTR_DVBT_SYSTEM_NETWORK_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1088
#define RSBTC_ATTR_DVBT_CODING_CONTROL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1089
#define RSBTC_ATTR_DVBT_CODING_CHANNEL_BANDWIDTH                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 214
#define RSBTC_ATTR_DVBT_CODING_USED_BANDWIDTH                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 215
#define RSBTC_ATTR_DVBT_CODING_FFT_MODE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 216
#define RSBTC_ATTR_DVBT_CODING_GUARD_INTERVAL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 217
#define RSBTC_ATTR_DVBT_CODING_DVBH_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 218
#define RSBTC_ATTR_DVBT_CODING_DVBH_SYMBOL_INTERLEAVER                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 219
#define RSBTC_ATTR_DVBT_CODING_HIERARCHY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 220
#define RSBTC_ATTR_DVBT_CODING_CONSTELLATION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 221
#define RSBTC_ATTR_DVBT_CODING_CODE_RATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 222
#define RSBTC_ATTR_DVBT_CODING_CODE_RATE_LP                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 223
#define RSBTC_ATTR_DVBT_SETTINGS_TPS_CELL_ID                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 224
#define RSBTC_ATTR_DVBT_SETTINGS_TPS_RESERVED_STATE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 225
#define RSBTC_ATTR_DVBT_SETTINGS_TPS_RESERVED_BITS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 226
#define RSBTC_ATTR_DVBT_SETTINGS_TIME_SLICING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 227
#define RSBTC_ATTR_DVBT_SETTINGS_TIME_SLICING_LP                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 228
#define RSBTC_ATTR_DVBT_SETTINGS_MPE_FEC                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 229
#define RSBTC_ATTR_DVBT_SETTINGS_MPE_FEC_LP                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 230
#define RSBTC_ATTR_DVBT_SFN_NETWORK_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1090
#define RSBTC_ATTR_DVBT_SFN_PROCESSING_DELAY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1091
#define RSBTC_ATTR_DVBT_SFN_DYNAMIC_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1092
#define RSBTC_ATTR_DVBT_SFN_TOTAL_DELAY                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1093
#define RSBTC_ATTR_DVBT_SFN_MAX_DEVIATION_TIME                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1094
#define RSBTC_ATTR_DVBT_SFN_STATIC_DELAY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1095
#define RSBTC_ATTR_DVBT_SFN_MAXIMUM_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1096
#define RSBTC_ATTR_DVBT_SFN_DISPATCH_TIME                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1097
#define RSBTC_ATTR_DVBT_SFN_MUTE_1ST_OFDM_SYMBOL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1098
#define RSBTC_ATTR_DVBT_SFN_1PPS_ROUTING                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1233
#define RSBTC_ATTR_DVBT_SFN_OFFSET                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1234
#define RSBTC_ATTR_DVBT_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 231
#define RSBTC_ATTR_DVBT_SPECIAL_SCRAMBLER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 232
#define RSBTC_ATTR_DVBC_SPECIAL_SCRAMBLER_LP                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 233
#define RSBTC_ATTR_DVBT_SPECIAL_SYNC_BYTE_INVERSION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 234
#define RSBTC_ATTR_DVBT_SPECIAL_SYNC_BYTE_INVERSION_LP                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 235
#define RSBTC_ATTR_DVBT_SPECIAL_REED_SOLOMON                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 236
#define RSBTC_ATTR_DVBT_SPECIAL_REED_SOLOMON_LP                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 237
#define RSBTC_ATTR_DVBT_SPECIAL_CONVOLUTIONAL_INTERLEAVER                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 238
#define RSBTC_ATTR_DVBT_SPECIAL_CONVOLUTIONAL_INTERLEAVER_LP                RS_SPECIFIC_PUBLIC_ATTR_BASE + 239
#define RSBTC_ATTR_DVBT_SPECIAL_BIT_INTERLEAVER                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 240
#define RSBTC_ATTR_DVBT_SPECIAL_SYMBOL_INTERLEAVER                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 241
#define RSBTC_ATTR_DVBT_SPECIAL_GUARD_INSERTION                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 242
#define RSBTC_ATTR_DVBT_SPECIAL_DISABLE_CARRIERS                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 250
#define RSBTC_ATTR_DVBT_SPECIAL_LOAD_CARRIERS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 251
#define RSBTC_ATTR_DVBT_SPECIAL_STORE_CARRIERS                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 252
#define RSBTC_ATTR_DVBC_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 253
#define RSBTC_ATTR_DVBC_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 254
#define RSBTC_ATTR_DVBC_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 255
#define RSBTC_ATTR_DVBC_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 256
#define RSBTC_ATTR_DVBC_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 257
#define RSBTC_ATTR_READ_DVBC_INPUT_SIGNAL_PACKET_LENGTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 258
#define RSBTC_ATTR_DVBC_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 259
#define RSBTC_ATTR_READ_DVBC_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 260
#define RSBTC_ATTR_READ_DVBC_INPUT_SIGNAL_MEASURED_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 261
#define RSBTC_ATTR_READ_DVBC_INPUT_SIGNAL_USEFUL_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 262
#define RSBTC_ATTR_DVBC_CONFIGURING_TEST_TS_PACKET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 263
#define RSBTC_ATTR_DVBC_CONFIGURING_PID_TEST_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 264
#define RSBTC_ATTR_DVBC_CONFIGURING_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 265
#define RSBTC_ATTR_DVBC_CONFIGURING_PAYLOAD_TEST_STUFF                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 266
#define RSBTC_ATTR_DVBC_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 267
#define RSBTC_ATTR_DVBC_CODING_SYMBOL_RATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 268
#define RSBTC_ATTR_DVBC_CODING_CONSTELLATION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 269
#define RSBTC_ATTR_DVBC_CODING_ROLL_OFF                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 270
#define RSBTC_ATTR_DVBC_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 271
#define RSBTC_ATTR_DVBC_SPECIAL_SCRAMBLER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 272
#define RSBTC_ATTR_DVBC_SPECIAL_SYNC_BYTE_INVERSION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 273
#define RSBTC_ATTR_DVBC_SPECIAL_REED_SOLOMON                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 274
#define RSBTC_ATTR_DVBC_SPECIAL_CONVOLUTIONAL_INTERLEAVER                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 275
#define RSBTC_ATTR_DVBS_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 276
#define RSBTC_ATTR_DVBS_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 277
#define RSBTC_ATTR_DVBS_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 278
#define RSBTC_ATTR_DVBS_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 279
#define RSBTC_ATTR_DVBS_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 280
#define RSBTC_ATTR_READ_DVBS_INPUT_SIGNAL_PACKET_LENGTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 281
#define RSBTC_ATTR_DVBS_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 282
#define RSBTC_ATTR_READ_DVBS_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 283
#define RSBTC_ATTR_READ_DVBS_INPUT_SIGNAL_MEASURED_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 284
#define RSBTC_ATTR_READ_DVBS_INPUT_SIGNAL_USEFUL_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 285
#define RSBTC_ATTR_DVBS_CONFIGURING_TEST_TS_PACKET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 286
#define RSBTC_ATTR_DVBS_CONFIGURING_PID_TEST_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 287
#define RSBTC_ATTR_DVBS_CONFIGURING_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 288
#define RSBTC_ATTR_DVBS_CONFIGURING_PAYLOAD_TEST_STUFF                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 289
#define RSBTC_ATTR_DVBS_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 290
#define RSBTC_ATTR_DVBS_CODING_SYMBOL_RATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 291
#define RSBTC_ATTR_DVBS_CODING_CONSTELLATION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 292
#define RSBTC_ATTR_DVBS_CODING_ROLL_OFF                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 293
#define RSBTC_ATTR_DVBS_CODING_CODE_RATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 294
#define RSBTC_ATTR_DVBS_CID_ENABLED                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1300
#define RSBTC_ATTR_DVBS_CID_HOST_SIGNAL                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1301
#define RSBTC_ATTR_DVBS_CID_FREQUENCY_OFFSET                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1302
#define RSBTC_ATTR_DVBS_CID_CHIP_RATE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1303
#define RSBTC_ATTR_DVBS_CID_DELTA_PSD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1304
#define RSBTC_ATTR_DVBS_CID_GLOBAL_UNIQUE_ID                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1305
#define RSBTC_ATTR_DVBS_CID_GLOBAL_UNIQUE_ID_CRC                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1306
#define RSBTC_ATTR_DVBS_CID_CONTENT_LATITUDE_LONGITUDE_ENABLED              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1307
#define RSBTC_ATTR_DVBS_CID_CONTENT_LATITUDE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1308
#define RSBTC_ATTR_DVBS_CID_CONTENT_LONGITUDE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1309
#define RSBTC_ATTR_DVBS_CID_CONTENT_TELEPHONE_ENABLED                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1310
#define RSBTC_ATTR_DVBS_CID_CONTENT_TELEPHONE_NUMBER                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1311
#define RSBTC_ATTR_DVBS_CID_CONTENT_TELEPHONE_EXTENSION_ENABLED             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1312
#define RSBTC_ATTR_DVBS_CID_CONTENT_TELEPHONE_EXTENSION_NUMBER              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1313
#define RSBTC_ATTR_DVBS_PHASE_NOISE_ENABLED                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1314
#define RSBTC_ATTR_DVBS_PHASE_NOISE_CW_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1315
#define RSBTC_ATTR_DVBS_PHASE_NOISE_MAGNITUDE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1316
#define RSBTC_ATTR_DVBS_PHASE_NOISE_SHAPE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1317
#define RSBTC_ATTR_DVBS_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 295
#define RSBTC_ATTR_DVBS_SPECIAL_SCRAMBLER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 296
#define RSBTC_ATTR_DVBS_SPECIAL_SYNC_BYTE_INVERSION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 297
#define RSBTC_ATTR_DVBS_SPECIAL_REED_SOLOMON                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 298
#define RSBTC_ATTR_DVBS_SPECIAL_CONVOLUTIONAL_INTERLEAVER                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 299
#define RSBTC_ATTR_8VBS_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 300
#define RSBTC_ATTR_8VSB_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 301
#define RSBTC_ATTR_8VSB_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 302
#define RSBTC_ATTR_8VSB_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 303
#define RSBTC_ATTR_8VSB_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 304
#define RSBTC_ATTR_READ_8VSB_INPUT_SIGNAL_PACKET_LENGTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 305
#define RSBTC_ATTR_8VSB_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 306
#define RSBTC_ATTR_READ_8VSB_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 307
#define RSBTC_ATTR_READ_8VSB_INPUT_SIGNAL_MEASURED_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 308
#define RSBTC_ATTR_READ_8VSB_INPUT_SIGNAL_USEFUL_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 309
#define RSBTC_ATTR_8VSB_CONFIGURING_TEST_TS_PACKET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 310
#define RSBTC_ATTR_8VSB_CONFIGURING_PID_TEST_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 311
#define RSBTC_ATTR_8VSB_CONFIGURING_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 312
#define RSBTC_ATTR_8VSB_CONFIGURING_PAYLOAD_TEST_STUFF                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 313
#define RSBTC_ATTR_8VSB_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 314
#define RSBTC_ATTR_8VSB_CODING_SYMBOL_RATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 315
#define RSBTC_ATTR_8VSB_CODING_CONSTELLATION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 316
#define RSBTC_ATTR_8VSB_CODING_ROLL_OFF                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 317
#define RSBTC_ATTR_8VSB_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 318
#define RSBTC_ATTR_8VSB_SPECIAL_RANDOMIZER                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 319
#define RSBTC_ATTR_8VSB_SPECIAL_REED_SOLOMON                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 320
#define RSBTC_ATTR_8VSB_SPECIAL_INTERLEAVER                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 321
#define RSBTC_ATTR_8VSB_SPECIAL_PILOT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 322
#define RSBTC_ATTR_8VSB_SPECIAL_PILOT_VALUE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 323
#define RSBTC_ATTR_J83B_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 324
#define RSBTC_ATTR_J83B_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 325
#define RSBTC_ATTR_J83B_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 326
#define RSBTC_ATTR_J83B_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 327
#define RSBTC_ATTR_J83B_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 328
#define RSBTC_ATTR_READ_J83B_INPUT_SIGNAL_PACKET_LENGTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 329
#define RSBTC_ATTR_J83B_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 330
#define RSBTC_ATTR_READ_J83B_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 331
#define RSBTC_ATTR_READ_J83B_INPUT_SIGNAL_MEASURED_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 332
#define RSBTC_ATTR_READ_J83B_INPUT_SIGNAL_USEFUL_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 333
#define RSBTC_ATTR_J83B_CONFIGURING_TEST_TS_PACKET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 334
#define RSBTC_ATTR_J83B_CONFIGURING_PID_TEST_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 335
#define RSBTC_ATTR_J83B_CONFIGURING_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 336
#define RSBTC_ATTR_J83B_CONFIGURING_PAYLOAD_TEST_STUFF                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 337
#define RSBTC_ATTR_J83B_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 338
#define RSBTC_ATTR_J83B_CODING_SYMBOL_RATE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 339
#define RSBTC_ATTR_J83B_CODING_CONSTELLATION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 340
#define RSBTC_ATTR_J83B_CODING_ROLL_OFF                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 341
#define RSBTC_ATTR_J83B_CODING_INTERLEAVER_MODE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 342
#define RSBTC_ATTR_J83B_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 343
#define RSBTC_ATTR_J83B_SPECIAL_CHECKSUM_GENERATOR                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 344
#define RSBTC_ATTR_J83B_SPECIAL_REED_SOLOMON                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 345
#define RSBTC_ATTR_J83B_SPECIAL_INTERLEAVER                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 346
#define RSBTC_ATTR_J83B_SPECIAL_RANDOMIZER                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 347
#define RSBTC_ATTR_READ_ISDBT_EXTERNAL_INPUT_SIGNAL_MEASURED_DATA_RATE      RS_SPECIFIC_PUBLIC_ATTR_BASE + 348
#define RSBTC_ATTR_ISDBT_INPUT_SIGNAL_SOURCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 349
#define RSBTC_ATTR_ISDBT_INPUT_SIGNAL_TEST_SIGNAL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 350
#define RSBTC_ATTR_READ_ISDBT_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 351
#define RSBTC_ATTR_READ_ISDBT_INPUT_SIGNAL_MEAS_USEFUL_DATA_RATE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 352
#define RSBTC_ATTR_ISDBT_EXTERNAL_INPUT_SIGNAL_INPUT                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 353
#define RSBTC_ATTR_ISDBT_EXTERNAL_INPUT_SIGNAL_IP_TS_CHANNEL                RS_SPECIFIC_PUBLIC_ATTR_BASE + 354
#define RSBTC_ATTR_ISDBT_EXTERNAL_INPUT_SIGNAL_INPUT_FORMAT                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 355
#define RSBTC_ATTR_ISDBT_EXTERNAL_INPUT_SIGNAL_REMUX                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 356
#define RSBTC_ATTR_READ_ISDBT_EXTERNAL_INPUT_SIGNAL_PACKET_LENGTH           RS_SPECIFIC_PUBLIC_ATTR_BASE + 357
#define RSBTC_ATTR_ISDBT_EXTERNAL_INPUT_SIGNAL_STUFFING                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 358
#define RSBTC_ATTR_ISDBT_CONFIGURING_TEST_TS_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 359
#define RSBTC_ATTR_ISDBT_CONFIGURING_PID_TEST_PACKET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 360
#define RSBTC_ATTR_ISDBT_CONFIGURING_PID                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 361
#define RSBTC_ATTR_ISDBT_CONFIGURING_PAYLOAD_TEST_STUFF                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 362
#define RSBTC_ATTR_ISDBT_CONFIGURING_PRBS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 363
#define RSBTC_ATTR_ISDBT_SYSTEM_NETWORK_MODE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1099
#define RSBTC_ATTR_ISDBT_CODING_SYSTEM                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 364
#define RSBTC_ATTR_ISDBT_CODING_CONTROL                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 365
#define RSBTC_ATTR_ISDBT_CODING_IIP_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1100
#define RSBTC_ATTR_ISDBT_CODING_PORTION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 366
#define RSBTC_ATTR_ISDBT_CODING_CONSTELLATION                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 367
#define RSBTC_ATTR_ISDBT_CODING_SEGMENTS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 368
#define RSBTC_ATTR_ISDBT_CODING_CODE_RATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 369
#define RSBTC_ATTR_ISDBT_CODING_TIME_INTERLEAVING                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 370
#define RSBTC_ATTR_ISDBT_CODING_ISDBT_MODE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 371
#define RSBTC_ATTR_ISDBT_CODING_GUARD_INTERVAL                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 372
#define RSBTC_ATTR_ISDBT_CODING_SUBCHANNEL                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 373
#define RSBTC_ATTR_ISDBT_CODING_BANDWIDTH_VARIATION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 374
#define RSBTC_ATTR_ISDBT_CODING_CHANNEL_BANDWIDTH                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 375
#define RSBTC_ATTR_ISDBT_CODING_USED_BANDWIDTH                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 376
#define RSBTC_ATTR_ISDBT_SFN_NETWORK_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1101
#define RSBTC_ATTR_ISDBT_SFN_PROCESSING_DELAY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1102
#define RSBTC_ATTR_ISDBT_SFN_DYNAMIC_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1103
#define RSBTC_ATTR_ISDBT_SFN_TOTAL_DELAY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1104
#define RSBTC_ATTR_ISDBT_SFN_MAX_DEVIATION_TIME                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1105
#define RSBTC_ATTR_ISDBT_SFN_STATIC_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1106
#define RSBTC_ATTR_ISDBT_SFN_MAXIMUM_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1107
#define RSBTC_ATTR_ISDBT_SFN_DISPATCH_TIME                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1108
#define RSBTC_ATTR_ISDBT_SFN_MUTE_1ST_OFDM_SYMBOL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1109
#define RSBTC_ATTR_ISDBT_SFN_1PPS_ROUTING                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1235
#define RSBTC_ATTR_ISDBT_SFN_OFFSET                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1236
#define RSBTC_ATTR_ISDBT_EEW_STATE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1237
#define RSBTC_ATTR_ISDBT_EEW_SIGNAL_TYPE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1238
#define RSBTC_ATTR_ISDBT_EEW_AREA_INFORMATION                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1239
#define RSBTC_ATTR_ISDBT_EEW_APPLY                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1240
#define RSBTC_ATTR_ISDBT_EEW_NUMBER_OF_EPICENTERS                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1241
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_INFORMATION_TYPE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1242
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_WARNING_ID                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1243
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_LATITUDE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1244
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_LONGITUDE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1245
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_DEPTH                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1246
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_OCCURRENCE_TIME                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1247
#define RSBTC_ATTR_ISDBT_EEW_EPICENTER_APPLY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1248
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 377
#define RSBTC_ATTR_ISDBT_SPECIAL_SCRAMBLER                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 378
#define RSBTC_ATTR_ISDBT_SPECIAL_REED_SOLOMON                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 379
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_BYTE_INTERLEAVER                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 380
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_BIT_INTERLEAVER                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 381
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_FREQUENCY_INTERLEAVER             RS_SPECIFIC_PUBLIC_ATTR_BASE + 382
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_ALERT_BROADCASTING_FLAG           RS_SPECIFIC_PUBLIC_ATTR_BASE + 383
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_AC_DATA_AC1                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 384
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_AC_DATA_AC2                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 385
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_GUARD_INSERTION                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 386
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_TX_PARAM_SWITCH_INDICATOR         RS_SPECIFIC_PUBLIC_ATTR_BASE + 387
#define RSBTC_ATTR_ISDBT_SPECIAL_SETTINGS_MOD_TMCC_NEXT                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 388
#define RSBTC_ATTR_DTMB_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 389
#define RSBTC_ATTR_DTMB_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 390
#define RSBTC_ATTR_DTMB_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 391
#define RSBTC_ATTR_DTMB_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 392
#define RSBTC_ATTR_DTMB_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 393
#define RSBTC_ATTR_READ_DTMB_INPUT_SIGNAL_PACKET_LENGTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 394
#define RSBTC_ATTR_DTMB_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 395
#define RSBTC_ATTR_READ_DTMB_INPUT_SIGNAL_MAXIMUM_USEFUL_DATA_RATE          RS_SPECIFIC_PUBLIC_ATTR_BASE + 396
#define RSBTC_ATTR_READ_DTMB_INPUT_SIGNAL_MEASURED_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 397
#define RSBTC_ATTR_READ_DTMB_INPUT_SIGNAL_USEFUL_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 398
#define RSBTC_ATTR_DTMB_CONFIGURING_TEST_TS_PACKET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 399
#define RSBTC_ATTR_DTMB_CONFIGURING_PID_TEST_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 400
#define RSBTC_ATTR_DTMB_CONFIGURING_PID                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 401
#define RSBTC_ATTR_DTMB_CONFIGURING_PAYLOAD_TEST_STUFF                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 402
#define RSBTC_ATTR_DTMB_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 403
#define RSBTC_ATTR_DTMB_CODING_NETWORK_MODE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 404
#define RSBTC_ATTR_DTMB_CODING_CONTROL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1318
#define RSBTC_ATTR_DTMB_CODING_SINGLE_CARRIER_MODE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 405
#define RSBTC_ATTR_DTMB_CODING_DUAL_PILOT_TONE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 406
#define RSBTC_ATTR_DTMB_CODING_CONTROL_FRAMES                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 407
#define RSBTC_ATTR_DTMB_CODING_CONSTELLATION                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 408
#define RSBTC_ATTR_DTMB_CODING_CODE_RATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 409
#define RSBTC_ATTR_DTMB_CODING_GUARD_INTERVAL                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 410
#define RSBTC_ATTR_DTMB_CODING_GUARD_INTERVAL_PN                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 411
#define RSBTC_ATTR_DTMB_CODING_TIME_INTERLEAVER                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 412
#define RSBTC_ATTR_DTMB_CODING_CHANNEL_BANDWIDTH                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 413
#define RSBTC_ATTR_DTMB_SFN_NETWORK_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1319
#define RSBTC_ATTR_DTMB_SFN_PROCESSING_DELAY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1320
#define RSBTC_ATTR_DTMB_SFN_DYNAMIC_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1321
#define RSBTC_ATTR_DTMB_SFN_TOTAL_DELAY                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1322
#define RSBTC_ATTR_DTMB_SFN_MAX_DEVIATION_TIME                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1323
#define RSBTC_ATTR_DTMB_SFN_STATIC_DELAY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1324
#define RSBTC_ATTR_READ_DTMB_CODING_LOCAL_MAXIMUM_DELAY                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 414
#define RSBTC_ATTR_DTMB_SFN_DISPATCH_TIME                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1325
#define RSBTC_ATTR_DTMB_SFN_SIP_RESTAMPING                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1326
#define RSBTC_ATTR_DTMB_SFN_MUTE_1ST_SIGNAL_FRAME                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1327
#define RSBTC_ATTR_DTMB_SFN_1PPS_ROUTING                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1447
#define RSBTC_ATTR_DTMB_SFN_DISTRIBUTION_DELAY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 415
#define RSBTC_ATTR_DTMB_SFN_RF_DELAY                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 416
#define RSBTC_ATTR_DTMB_SFN_DELAY_OFFSET                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 417
#define RSBTC_ATTR_READ_DTMB_CODING_LOCAL_DELAY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 418
#define RSBTC_ATTR_DTMB_SFN_TX_ADDRESS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 419
#define RSBTC_ATTR_DTMB_SFN_DELAY_ADJUST                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 420
#define RSBTC_ATTR_DTMB_SFN_FREQUENCY_OFFSET                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 421
#define RSBTC_ATTR_DTMB_SFN_TX_POWER                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 422
#define RSBTC_ATTR_DTMB_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 423
#define RSBTC_ATTR_DTMB_SPECIAL_GI_POWER_BOOST                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 424
#define RSBTC_ATTR_DTMB_SPECIAL_SI_POWER_NORMALIZATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 425
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_CM_MODE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 426
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_NUM_INPUT_STREAMS                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1062
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_SOURCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1063
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_INPUT                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 427
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_IP_TS_CHANNEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 428
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_INPUT_FORMAT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 429
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_STUFFING                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 430
#define RSBTC_ATTR_READ_DVBS2_INPUT_SIGNAL_PACKET_LENGTH                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 431
#define RSBTC_ATTR_DVBS2_INPUT_SIGNAL_TEST_SIGNAL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 432
#define RSBTC_ATTR_READ_DVBS2_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 433
#define RSBTC_ATTR_READ_DVBS2_INPUT_SIGNAL_MEASURED_DATA_RATE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 434
#define RSBTC_ATTR_READ_DVBS2_INPUT_SIGNAL_USEFUL_DATA_RATE                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 435
#define RSBTC_ATTR_DVBS2_CONFIGURING_TEST_TS_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 436
#define RSBTC_ATTR_DVBS2_CONFIGURING_PID_TEST_PACKET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 437
#define RSBTC_ATTR_DVBS2_CONFIGURING_PID                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 438
#define RSBTC_ATTR_DVBS2_CONFIGURING_PAYLOAD_TEST_STUFF                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 439
#define RSBTC_ATTR_DVBS2_CONFIGURING_PRBS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 440
#define RSBTC_ATTR_DVBS2_CODING_SYMBOL_RATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 441
#define RSBTC_ATTR_DVBS2_CODING_ROLL_OFF                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 444
#define RSBTC_ATTR_DVBS2_CODING_ANNEX_M                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1110
#define RSBTC_ATTR_DVBS2_CODING_NUM_TIME_SLICE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1111
#define RSBTC_ATTR_DVBS2_CODING_S2_X                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1249
#define RSBTC_ATTR_DVBS2_CODING_S2_X_SUPER_FRAME                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1250
#define RSBTC_ATTR_DVBS2_CODING_S2_X_CHANNEL_BONDING                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1251
#define RSBTC_ATTR_DVBS2_CODING_TSN                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1112
#define RSBTC_ATTR_DVBS2_CODING_ISI                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1113
#define RSBTC_ATTR_DVBS2_CODING_MODULATION_CODING                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 445
#define RSBTC_ATTR_DVBS2_CODING_FECFRAME                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 442
#define RSBTC_ATTR_DVBS2_CODING_PILOTS                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 443
#define RSBTC_ATTR_DVBS2_CODING_NUM_SYMBOLS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1114
#define RSBTC_ATTR_DVBS2_CODING_CONSTELLATION                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 446
#define RSBTC_ATTR_DVBS2_CODING_CODE_RATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 447
#define RSBTC_ATTR_DVBS2_CID_ENABLED                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1328
#define RSBTC_ATTR_DVBS2_CID_HOST_SIGNAL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1329
#define RSBTC_ATTR_DVBS2_CID_FREQUENCY_OFFSET                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1330
#define RSBTC_ATTR_DVBS2_CID_CHIP_RATE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1331
#define RSBTC_ATTR_DVBS2_CID_DELTA_PSD                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1332
#define RSBTC_ATTR_DVBS2_CID_GLOBAL_UNIQUE_ID                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1333
#define RSBTC_ATTR_DVBS2_CID_GLOBAL_UNIQUE_ID_CRC                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1334
#define RSBTC_ATTR_DVBS2_CID_CONTENT_LATITUDE_LONGITUDE_ENABLED             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1335
#define RSBTC_ATTR_DVBS2_CID_CONTENT_LATITUDE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1336
#define RSBTC_ATTR_DVBS2_CID_CONTENT_LONGITUDE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1337
#define RSBTC_ATTR_DVBS2_CID_CONTENT_TELEPHONE_ENABLED                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1338
#define RSBTC_ATTR_DVBS2_CID_CONTENT_TELEPHONE_NUMBER                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1339
#define RSBTC_ATTR_DVBS2_CID_CONTENT_TELEPHONE_EXTENSION_ENABLED            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1340
#define RSBTC_ATTR_DVBS2_CID_CONTENT_TELEPHONE_EXTENSION_NUMBER             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1341
#define RSBTC_ATTR_DVBS2_PHASE_NOISE_ENABLED                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1342
#define RSBTC_ATTR_DVBS2_PHASE_NOISE_CW_MODE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1343
#define RSBTC_ATTR_DVBS2_PHASE_NOISE_MAGNITUDE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1344
#define RSBTC_ATTR_DVBS2_PHASE_NOISE_SHAPE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1345
#define RSBTC_ATTR_DVBS2_SPECIAL_SETTINGS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 448
#define RSBTC_ATTR_DVBS2_SPECIAL_PL_SCRAMBLING_SEQUENCE_ID                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 449
#define RSBTC_ATTR_DVBS2_SPECIAL_PL_GOLD_CODE_INDEX                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1346
#define RSBTC_ATTR_DVBS2_SPECIAL_PL_SCRAMBLER                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1448
#define RSBTC_ATTR_DVBS2_SPECIAL_PL_DATASLICES_PRBS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1449
#define RSBTC_ATTR_DVBS2_SPECIAL_BCH                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 450
#define RSBTC_ATTR_DVBS2_SPECIAL_LDPC_CC                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 451
#define RSBTC_ATTR_DVBS2_SPECIAL_CRC_8_IN_PAYLOAD                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 452
#define RSBTC_ATTR_DVBS2_SPECIAL_CRC_8_IN_HEADER                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 453
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_SOURCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 454
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_INPUT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 455
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_IP_TS_CHANNEL                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 456
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_INPUT_FORMAT                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 457
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_STUFFING                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 458
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_PACKET_LENGTH                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 459
#define RSBTC_ATTR_DIRECTV_INPUT_SIGNAL_TEST_SIGNAL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 460
#define RSBTC_ATTR_READ_DIRECTV_INPUT_SIGNAL_REQ_SOURCE_DATA_RATE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 461
#define RSBTC_ATTR_READ_DIRECTV_INPUT_SIGNAL_MAXIMUM_USEFUL_DATA_RATE       RS_SPECIFIC_PUBLIC_ATTR_BASE + 462
#define RSBTC_ATTR_READ_DIRECTV_INPUT_SIGNAL_MEASURED_DATA_RATE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 463
#define RSBTC_ATTR_READ_DIRECTV_INPUT_SIGNAL_USEFUL_DATA_RATE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 464
#define RSBTC_ATTR_DIRECTV_CONFIGURING_TEST_TS_PACKET                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 465
#define RSBTC_ATTR_DIRECTV_CONFIGURING_PRBS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 466
#define RSBTC_ATTR_DIRECTV_CODING_SYMBOL_RATE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 467
#define RSBTC_ATTR_DIRECTV_CODING_CONSTELLATION                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 468
#define RSBTC_ATTR_DIRECTV_CODING_ROLL_OFF                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 469
#define RSBTC_ATTR_DIRECTV_CODING_CODE_RATE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 470
#define RSBTC_ATTR_DIRECTV_SPECIAL_SETTINGS_STATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 471
#define RSBTC_ATTR_DIRECTV_SPECIAL_INSERT_ERRORS_AFTER_CONV_ENC             RS_SPECIFIC_PUBLIC_ATTR_BASE + 472
#define RSBTC_ATTR_DIRECTV_PHASE_NOISE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1115
#define RSBTC_ATTR_DIRECTV_PHASE_NOISE_MAGNITUDE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1116
#define RSBTC_ATTR_DIRECTV_PHASE_NOISE_CW_MODE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1347
#define RSBTC_ATTR_DIRECTV_PHASE_NOISE_SHAPE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1348
#define RSBTC_ATTR_ATSCMH_INPUT_SIGNAL_SOURCE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 473
#define RSBTC_ATTR_ATSCMH_INPUT_SIGNAL_INPUT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 474
#define RSBTC_ATTR_ATSCMH_INPUT_SIGNAL_IP_TS_CHANNEL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 475
#define RSBTC_ATTR_ATSCMH_INPUT_SIGNAL_INPUT_FORMAT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 476
#define RSBTC_ATTR_ATSCMH_INPUT_SIGNAL_STUFFING                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 477
#define RSBTC_ATTR_READ_ATSCMH_INPUT_SIGNAL_PACKET_LENGTH                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 478
#define RSBTC_ATTR_ATSCMH_INPUT_SIGNAL_TEST_SIGNAL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 479
#define RSBTC_ATTR_READ_ATSCMH_INPUT_SIGNAL_MAX_USEFUL_DATA_RATE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 480
#define RSBTC_ATTR_READ_ATSCMH_INPUT_SIGNAL_MEASURED_DATA_RATE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 481
#define RSBTC_ATTR_READ_ATSCMH_INPUT_SIGNAL_USEFUL_DATA_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 482
#define RSBTC_ATTR_ATSCMH_CONFIGURING_TEST_TS_PACKET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 483
#define RSBTC_ATTR_ATSCMH_CONFIGURING_PID_TEST_PACKET                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 484
#define RSBTC_ATTR_ATSCMH_CONFIGURING_PID                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 485
#define RSBTC_ATTR_ATSCMH_CONFIGURING_PAYLOAD_TEST_STUFF                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 486
#define RSBTC_ATTR_ATSCMH_CONFIGURING_PRBS                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 487
#define RSBTC_ATTR_ATSCMH_CODING_SYMBOL_RATE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 488
#define RSBTC_ATTR_ATSCMH_CODING_CONSTELLATION                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 489
#define RSBTC_ATTR_ATSCMH_CODING_CONTROL                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1118
#define RSBTC_ATTR_ATSCMH_CODING_ROLL_OFF                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 490
#define RSBTC_ATTR_ATSCMH_CODING_MH_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 491
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_MODE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1117
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_MHE_PID                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 492
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_RF_WATERMARK                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 493
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_TX_ADDRESS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 494
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_ID                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 495
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_BURY_RATIO                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 496
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_MTXID_TRANSMISSION                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 497
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_TRANSMITTER_ID                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 498
#define RSBTC_ATTR_ATSCMH_CODING_NETWORK_MARKET_ID                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 499
#define RSBTC_ATTR_ATSCMH_SFN_PROCESSING_DELAY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1120
#define RSBTC_ATTR_ATSCMH_SFN_DYNAMIC_DELAY                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1121
#define RSBTC_ATTR_ATSCMH_SFN_TOTAL_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1122
#define RSBTC_ATTR_ATSCMH_SFN_MAX_DEVIATION_TIME                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1123
#define RSBTC_ATTR_ATSCMH_SFN_STATIC_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1124
#define RSBTC_ATTR_ATSCMH_SFN_MAXIMUM_DELAY                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1125
#define RSBTC_ATTR_ATSCMH_SFN_DISPATCH_TIME                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1126
#define RSBTC_ATTR_ATSCMH_SFN_MUTE_1ST_OFDM_SYMBOL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1127
#define RSBTC_ATTR_ATSCMH_SPECIAL_SETTINGS                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 500
#define RSBTC_ATTR_ATSCMH_SPECIAL_RANDOMIZER                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 501
#define RSBTC_ATTR_ATSCMH_SPECIAL_REED_SOLOMON                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 502
#define RSBTC_ATTR_ATSCMH_SPECIAL_INTERLEAVER                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 503
#define RSBTC_ATTR_ATSCMH_SPECIAL_SFN_INITIALIZATION                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 504
#define RSBTC_ATTR_ATSCMH_SPECIAL_TRELLIS_INITIALIZATION                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 505
#define RSBTC_ATTR_ATSCMH_SPECIAL_PILOT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 506
#define RSBTC_ATTR_ATSCMH_SPECIAL_PILOT_VALUE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 507
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_STATUS_INFO                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1450
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_CONFORMANCE_CHECK                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1451
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_STL_INTERFACE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1452
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_STL_RESET_LOG_FILE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1453
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_NUMBER_OF_PLPS                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1454
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_SOURCE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1455
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_DESTINATION_IP_ADDRESS                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1457
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_DESTINATION_IP_PORT                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1458
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_INPUT_TYPE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1459
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_INPUT_PROTOCOL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1460
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_PLP_ID                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1461
#define RSBTC_ATTR_ATSC3_INPUT_SIGNAL_PLP_TEST_SIGNAL                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1462
#define RSBTC_ATTR_READ_ATSC3_INPUT_SIGNAL_PLP_PACKET_LENGTH                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1463
#define RSBTC_ATTR_READ_ATSC3_INPUT_SIGNAL_PLP_MAX_USEFUL_DATA_RATE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1464
#define RSBTC_ATTR_READ_ATSC3_INPUT_SIGNAL_PLP_MEASURED_DATA_RATE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1465
#define RSBTC_ATTR_READ_ATSC3_INPUT_SIGNAL_PLP_USEFUL_DATA_RATE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1466
#define RSBTC_ATTR_ATSC3_CONFIGURING_TEST_TS_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1467
#define RSBTC_ATTR_ATSC3_CONFIGURING_PID_TEST_PACKET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1468
#define RSBTC_ATTR_ATSC3_CONFIGURING_PID                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1469
#define RSBTC_ATTR_ATSC3_CONFIGURING_TEST_IP_PACKET                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1470
#define RSBTC_ATTR_ATSC3_CONFIGURING_PAYLOAD_TEST_STUFF                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1471
#define RSBTC_ATTR_ATSC3_CONFIGURING_PRBS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1472
#define RSBTC_ATTR_ATSC3_ALP_ENCAPSULATION_TYPE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1473
#define RSBTC_ATTR_ATSC3_ENCAPS_BB_FRAME_PADDING                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1474
#define RSBTC_ATTR_ATSC3_ENCAPS_BB_FRAME_COUNTER                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1475
#define RSBTC_ATTR_ATSC3_BICM_LOW_LEVEL_SIGNALING                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1476
#define RSBTC_ATTR_ATSC3_BICM_LAYER                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1477
#define RSBTC_ATTR_ATSC3_BICM_LDM_INJECTION_LEVEL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1478
#define RSBTC_ATTR_ATSC3_BICM_SIZE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1479
#define RSBTC_ATTR_ATSC3_BICM_SCRAMBLER_TYPE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1480
#define RSBTC_ATTR_ATSC3_BICM_FEC_TYPE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1481
#define RSBTC_ATTR_ATSC3_BICM_CODE_RATE                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1482
#define RSBTC_ATTR_ATSC3_BICM_CONSTELLATION                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1483
#define RSBTC_ATTR_ATSC3_BICM_SUBSLICING_TYPE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1484
#define RSBTC_ATTR_ATSC3_BICM_NUMBER_OF_SUBSLICES                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1485
#define RSBTC_ATTR_ATSC3_BICM_SUBSLICE_INTERVAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1486
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTERLEAVER_MODE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1487
#define RSBTC_ATTR_ATSC3_BICM_EXTENDED_INTERLEAVING                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1488
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTERLEAVER_DEPTH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1489
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTER_SUBFRAME                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1490
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTERLEAVER_NUMBER_OF_TI_BLOCKS          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1491
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTERLEAVER_NUMBER_OF_FEC_BLOCKS         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1581
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTERLEAVER_MAX_NUMBER_OF_FEC_BLOCKS     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1492
#define RSBTC_ATTR_ATSC3_BICM_TIME_INTERLEAVER_CELL_INTERLEAVER             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1493
#define RSBTC_ATTR_ATSC3_NUMBER_OF_SUBFRAMES                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1494
#define RSBTC_ATTR_ATSC3_OFDM_CHANNEL_BANDWIDTH                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1495
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_SUBFRAME_MIMO                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1496
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_SUBFRAME_MISO                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1497
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_FFT_SIZE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1498
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_REDUCED_CARRIER_MODE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1499
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_GUARD_INTERVAL                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1500
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_TOTAL_NUMBER_OF_DATA_OFDM_SYMBOLS  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1501
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_SISO_PILOT_PATTERN                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1502
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_PILOT_BOOST_MODE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1503
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_SUBFRAME_BOUNDARY_SYMBOL_FIRST     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1504
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_SUBFRAME_BOUNDARY_SYMBOL_LAST      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1505
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_SUBFRAME_BOUNDARY_SYMBOL_NULL_CELLS RS_SPECIFIC_PUBLIC_ATTR_BASE + 1506
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_FREQUENCY_INTERLEAVER              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1507
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_DURATION                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1508
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_USED_BANDWIDTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1509
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_NUMBER_OF_PLPS_IN_SUBFRAME         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1510
#define RSBTC_ATTR_ATSC3_SUBFRAMING_OFDM_PLP_IDS_IN_SUBFRAME                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1511
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_MODE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1512
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_LOW_LEVEL_SIGNALING                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1513
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_TIME_INFO                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1514
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_RETURN_CHANNEL                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1515
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_NUMBER_OF_TRANSMITTERS              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1516
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_TRANSMITTER_INDEX                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1517
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_TXID_MODE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1518
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_TXID_ADDRESS                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1519
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_TXID_INJECTION_LEVEL                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1520
#define RSBTC_ATTR_ATSC3_SYSTEM_NETWORK_BROADCAST_STREAM_ID                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1521
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_PEAK_TO_AVERAGE_POWER_RATIO           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1522
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_MODE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1523
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_LENGTH                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1524
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_EXCESS_SAMPLES_PER_SYMBOL             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1525
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_FINAL_EXCESS_SAMPLES                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1526
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_TIME_OFFSET                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1527
#define RSBTC_ATTR_ATSC3_SYSTEM_FRAME_ADDITIONAL_SAMPLES                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1528
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_BASIC_VERSION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1529
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_BASIC_FEC_TYPE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1530
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_NUMBER_OF_PREAMBLE_SYMBOLS               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1531
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_REDUCED_CARRIER_MODE_PREAMBLE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1532
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_PILOT_PATTERN_DX_PREAMBLE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1533
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_DETAIL_VERSION                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1534
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_DETAIL_FEC_TYPE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1535
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_DETAIL_ADDITIONAL_PARITY_MODE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1536
#define RSBTC_ATTR_ATSC3_SYSTEM_L1_NUMBER_OF_RFS                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1537
#define RSBTC_ATTR_ATSC3_FRAME_INFO_FRAME_DURATION                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1538
#define RSBTC_ATTR_ATSC3_FRAME_INFO_L1_BASIC_BYTES                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1539
#define RSBTC_ATTR_ATSC3_FRAME_INFO_L1_BASIC_CELLS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1540
#define RSBTC_ATTR_ATSC3_FRAME_INFO_L1_DETAIL_BYTES                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1541
#define RSBTC_ATTR_ATSC3_FRAME_INFO_L1_DETAIL_CELLS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1542
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_DURATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1543
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_MAJOR_VERSION                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1544
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_MINOR_VERSION                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1545
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_EMERGENCY_ALERT_SIGNALING     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1546
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_MIN_TIME_TO_NEXT              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1547
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_SYSTEM_BANDWIDTH              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1548
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_BSR_COEFFICIENT               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1549
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_PREAMBLE_STRUCTURE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1550
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_PREAMBLE_FFT_SIZE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1551
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_PREAMBLE_GUARD_INTERVAL       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1552
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_PREAMBLE_PILOT_PATTERN_DX     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1553
#define RSBTC_ATTR_ATSC3_FRAME_INFO_BOOTSTRAP_PREAMBLE_L1_BASIC_FEC_TYPE    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1554
#define RSBTC_ATTR_ATSC3_SFN_PROCESSING_DELAY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1555
#define RSBTC_ATTR_ATSC3_SFN_DYNAMIC_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1556
#define RSBTC_ATTR_ATSC3_SFN_TOTAL_DELAY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1557
#define RSBTC_ATTR_ATSC3_SFN_MAX_DEVIATION_TIME                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1558
#define RSBTC_ATTR_ATSC3_SFN_STATIC_DELAY                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1559
#define RSBTC_ATTR_ATSC3_SFN_MUTE                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1560
#define RSBTC_ATTR_ATSC3_SFN_NETWORK_DELAY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1561
#define RSBTC_ATTR_ATSC3_SFN_NETWORK_DELAY_JITTER                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1562
#define RSBTC_ATTR_ATSC3_SPECIAL_SETTINGS                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1563
#define RSBTC_ATTR_ATSC3_SPECIAL_PREAMBLE_COMPATIBILITY_MODE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1564
#define RSBTC_ATTR_ATSC3_SPECIAL_TMP_COMPATIBILITY_MODE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1565
#define RSBTC_ATTR_ATSC3_SPECIAL_LMT_COMPATIBILITY_MODE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1566
#define RSBTC_ATTR_ATSC3_SPECIAL_STL_MISO                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1567
#define RSBTC_ATTR_ATSC3_SPECIAL_BOOTSTRAP_MINOR_VERSION                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1568
#define RSBTC_ATTR_ATSC3_SPECIAL_BOOTSTRAP_EMERGENCY_ALERT_SIGNALING        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1569
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1010
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1128
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1129
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1130
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_STUFFING                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1011
#define RSBTC_ATTR_READ_CMMB_MAX_USEFUL_DATA_RATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1012
#define RSBTC_ATTR_READ_CMMB_MEASURED_DATA_RATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1013
#define RSBTC_ATTR_READ_CMMB_MEASURED_REQUIRED_DATA_RATE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1252
#define RSBTC_ATTR_READ_CMMB_STREAM_FORMAT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1253
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_NUMBER_OF_SERVICES                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1014
#define RSBTC_ATTR_CMMB_INPUT_SIGNAL_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1015
#define RSBTC_ATTR_CMMB_CONFIGURING_TRANSMITTER_ID                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1016
#define RSBTC_ATTR_CMMB_CONFIGURING_AREA_ID                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1017
#define RSBTC_ATTR_CMMB_CONFIGURING_PAYLOAD_TEST_SIGNAL                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1018
#define RSBTC_ATTR_CMMB_CONFIGURING_PRBS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1019
#define RSBTC_ATTR_CMMB_CODING_CONTROL                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1020
#define RSBTC_ATTR_CMMB_CODING_BANDWIDTH                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1021
#define RSBTC_ATTR_CMMB_CODING_OFDM_MODE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1022
#define RSBTC_ATTR_CMMB_SPECIAL_SETTINGS_STATE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1023
#define RSBTC_ATTR_CMMB_SPECIAL_REED_SOLOMON                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1024
#define RSBTC_ATTR_CMMB_SPECIAL_BYTE_INTERLEAVER                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1025
#define RSBTC_ATTR_CMMB_SPECIAL_BIT_INTERLEAVER                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1026
#define RSBTC_ATTR_CMMB_SPECIAL_SCRAMBLING                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1027
#define RSBTC_ATTR_ISDBTMM_INPUT_SIGNAL_SOURCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1028
#define RSBTC_ATTR_ISDBTMM_INPUT_SIGNAL_TEST_SIGNAL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1029
#define RSBTC_ATTR_ISDBTMM_INPUT_SIGNAL_INPUT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1030
#define RSBTC_ATTR_ISDBTMM_INPUT_SIGNAL_IP_TS_CHANNEL                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1131
#define RSBTC_ATTR_ISDBTMM_INPUT_SIGNAL_STUFFING                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1031
#define RSBTC_ATTR_READ_ISDBTMM_INPUT_DATA_RATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1032
#define RSBTC_ATTR_READ_ISDBTMM_MAX_USEFUL_DATA_RATE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1033
#define RSBTC_ATTR_ISDBTMM_CONFIGURING_TEST_TS_PACKET                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1034
#define RSBTC_ATTR_ISDBTMM_CONFIGURING_PAYLOAD_TEST_TS_PACKET               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1035
#define RSBTC_ATTR_ISDBTMM_CONFIGURING_PID_TEST_PACKET                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1036
#define RSBTC_ATTR_ISDBTMM_CONFIGURING_PID                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1037
#define RSBTC_ATTR_ISDBTMM_CONFIGURING_PRBS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1038
#define RSBTC_ATTR_ISDBTMM_CODING_CONTROL                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1039
#define RSBTC_ATTR_ISDBTMM_CODING_PARTIAL_RECEPTION                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1040
#define RSBTC_ATTR_ISDBTMM_CODING_NUMBER_OF_SUPER_SEGMENTS                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1041
#define RSBTC_ATTR_ISDBTMM_CODING_LAYER_OFDM_SEGMENTS                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1044
#define RSBTC_ATTR_ISDBTMM_CODING_LAYER_CONSTELLATION                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1046
#define RSBTC_ATTR_ISDBTMM_CODING_LAYER_CODE_RATE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1047
#define RSBTC_ATTR_ISDBTMM_CODING_LAYER_TIME_INTERLEAVING                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1048
#define RSBTC_ATTR_ISDBTMM_CODING_LAYER_CENTER_SUBCHANNEL                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1045
#define RSBTC_ATTR_ISDBTMM_CODING_SUPER_SEGMENTS_TYPE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1042
#define RSBTC_ATTR_ISDBTMM_CODING_NUMBER_SEGMENTS                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1043
#define RSBTC_ATTR_ISDBTMM_CODING_EMERGENCY_ALARM                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1049
#define RSBTC_ATTR_ISDBTMM_CODING_ISDBT_MODE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1050
#define RSBTC_ATTR_ISDBTMM_CODING_GUARD_INTERVAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1051
#define RSBTC_ATTR_ISDBTMM_CODING_REFERENCE_CHANNEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1052
#define RSBTC_ATTR_ISDBTMM_CODING_BANDWIDTH                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1053
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SETTINGS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1054
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SETTINGS_BYTE_INTERLEAVER                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1055
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SETTINGS_BIT_INTERLEAVER                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1056
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SETTINGS_FREQUENCY_INTERLEAVER           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1057
#define RSBTC_ATTR_ISDBTMM_SPECIAL_REED_SOLOMON                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1058
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SCRAMBLER                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1059
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SETTINGS_TX_PARAM_SWITCHING_INDICATOR    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1060
#define RSBTC_ATTR_ISDBTMM_SPECIAL_SETTINGS_TMCC_NEXT                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1061
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1065
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_INPUT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1066
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_IP_TS_CHANNEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1067
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_INPUT_FORMAT                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1068
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_ETI_SIGNAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1069
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_NUMBER_OF_STREAMS                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1070
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_MODE_IDENTITY                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1071
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_INFO_SUBCHANNEL_ID                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1072
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_INFO_PROTECTION_PROFILE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1073
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_INFO_PROTECTION_LEVEL                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1074
#define RSBTC_ATTR_TDMB_INPUT_SIGNAL_INFO_DATA_RATE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1075
#define RSBTC_ATTR_TDMB_NETWORK_MODE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1076
#define RSBTC_ATTR_TDMB_NETWORK_TTI                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1077
#define RSBTC_ATTR_TDMB_NETWORK_TII_MAIN_ID                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1078
#define RSBTC_ATTR_TDMB_NETWORK_TII_SUB_ID                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1079
#define RSBTC_ATTR_TDMB_SFN_NETWORK_COMPENSATION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1254
#define RSBTC_ATTR_TDMB_SFN_TX_COMPENSATION_DELAY                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1255
#define RSBTC_ATTR_TDMB_SFN_TX_PROCESSING_DELAY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1256
#define RSBTC_ATTR_TDMB_SFN_TX_DELAY                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1257
#define RSBTC_ATTR_TDMB_SFN_STATIC_DELAY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1258
#define RSBTC_ATTR_TDMB_SFN_TOTAL_DELAY                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1259
#define RSBTC_ATTR_TDMB_SFN_MAX_DEVIATION_TIME                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1260
#define RSBTC_ATTR_TDMB_SFN_1PPS_ROUTING                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1261
#define RSBTC_ATTR_TDMB_SPECIAL_SETTINGS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1080
#define RSBTC_ATTR_TDMB_SPECIAL_TRANSMISSION_MODE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1081
#define RSBTC_ATTR_TDMB_SPECIAL_SELECT_TRANSMISSION_MODE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1082
#define RSBTC_ATTR_TDMB_SPECIAL_PRBS_TEST_SIGNAL                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1083
#define RSBTC_ATTR_TDMB_SPECIAL_PRBS_TEST_SIGNAL_SETTINGS                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1084
#define RSBTC_ATTR_TDMB_SPECIAL_TEST_SIGNAL_SCID                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1085
#define RSBTC_ATTR_ANALOG_TV_VIDEO_INPUT_SIGNAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 508
#define RSBTC_ATTR_ANALOG_TV_VIDEO_VISION_PICTURE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 509
#define RSBTC_ATTR_ANALOG_TV_VIDEO_LOAD_FILE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 510
#define RSBTC_ATTR_ANALOG_TV_AUDIO_INPUT_SIGNAL                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 511
#define RSBTC_ATTR_ANALOG_TV_AUDIO_AF_1                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 512
#define RSBTC_ATTR_ANALOG_TV_AUDIO_AF_2                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 513
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_FREQUENCY_SOUND_MONO           RS_SPECIFIC_PUBLIC_ATTR_BASE + 514
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_FREQUENCY_SOUND_LEFT           RS_SPECIFIC_PUBLIC_ATTR_BASE + 515
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_FREQUENCY_SOUND_CHANNEL_1      RS_SPECIFIC_PUBLIC_ATTR_BASE + 516
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_FREQUENCY_SOUND_RIGHT          RS_SPECIFIC_PUBLIC_ATTR_BASE + 517
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_FREQUENCY_SOUND_CHANNEL_2      RS_SPECIFIC_PUBLIC_ATTR_BASE + 518
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_LEVEL_SOUND_MONO               RS_SPECIFIC_PUBLIC_ATTR_BASE + 519
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_LEVEL_SOUND_LEFT               RS_SPECIFIC_PUBLIC_ATTR_BASE + 520
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_LEVEL_SOUND_CHANNEL_1          RS_SPECIFIC_PUBLIC_ATTR_BASE + 521
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_LEVEL_SOUND_RIGHT              RS_SPECIFIC_PUBLIC_ATTR_BASE + 522
#define RSBTC_ATTR_ANALOG_TV_AUDIO_GENERATOR_LEVEL_SOUND_CHANNEL_2          RS_SPECIFIC_PUBLIC_ATTR_BASE + 523
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_LOAD_FILE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 524
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_LEVEL_SOUND_MONO                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 525
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_LEVEL_SOUND_LEFT                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 526
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_LEVEL_SOUND_CHANNEL_1             RS_SPECIFIC_PUBLIC_ATTR_BASE + 527
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_LEVEL_SOUND_RIGHT                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 528
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_LEVEL_SOUND_CHANNEL_2             RS_SPECIFIC_PUBLIC_ATTR_BASE + 529
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_ATTENUATION_1                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 530
#define RSBTC_ATTR_ANALOG_TV_AUDIO_PLAYER_ATTENUATION_2                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 531
#define RSBTC_ATTR_ANALOG_TV_NICAM_SOURCE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 532
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_SOUND_MODE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 533
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_AF_1                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 534
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_AF_2                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 535
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_FREQ_SOUND_MONO                RS_SPECIFIC_PUBLIC_ATTR_BASE + 536
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_FREQ_SOUND_LEFT                RS_SPECIFIC_PUBLIC_ATTR_BASE + 537
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_FREQ_SOUND_CHANNEL_1           RS_SPECIFIC_PUBLIC_ATTR_BASE + 538
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_FREQ_SOUND_RIGHT               RS_SPECIFIC_PUBLIC_ATTR_BASE + 539
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_FREQ_SOUND_CHANNEL_2           RS_SPECIFIC_PUBLIC_ATTR_BASE + 540
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_LEVEL_SOUND_MONO               RS_SPECIFIC_PUBLIC_ATTR_BASE + 541
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_LEVEL_SOUND_LEFT               RS_SPECIFIC_PUBLIC_ATTR_BASE + 542
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_LEVEL_SOUND_CHANNEL_1          RS_SPECIFIC_PUBLIC_ATTR_BASE + 543
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_LEVEL_SOUND_RIGHT              RS_SPECIFIC_PUBLIC_ATTR_BASE + 544
#define RSBTC_ATTR_ANALOG_TV_NICAM_GENERATOR_LEVEL_SOUND_CHANNEL_2          RS_SPECIFIC_PUBLIC_ATTR_BASE + 545
#define RSBTC_ATTR_ANALOG_TV_SOUND_MODULATION_MODE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 546
#define RSBTC_ATTR_ANALOG_TV_SOUND_MOD_FREQ_DEVIATION_MONO                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 547
#define RSBTC_ATTR_ANALOG_TV_SOUND_MOD_FREQ_DEVIATION_SUBCARRIER_1          RS_SPECIFIC_PUBLIC_ATTR_BASE + 548
#define RSBTC_ATTR_ANALOG_TV_SOUND_MOD_FREQ_DEVIATION_SUBCARRIER_2          RS_SPECIFIC_PUBLIC_ATTR_BASE + 549
#define RSBTC_ATTR_ANALOG_TV_SOUND_MOD_FREQ_DEVIATION_PILOT                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 550
#define RSBTC_ATTR_ANALOG_TV_BTSC_LOAD_FILE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1262
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SETTINGS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 551
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_CARRIER                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 552
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_VIDEO                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 553
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_VSB_FILTER                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 554
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_VSB_CHARACTERISTIC                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 555
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_GD_PRECORRECTION                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 556
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_RESIDUAL_CARRIER                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 557
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_PREEMPHASIS                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 558
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_NICAM_PREEMPHASIS                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 559
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_NICAM_GAIN_ADJUST_1                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 560
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_NICAM_GAIN_ADJUST_2                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 561
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_NICAM_ROLL_OFF                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 562
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_SUBCARRIER_MONO       RS_SPECIFIC_PUBLIC_ATTR_BASE + 563
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_SUBCARRIER_1          RS_SPECIFIC_PUBLIC_ATTR_BASE + 564
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_SUBCARRIER_2          RS_SPECIFIC_PUBLIC_ATTR_BASE + 565
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_FREQUENCY_SUBCARRIER_MONO RS_SPECIFIC_PUBLIC_ATTR_BASE + 566
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_FREQUENCY_SUBCARRIER_1 RS_SPECIFIC_PUBLIC_ATTR_BASE + 567
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_FREQUENCY_SUBCARRIER_2 RS_SPECIFIC_PUBLIC_ATTR_BASE + 568
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_LEVEL_SUBCARRIER_MONO RS_SPECIFIC_PUBLIC_ATTR_BASE + 569
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_LEVEL_SUBCARRIER_1    RS_SPECIFIC_PUBLIC_ATTR_BASE + 570
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_MODULATION_LEVEL_SUBCARRIER_2    RS_SPECIFIC_PUBLIC_ATTR_BASE + 571
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_PILOT                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 572
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_NOMINAL_FREQUENCY_DEVIATION_PILOT RS_SPECIFIC_PUBLIC_ATTR_BASE + 573
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_NOM_FREQUENCY_DEVIATION_MONO     RS_SPECIFIC_PUBLIC_ATTR_BASE + 574
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_NOM_FREQUENCY_DEVIATION_SUBCARRIER_1 RS_SPECIFIC_PUBLIC_ATTR_BASE + 575
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_NOM_FREQUENCY_DEVIATION_SUBCARRIER_2 RS_SPECIFIC_PUBLIC_ATTR_BASE + 576
#define RSBTC_ATTR_ANALOG_TV_SPECIAL_SOUND_AM_MODULATION_DEPTH              RS_SPECIFIC_PUBLIC_ATTR_BASE + 577
#define RSBTC_ATTR_AUDIO_BC_FM_INPUT_SIGNAL_SOURCE_1                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 578
#define RSBTC_ATTR_AUDIO_BC_FM_INPUT_SIGNAL_SOURCE_2                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 579
#define RSBTC_ATTR_AUDIO_BC_FM_AUDIO_CHANNEL_1                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 580
#define RSBTC_ATTR_AUDIO_BC_FM_AUDIO_CHANNEL_2                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 581
#define RSBTC_ATTR_AUDIO_BC_FM_AF_MODE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 582
#define RSBTC_ATTR_AUDIO_BC_FM_GENERATOR_FREQUENCY_MONO                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 583
#define RSBTC_ATTR_AUDIO_BC_FM_GENERATOR_FREQUENCY_LEFT                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 584
#define RSBTC_ATTR_AUDIO_BC_FM_GENERATOR_FREQUENCY_RIGHT                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 585
#define RSBTC_ATTR_AUDIO_BC_FM_GENERATOR_LEVEL_MONO                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 586
#define RSBTC_ATTR_AUDIO_BC_FM_GENERATOR_LEVEL_LEFT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 587
#define RSBTC_ATTR_AUDIO_BC_FM_GENERATOR_LEVEL_RIGHT                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 588
#define RSBTC_ATTR_AUDIO_BC_FM_AUDIO_PLAYER_LOAD_FILE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 589
#define RSBTC_ATTR_AUDIO_BC_FM_AUDIO_PLAYER_ATTENUATION_1                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 590
#define RSBTC_ATTR_AUDIO_BC_FM_AUDIO_PLAYER_ATTENUATION_2                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 591
#define RSBTC_ATTR_AUDIO_BC_FM_MODULATION_MODE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 592
#define RSBTC_ATTR_AUDIO_BC_FM_MODULATION_PREEMPHASIS                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 593
#define RSBTC_ATTR_AUDIO_BC_FM_MODULATION_RDS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 594
#define RSBTC_ATTR_AUDIO_BC_FM_MODULATION_NOM_FREQ_DEV_AUDIO                RS_SPECIFIC_PUBLIC_ATTR_BASE + 595
#define RSBTC_ATTR_READ_AUDIO_BC_FM_MODULATION_FREQ_DEV_AUDIO               RS_SPECIFIC_PUBLIC_ATTR_BASE + 596
#define RSBTC_ATTR_AUDIO_BC_FM_MODULATION_FREQ_DEV_RDS                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 597
#define RSBTC_ATTR_AUDIO_BC_FM_MODULATION_FREQ_DEVIATION_PILOT              RS_SPECIFIC_PUBLIC_ATTR_BASE + 598
#define RSBTC_ATTR_AUDIO_BC_RDS_GROUP_SEQUENCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 599
#define RSBTC_ATTR_AUDIO_BC_RDS_PI                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 600
#define RSBTC_ATTR_AUDIO_BC_RDS_PS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 601
#define RSBTC_ATTR_AUDIO_BC_RDS_TP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 602
#define RSBTC_ATTR_AUDIO_BC_RDS_TA                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 603
#define RSBTC_ATTR_AUDIO_BC_RDS_PTY                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 604
#define RSBTC_ATTR_AUDIO_BC_RDS_PTYN                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 605
#define RSBTC_ATTR_AUDIO_BC_RDS_RT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 606
#define RSBTC_ATTR_AUDIO_BC_RDS_RT_READY                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 607
#define RSBTC_ATTR_AUDIO_BC_RDS_MS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 608
#define RSBTC_ATTR_AUDIO_BC_RDS_DI_DYNAMIC_PTY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 609
#define RSBTC_ATTR_AUDIO_BC_RDS_DI_COMPRESSED                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 610
#define RSBTC_ATTR_AUDIO_BC_RDS_DI_ART_HEAD                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 611
#define RSBTC_ATTR_AUDIO_BC_RDS_DI_STEREO                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 612
#define RSBTC_ATTR_AUDIO_BC_RDS_DI_CT                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 613
#define RSBTC_ATTR_AUDIO_BC_RDS_DI_CT_OFFSET                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 614
#define RSBTC_ATTR_AUDIO_BC_AF_METHOD                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 615
#define RSBTC_ATTR_AUDIO_BC_AF_A_NUMBER_OF_FREQUENCIES                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 616
#define RSBTC_ATTR_AUDIO_BC_AF_A_FREQUENCY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 617
#define RSBTC_ATTR_AUDIO_BC_AF_B_NUMBER_OF_LISTS                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 618
#define RSBTC_ATTR_AUDIO_BC_AF_B_NUMBER_OF_FREQUENCIES                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 619
#define RSBTC_ATTR_AUDIO_BC_AF_B_TUNING_FREQUENCY                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 620
#define RSBTC_ATTR_AUDIO_BC_AF_B_AF_FREQUENCY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 621
#define RSBTC_ATTR_AUDIO_BC_AF_B_AF_ORDER                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 622
#define RSBTC_ATTR_AUDIO_BC_EON_PI                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 623
#define RSBTC_ATTR_AUDIO_BC_EON_PS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 624
#define RSBTC_ATTR_AUDIO_BC_EON_TP                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 625
#define RSBTC_ATTR_AUDIO_BC_EON_TA                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 626
#define RSBTC_ATTR_AUDIO_BC_EON_LA                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 627
#define RSBTC_ATTR_AUDIO_BC_EON_EG                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 628
#define RSBTC_ATTR_AUDIO_BC_EON_ILS                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 629
#define RSBTC_ATTR_AUDIO_BC_EON_LSN                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 630
#define RSBTC_ATTR_AUDIO_BC_EON_PTY                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 631
#define RSBTC_ATTR_AUDIO_BC_EON_PIN                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 632
#define RSBTC_ATTR_AUDIO_BC_EON_AF_METHOD                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 633
#define RSBTC_ATTR_AUDIO_BC_EON_AF_A_NUMBER_OF_FREQUENCIES                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 634
#define RSBTC_ATTR_AUDIO_BC_EON_AF_A_FREQUENCY                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 635
#define RSBTC_ATTR_AUDIO_BC_EON_AF_MAPPED_FREQ_NUMBER_OF_FREQUENCIES        RS_SPECIFIC_PUBLIC_ATTR_BASE + 636
#define RSBTC_ATTR_AUDIO_BC_EON_AF_MAPPED_FREQ_TUNING_FREQUENCY             RS_SPECIFIC_PUBLIC_ATTR_BASE + 637
#define RSBTC_ATTR_AUDIO_BC_EON_AF_MAPPED_FREQ_AF_FREQUENCY                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 638
#define RSBTC_ATTR_AUDIO_BC_TMC                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 639
#define RSBTC_ATTR_AUDIO_BC_TMC_GROUP_3A_VAR_00                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 640
#define RSBTC_ATTR_AUDIO_BC_TMC_GROUP_3A_VAR_01                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 641
#define RSBTC_ATTR_AUDIO_BC_TMC_NUMBER_OF_8A_GROUPS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 642
#define RSBTC_ATTR_AUDIO_BC_TMC_8A_BLOCK                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 643
#define RSBTC_ATTR_AUDIO_BC_TMC_APPLY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 644
#define RSBTC_ATTR_AUDIO_BC_TMC_READY                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 645
#define RSBTC_ATTR_AUDIO_BC_OPEN_FORMAT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 646
#define RSBTC_ATTR_AUDIO_BC_OPEN_FORMAT_GROUP_A_READY                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 647
#define RSBTC_ATTR_AUDIO_BC_OPEN_FORMAT_GROUP_A                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 648
#define RSBTC_ATTR_AUDIO_BC_OPEN_FORMAT_GROUP_B_READY                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 649
#define RSBTC_ATTR_AUDIO_BC_OPEN_FORMAT_GROUP_B                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 650
#define RSBTC_ATTR_AUDIO_BC_OPEN_FORMAT_APPLY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 651
#define RSBTC_ATTR_AUDIO_BC_FM_SPECIAL_SETTINGS                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 652
#define RSBTC_ATTR_AUDIO_BC_FM_SPECIAL_PILOT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 653
#define RSBTC_ATTR_AUDIO_BC_FM_SPECIAL_PHASE_OFFSET_PILOT                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 654
#define RSBTC_ATTR_AUDIO_BC_FM_SPECIAL_PHASE_OFFSET_RDS                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 655
#define RSBTC_ATTR_AUDIO_BC_AM_INPUT_SIGNAL_SOURCE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 656
#define RSBTC_ATTR_AUDIO_BC_AM_AUDIO_CHANNEL                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 657
#define RSBTC_ATTR_AUDIO_BC_AM_AUDIO_GENERATOR_FREQUENCY                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 658
#define RSBTC_ATTR_AUDIO_BC_AM_AUDIO_GENERATOR_LEVEL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 659
#define RSBTC_ATTR_AUDIO_BC_AM_AUDIO_PLAYER_LOAD_FILE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 660
#define RSBTC_ATTR_AUDIO_BC_AM_AUDIO_PLAYER_ATTENUATION                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 661
#define RSBTC_ATTR_AUDIO_BC_AM_NOMINAL_MODULATION_DEPTH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 662
#define RSBTC_ATTR_READ_AUDIO_BC_AM_MODULATION_DEPTH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 663
#define RSBTC_ATTR_ARB_MEMORY_USAGE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 664
#define RSBTC_ATTR_ARB_WAVEFORM_DELETE                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 665
#define RSBTC_ATTR_ARB_DEFRAG                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 666
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_STATE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 667
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_USAGE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 668
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_WAVEFORM                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 669
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_SAMPLES                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 670
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_SAMPLE_RATE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 671
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_SEQUENCE_DURATION          RS_SPECIFIC_PUBLIC_ATTR_BASE + 672
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_LEVEL                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 673
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_ATTENUATION                RS_SPECIFIC_PUBLIC_ATTR_BASE + 674
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_FREQUENCY_OFFSET           RS_SPECIFIC_PUBLIC_ATTR_BASE + 675
#define RSBTC_ATTR_ARBITRARY_WAVEFORM_GENERATION_POINTS                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1349
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_IQ_SWAP                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1133
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_CONNECTED_INSTRUMENT           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1134
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_SAMPLE_RATE_SOURCE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1135
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_SAMPLE_RATE_VALUE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1136
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_LEVEL_IQ_AUTO_SETTINGS         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1137
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_LEVEL_MEASUREMENT_PERIOD       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1138
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_LEVEL_SET_AUTO_LEVEL           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1139
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_LEVEL_CREST_FACTOR             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1140
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_LEVEL_PEAK                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1141
#define RSBTC_ATTR_AUDIO_BC_DIGITAL_IQ_INPUT_LEVEL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1142
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_ENABLED                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1350
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_IQ_SWAP                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1351
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_SAMPLE_RATE_SOURCE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1352
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_SAMPLE_RATE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1353
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_CONNECTED_DEVICE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1354
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_SET_LEVEL_VIA                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1355
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_PEAK_LEVEL                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1356
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_LEVEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1357
#define RSBTC_ATTR_BASEBAND_DIGITAL_IQ_OUTPUT_CREST_FACTOR                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1358
#define RSBTC_ATTR_INTERFERER_SIGNAL_SOURCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 694
#define RSBTC_ATTR_INTERFERER_SIGNAL_ROUTING                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 695
#define RSBTC_ATTR_INTERFERER_SIGNAL_FREQUENCY_OFFSET                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 696
#define RSBTC_ATTR_INTERFERER_SIGNAL_REFERENCE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 697
#define RSBTC_ATTR_INTERFERER_SIGNAL_POLARITY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 698
#define RSBTC_ATTR_INTERFERER_SIGNAL_ATTENUATION                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 699
#define RSBTC_ATTR_INTERFERER_SIGNAL_LEVEL                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 700
#define RSBTC_ATTR_TX_DISTORTION_NONLINEARITY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1199
#define RSBTC_ATTR_RX_DISTORTION_NONLINEARITY                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1263
#define RSBTC_ATTR_TX_DISTORTION                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1264
#define RSBTC_ATTR_RX_DISTORTION                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1265
#define RSBTC_ATTR_TX_DISTORTION_LIMITER                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1200
#define RSBTC_ATTR_TX_DISTORTION_LIMITER_POSITION                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1201
#define RSBTC_ATTR_TX_DISTORTION_NONLINEARITY_GAIN                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1205
#define RSBTC_ATTR_RX_DISTORTION_NONLINEARITY_GAIN                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1266
#define RSBTC_ATTR_TX_DISTORTION_NONLINEARITY_PHASE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1206
#define RSBTC_ATTR_RX_DISTORTION_NONLINEARITY_PHASE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1267
#define RSBTC_ATTR_TX_DISTORTION_FREQUENCY_RESPONSE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1207
#define RSBTC_ATTR_FADING_SIMULATOR_SIGNAL_ROUTING                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 702
#define RSBTC_ATTR_FADING_SIMULATOR_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 701
#define RSBTC_ATTR_FADING_BIRTH_DEATH_PROPAGATION_STATE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 703
#define RSBTC_ATTR_FADING_MOVING_PROPAGATION_STATE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 704
#define RSBTC_ATTR_FADING_2_CHANNEL_INTERFERER_STATE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 705
#define RSBTC_ATTR_FADING_STANDARD_PARAMETER_SET                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 706
#define RSBTC_ATTR_FADING_CONFIGURATION                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 707
#define RSBTC_ATTR_FADING_LOAD_FILE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 709
#define RSBTC_ATTR_FADING_STORE_FILE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 710
#define RSBTC_ATTR_FADING_DELETE_FILE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 711
#define RSBTC_ATTR_FADING_STANDARD_DELAY_BASIC_DELAY                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 712
#define RSBTC_ATTR_FADING_STANDARD_DELAY_STATE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 713
#define RSBTC_ATTR_FADING_STANDARD_DELAY_PROFILE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 714
#define RSBTC_ATTR_FADING_STANDARD_DELAY_PATH_LOSS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 715
#define RSBTC_ATTR_FADING_STANDARD_DELAY_ADDITIONAL_DELAY                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 716
#define RSBTC_ATTR_FADING_STANDARD_DELAY_RESULTING_DELAY                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 717
#define RSBTC_ATTR_FADING_STANDARD_DELAY_POWER_RATIO                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 718
#define RSBTC_ATTR_FADING_STANDARD_DELAY_CONST_PHASE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 719
#define RSBTC_ATTR_FADING_STANDARD_DELAY_SPEED                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 720
#define RSBTC_ATTR_FADING_STANDARD_DELAY_FREQUENCY_RATIO                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 721
#define RSBTC_ATTR_FADING_STANDARD_DELAY_RESULTING_DOPPLER_SHIFT            RS_SPECIFIC_PUBLIC_ATTR_BASE + 722
#define RSBTC_ATTR_FADING_STANDARD_DELAY_FREQUENCY_SHIFT                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1268
#define RSBTC_ATTR_FADING_STANDARD_DELAY_FREQUENCY_SPREAD                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1269
#define RSBTC_ATTR_FADING_STANDARD_DELAY_CORRELATION_PATH                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 723
#define RSBTC_ATTR_FADING_STANDARD_DELAY_CORRELATION_PATH_COEFFICIENT       RS_SPECIFIC_PUBLIC_ATTR_BASE + 724
#define RSBTC_ATTR_FADING_STANDARD_DELAY_CORRELATION_PATH_PHASE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 725
#define RSBTC_ATTR_FADING_STANDARD_DELAY_LOGNORMAL_STATE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 726
#define RSBTC_ATTR_FADING_STANDARD_DELAY_LOGNORMAL_STATE_LOCAL_CONSTANT     RS_SPECIFIC_PUBLIC_ATTR_BASE + 727
#define RSBTC_ATTR_FADING_STANDARD_DELAY_LOGNORMAL_STATE_STANDARD_DEVIATION RS_SPECIFIC_PUBLIC_ATTR_BASE + 728
#define RSBTC_ATTR_FADING_PAR_DEL30_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 729
#define RSBTC_ATTR_FADING_PAR_DEL30_PROFILE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 730
#define RSBTC_ATTR_FADING_PAR_DEL30_PATH_LOSS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 731
#define RSBTC_ATTR_FADING_PAR_DEL30_BASIC_DELAY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 732
#define RSBTC_ATTR_FADING_PAR_DEL30_ADDITIONAL_DELAY                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 733
#define RSBTC_ATTR_FADING_PAR_DEL30_RESULTING_DELAY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 734
#define RSBTC_ATTR_FADING_PAR_DEL30_POWER_RATIO                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 735
#define RSBTC_ATTR_FADING_PAR_DEL30_CONST_PHASE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 736
#define RSBTC_ATTR_FADING_PAR_DEL30_SPEED                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 737
#define RSBTC_ATTR_FADING_PAR_DEL30_FREQ_RATIO                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 738
#define RSBTC_ATTR_FADING_PAR_DEL30_RESULT_DOPPLER_SHIFT                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 739
#define RSBTC_ATTR_FADING_PAR_DEL30_CORRELATION_PATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 740
#define RSBTC_ATTR_FADING_PAR_DEL30_COEFFICIENT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 741
#define RSBTC_ATTR_FADING_PAR_DEL30_PHASE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 742
#define RSBTC_ATTR_FADING_PAR_DEL30_LOGNORMAL_STATE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 743
#define RSBTC_ATTR_FADING_PAR_DEL30_LOCAL_CONSTANT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 744
#define RSBTC_ATTR_FADING_PAR_DEL30_STANDARD_DEVIATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 745
#define RSBTC_ATTR_FADING_PAR_DEL30_STATE_MODE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 746
#define RSBTC_ATTR_FADING_PAR_DEL50_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 747
#define RSBTC_ATTR_FADING_PAR_DEL50_PROFILE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 748
#define RSBTC_ATTR_FADING_PAR_DEL50_PATH_LOSS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 749
#define RSBTC_ATTR_FADING_PAR_DEL50_BASIC_DELAY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 750
#define RSBTC_ATTR_FADING_PAR_DEL50_ADDITIONAL_DELAY                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 751
#define RSBTC_ATTR_FADING_PAR_DEL50_RESULTING_DELAY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 752
#define RSBTC_ATTR_FADING_PAR_DEL50_POWER_RATIO                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 753
#define RSBTC_ATTR_FADING_PAR_DEL50_CONST_PHASE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 754
#define RSBTC_ATTR_FADING_PAR_DEL50_SPEED                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 755
#define RSBTC_ATTR_FADING_PAR_DEL50_FREQ_RATIO                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 756
#define RSBTC_ATTR_FADING_PAR_DEL50_RESULT_DOPPLER_SHIFT                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 757
#define RSBTC_ATTR_FADING_PAR_DEL50_CORRELATION_PATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 758
#define RSBTC_ATTR_FADING_PAR_DEL50_COEFFICIENT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 759
#define RSBTC_ATTR_FADING_PAR_DEL50_PHASE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 760
#define RSBTC_ATTR_FADING_PAR_DEL50_LOGNORMAL_STATE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 761
#define RSBTC_ATTR_FADING_PAR_DEL50_LOCAL_CONSTANT                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 762
#define RSBTC_ATTR_FADING_PAR_DEL50_STANDARD_DEVIATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 763
#define RSBTC_ATTR_FADING_PAR_DEL50_STATE_MODE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 764
#define RSBTC_ATTR_FADING_PAR_PROFILE_BIRTH_DEATH                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 765
#define RSBTC_ATTR_FADING_PAR_PATH_LOSS_BIRTH_DEATH                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 766
#define RSBTC_ATTR_FADING_PAR_MIN_DELAY_BIRTH_DEATH                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 767
#define RSBTC_ATTR_FADING_PAR_DELAY_GRID_BIRTH_DEATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 768
#define RSBTC_ATTR_FADING_PAR_POSITIONS_BIRTH_DEATH                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 769
#define RSBTC_ATTR_FADING_PAR_MAX_DELAY_BIRTH_DEATH                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 770
#define RSBTC_ATTR_FADING_PAR_START_OFFSET_BIRTH_DEATH                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 771
#define RSBTC_ATTR_FADING_PAR_HOPPING_DWELL_BIRTH_DEATH                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 772
#define RSBTC_ATTR_FADING_PAR_SPEED_BIRTH_DEATH                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 773
#define RSBTC_ATTR_FADING_PAR_FREQ_RATIO_BIRTH_DEATH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 774
#define RSBTC_ATTR_FADING_PAR_RES_DOPPLER_SHIFT_BIRTH_DEATH                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 775
#define RSBTC_ATTR_FADING_PAR_REF_PATH_STATE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 776
#define RSBTC_ATTR_FADING_PAR_REF_PATH_LOSS                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 777
#define RSBTC_ATTR_FADING_PAR_REF_PATH_DELAY                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 778
#define RSBTC_ATTR_FADING_PAR_MOVING_PATH_STATE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 779
#define RSBTC_ATTR_FADING_PAR_MOVING_PATH_LOSS                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 780
#define RSBTC_ATTR_FADING_PAR_MOVING_PATH_DELAY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 781
#define RSBTC_ATTR_FADING_PAR_MOVING_PATH_VARIATION_PK                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 782
#define RSBTC_ATTR_FADING_PAR_MOVING_PATH_VARIATION_PERIOD                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 783
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_REF_PATH_STATE                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 784
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_REF_PATH_PROFILE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 785
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_REF_PATH_LOSS                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 786
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_SPEED                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 787
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_REF_PATH_FREQ_RATIO                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 788
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_REF_PATH_DOPLER_SHIFT               RS_SPECIFIC_PUBLIC_ATTR_BASE + 789
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_REF_PATH_MIN_DELAY                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 790
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_STATE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 791
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_PROFILE                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 792
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_LOSS                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 793
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_FREQ_RATIO              RS_SPECIFIC_PUBLIC_ATTR_BASE + 794
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_DOPLER_SHIFT            RS_SPECIFIC_PUBLIC_ATTR_BASE + 795
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_MIN_DELAY               RS_SPECIFIC_PUBLIC_ATTR_BASE + 796
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PATH_MAX_DELAY               RS_SPECIFIC_PUBLIC_ATTR_BASE + 797
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_MODE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 798
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_PERIOD                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 799
#define RSBTC_ATTR_FADING_2PATH_DYN_DEL_MOVING_DWELL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 800
#define RSBTC_ATTR_FADING_MIMO_MATRIX_VALUE_COLUMN_MAGNITUDE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1270
#define RSBTC_ATTR_FADING_MIMO_MATRIX_VALUE_COLUMN_PHASE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1271
#define RSBTC_ATTR_FADING_MIMO_MATRIX_VALUE_CONFLICT_COLUMN                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1272
#define RSBTC_ATTR_FADING_MIMO_MATRIX_ACCEPT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1273
#define RSBTC_ATTR_FADING_MIMO_RELATIVE_TAP_VECTOR_GAIN                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1274
#define RSBTC_ATTR_FADING_MIMO_RELATIVE_TAP_VECTOR_PHASE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1275
#define RSBTC_ATTR_FADING_MIMO_RELATIVE_TAP_VECTOR_SET_TO_UNITY             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1359
#define RSBTC_ATTR_FADING_GENERAL_SPEED_UNIT                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 801
#define RSBTC_ATTR_FADING_GENERAL_REFERENCE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 802
#define RSBTC_ATTR_FADING_GENERAL_COMMON_SPEED_FOR_ALL_PATHS                RS_SPECIFIC_PUBLIC_ATTR_BASE + 803
#define RSBTC_ATTR_FADING_GENERAL_SIGNAL_DEDICATED_TO                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 805
#define RSBTC_ATTR_FADING_GENERAL_VIRTUAL_RF                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 806
#define RSBTC_ATTR_FADING_GENERAL_IGNORE_RF_CHANGES                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 807
#define RSBTC_ATTR_FADING_GENERAL_FREQUENCY_HOPPING_MODE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 808
#define RSBTC_ATTR_FADING_GENERAL_RESTART_EVENT                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 809
#define RSBTC_ATTR_FADING_GENERAL_RESTART                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 810
#define RSBTC_ATTR_FADING_INSERTION_LOSS_MODE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 811
#define RSBTC_ATTR_FADING_INSERTION_LOSS                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 812
#define RSBTC_ATTR_FADING_INSERTION_LOSS_CLIPPED_SAMPLES                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 813
#define RSBTC_ATTR_FADING_GLOBAL_FADER_COUPLING_START_SEED                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 814
#define RSBTC_ATTR_FADING_STANDARD_DEVIATION_COUPLED                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 815
#define RSBTC_ATTR_FADING_LOCAL_CONSTANT_COUPLED                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 816
#define RSBTC_ATTR_FADING_DEFAULT                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 817
#define RSBTC_ATTR_FADING_SPEED_COUPLED                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 820
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_TARGET_NAME                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1570
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_REMOTE_CONNECTION_STATE    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1571
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_REMOTE_CONNECT             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1572
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_REMOTE_DISCONNECT          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1573
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_DIGITAL_IQ_CONNECTION_STATE RS_SPECIFIC_PUBLIC_ATTR_BASE + 1574
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_DIGITAL_IQ_CONNECTED_INSTRUMENT RS_SPECIFIC_PUBLIC_ATTR_BASE + 1575
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_RF_FREQUENCY               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1576
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_RF_LEVEL                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1577
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_RF_STATE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1578
#define RSBTC_ATTR_FADING_EXTERNAL_RF_SIGNAL_GEN_RF_COUPLING                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1579
#define RSBTC_ATTR_NOISE_A_NOISE_B                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 822
#define RSBTC_ATTR_NOISE_SAVE_USER_PROFILE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 823
#define RSBTC_ATTR_NOISE_SELECT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 824
#define RSBTC_ATTR_NOISE_AWGN                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 825
#define RSBTC_ATTR_NOISE_AWGN_CN                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 826
#define RSBTC_ATTR_NOISE_AWGN_EB_N0_RATIO                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 827
#define RSBTC_ATTR_NOISE_AWGN_SETTINGS_BANDWIDTH_COUPLING                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 828
#define RSBTC_ATTR_NOISE_AWGN_SETTINGS_RECEIVER_BANDWIDTH                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 829
#define RSBTC_ATTR_NOISE_AWGN_SETTINGS_GENERATED_NOISE_BANDWIDTH            RS_SPECIFIC_PUBLIC_ATTR_BASE + 830
#define RSBTC_ATTR_NOISE_IMPULSIVE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 831
#define RSBTC_ATTR_NOISE_IMPULSIVE_CI                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 832
#define RSBTC_ATTR_NOISE_IMPULSIVE_FRAME_DURATION                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 833
#define RSBTC_ATTR_NOISE_IMPULSIVE_PULSES_PER_BURST                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 834
#define RSBTC_ATTR_NOISE_IMPULSIVE_PULSE_SPACING_MIN                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 835
#define RSBTC_ATTR_NOISE_IMPULSIVE_PULSE_SPACING_MAX                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 836
#define RSBTC_ATTR_PHASE_NOISE                                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 837
#define RSBTC_ATTR_PHASE_NOISE_SSB_SHAPE_PROFILE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 838
#define RSBTC_ATTR_PHASE_NOISE_AT_100_HZ                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 839
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_STATE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 840
#define RSBTC_ATTR_RF_SIGNAL_OUTPUT_STATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 841
#define RSBTC_ATTR_RF_SIGNAL_OUTPUT_PROTECTION_TRIPPED                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 842
#define RSBTC_ATTR_RF_SIGNAL_OUTPUT_PROTECTION_CLEAR                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 843
#define RSBTC_ATTR_RF_SIGNAL_OUTPUT_PROTECTION_CIRCUIT_STATUS               RS_SPECIFIC_PUBLIC_ATTR_BASE + 844
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_UNIT                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 845
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 846
#define RSBTC_ATTR_RF_SIGNAL_ACTUAL_CENTER_FREQUENCY                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 847
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_OFFSET                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 848
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_COUPLING_AB                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 849
#define RSBTC_ATTR_RF_SIGNAL_VSB_FREQUENCY                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 850
#define RSBTC_ATTR_RF_SIGNAL_DELTA_PHASE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 853
#define RSBTC_ATTR_RF_SIGNAL_LOCAL_OSCILLATOR_MODE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 854
#define RSBTC_ATTR_RF_SIGNAL_LOCAL_OSCILLATOR_OUT_STATE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 855
#define RSBTC_ATTR_RF_SIGNAL_CHANNEL_TABLE_CHANNEL_FILE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 856
#define RSBTC_ATTR_RF_SIGNAL_CHANNEL_NUMBER                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 857
#define RSBTC_ATTR_RF_SIGNAL_CHANNEL_NAME                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 858
#define RSBTC_ATTR_RF_SIGNAL_CHANNEL_TABLE_STORE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 859
#define RSBTC_ATTR_RF_SIGNAL_CHANNEL_TABLE_DELETE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 860
#define RSBTC_ATTR_RF_SIGNAL_EXTERNAL_REF_FREQ_SOURCE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 861
#define RSBTC_ATTR_RF_SIGNAL_EXTERNAL_REF_FREQ_ADJUSTMENT_STATE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 862
#define RSBTC_ATTR_RF_SIGNAL_EXTERNAL_REF_FREQ_ADJUSTMENT_DAC_VALUE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 863
#define RSBTC_ATTR_RF_SIGNAL_EXTERNAL_REF_FREQUENCY                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 864
#define RSBTC_ATTR_RF_SIGNAL_EXTERNAL_REF_FREQ_SYNC_BANDWIDTH               RS_SPECIFIC_PUBLIC_ATTR_BASE + 865
#define RSBTC_ATTR_RF_SIGNAL_LEVEL                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 868
#define RSBTC_ATTR_RF_SIGNAL_VOLTAGE_LEVEL                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1294
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_DIGITAL_ATTENUATION                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1276
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SETTINGS_MODE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 866
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SETTING_CHARACTERISTIC                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 867
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_PEP                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 869
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_USER_LIMIT_PEP                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 870
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_OFFSET                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 871
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_COUPLING_AB                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 872
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ATTENUATOR_MODE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 875
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ATTENUATION                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 876
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ATTENUATION_RANGE_MAXIMUM                RS_SPECIFIC_PUBLIC_ATTR_BASE + 877
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ATTENUATION_RANGE_MINIMUM                RS_SPECIFIC_PUBLIC_ATTR_BASE + 878
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ATTENUATION_RANGE_RESET                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 879
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ALC_STATE                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 880
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ALC_DETECTOR_SENSITIVITY                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 883
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ALC_SAMPLE_LEVEL                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 884
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_ALC_SEARCH_ONCE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 882
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_STATE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 885
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_MODE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 886
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_START                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 887
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_STOP                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 888
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_CENTER                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 889
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_SPAN                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 890
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_SPACING                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 891
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_SHAPE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 892
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_STEP_LINEAR                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 893
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_STEP_LOGARITHMIC               RS_SPECIFIC_PUBLIC_ATTR_BASE + 894
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_DWELL                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 895
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_RESET                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 896
#define RSBTC_ATTR_RF_SIGNAL_FREQUENCY_SWEEP_EXECUTE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 897
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_STATE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 899
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_MODE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 900
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_START                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 901
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_STOP                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 902
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_SHAPE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 903
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_STEP                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 904
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_DWELL                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 905
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_RESET                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 906
#define RSBTC_ATTR_RF_SIGNAL_LEVEL_SWEEP_EXECUTE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 907
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_SIGNAL_SOURCE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 908
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_IQ_WIDEBAND                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 909
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_OPTIMIZATION_MODE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 910
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_INTERNAL_BASEBAND_GAIN              RS_SPECIFIC_PUBLIC_ATTR_BASE + 911
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_WIDEBAND_IQ_GAIN                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 912
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_WIDEBAND_IQ_CREST_FACTOR            RS_SPECIFIC_PUBLIC_ATTR_BASE + 913
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_BASEBAND_IMPAIRMENTS_STATE          RS_SPECIFIC_PUBLIC_ATTR_BASE + 914
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_BASEBAND_IMPAIRMENTS_I_OFFSET       RS_SPECIFIC_PUBLIC_ATTR_BASE + 915
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_BASEBAND_IMPAIRMENTS_Q_OFFSET       RS_SPECIFIC_PUBLIC_ATTR_BASE + 916
#define RSBTC_ATTR_RF_SIGNAL_MOD_BB_IMPAIRMENTS_QUADRATURE_ERR              RS_SPECIFIC_PUBLIC_ATTR_BASE + 917
#define RSBTC_ATTR_RF_SIGNAL_MOD_BB_IMPAIR_AMPLITUDE_IMBALANCE              RS_SPECIFIC_PUBLIC_ATTR_BASE + 918
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_BASEBAND_IMPAIRMENTS_SKEW           RS_SPECIFIC_PUBLIC_ATTR_BASE + 919
#define RSBTC_ATTR_RF_SIGNAL_MODULATION_BASEBAND_IMPAIRMENTS_IQ_DELAY       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1143
#define RSBTC_ATTR_RF_SIGNAL_MODULATOR_IMPAIRMENTS_STATE                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 920
#define RSBTC_ATTR_RF_SIGNAL_MODULATOR_IMPAIRMENTS_I_OFFSET                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 921
#define RSBTC_ATTR_RF_SIGNAL_MODULATOR_IMPAIRMENTS_Q_OFFSET                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 922
#define RSBTC_ATTR_RF_SIGNAL_MODULATOR_IMPAIRMENTS_QUADRATURE_ERROR         RS_SPECIFIC_PUBLIC_ATTR_BASE + 923
#define RSBTC_ATTR_RF_SIGNAL_MODULATOR_IMPAIRMENTS_AMPLITUDE_IMBALANCE      RS_SPECIFIC_PUBLIC_ATTR_BASE + 924
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_STATE                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1144
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_SAMPLE_RATE_SOURCE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1145
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_SAMPLE_RATE_VALUE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1146
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_CONNECTED_DEVICE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1147
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_RESET                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1148
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_SET_LEVEL_VIA                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1149
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_PEAK_LEVEL                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1150
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_LEVEL                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1151
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_USER_VARIATION_ACTIVE        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1152
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_OUTPUT_USER_VARIATION_STEP          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1153
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_STATE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1154
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_I_OFFSET                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1155
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_Q_OFFSET                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1156
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_QUADRATURE_ERROR        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1157
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_AMPLITUDE_IMBALANCE     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1158
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_SKEW                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1159
#define RSBTC_ATTR_RF_SIGNAL_IQ_DIGITAL_IMPAIRMENTS_IQ_DELAY                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1160
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_STATE                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1161
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_TYPE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1162
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_MODE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1163
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_IQ_LEVEL_VP                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1164
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_COUPLE_IQ_BIAS                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1165
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_I_SETTINGS_BIAS               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1166
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_Q_SETTINGS_BIAS               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1167
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_I_SETTINGS_OFFSET             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1168
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_OUTPUT_Q_SETTINGS_OFFSET             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1169
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_STATE            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1170
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_I_OFFSET         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1171
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_Q_OFFSET         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1172
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_QUADRATURE_ERROR RS_SPECIFIC_PUBLIC_ATTR_BASE + 1173
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_AMPLITUDE_IMBALANCE RS_SPECIFIC_PUBLIC_ATTR_BASE + 1174
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_SKEW             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1175
#define RSBTC_ATTR_RF_SIGNAL_IQ_ANALOG_DIGITAL_IMPAIRMENTS_IQ_DELAY         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1176
#define RSBTC_ATTR_MM_GENERATOR_TRANSPORT_STREAM_REMUX                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1295
#define RSBTC_ATTR_MM_GENERATOR_TRANSPORT_STREAM_PLAYER_TYPE                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1296
#define RSBTC_ATTR_MM_GENERATOR_TRANSPORT_STREAM_PLAYER_INPUT_STATE         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1297
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_STATUS                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 925
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_FILE                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 926
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_FILE_DATE_AND_TIME                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 927
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_FILE_SIZE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 928
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_ORIGINAL_LOOP_TIME                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 929
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_ORIGINAL_DATA_RATE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 930
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_BYTE_RATE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1177
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_DATA_RATE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 931
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_PACKET_LENGTH                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 932
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_NULLPACKET_STUFFING           RS_SPECIFIC_PUBLIC_ATTR_BASE + 933
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_STOP_DATA                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 934
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_SETTINGS_TEST_TS_PACKET       RS_SPECIFIC_PUBLIC_ATTR_BASE + 935
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_SETTINGS_PID_TEST_PACKET      RS_SPECIFIC_PUBLIC_ATTR_BASE + 936
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_SETTINGS_PID                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 937
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_SETTINGS_PAYLOAD              RS_SPECIFIC_PUBLIC_ATTR_BASE + 938
#define RSBTC_ATTR_MM_GENERATOR_PLAYER_OUTPUT_SETTINGS_PRBS                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 939
#define RSBTC_ATTR_MM_GENERATOR_PLAY_WINDOW_START                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 940
#define RSBTC_ATTR_MM_GENERATOR_PLAY_WINDOW_STOP                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 941
#define RSBTC_ATTR_MM_GENERATOR_PLAY_WINDOW_RESET                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 942
#define RSBTC_ATTR_MM_GENERATOR_PLAY_WINDOW_POSITION                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1360
#define RSBTC_ATTR_MM_GENERATOR_SEAMLESS_LOOP_CONTINUITY_COUNTER            RS_SPECIFIC_PUBLIC_ATTR_BASE + 943
#define RSBTC_ATTR_MM_GENERATOR_SEAMLESS_LOOP_PCR_DTS_PTS                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 944
#define RSBTC_ATTR_MM_GENERATOR_SEAMLESS_LOOP_TDT_TOT                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 945
#define RSBTC_ATTR_MM_GENERATOR_OUTPUT_CONFIG_DATA_RATE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1298
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1277
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT_ALIAS                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1278
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT_TYPE                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1279
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT_MULTICAST_ADDRESS                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1280
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT_PORT                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1281
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT_IGMPV3_SOURCE_ADDRESS              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1282
#define RSBTC_ATTR_MM_GENERATOR_IP_INPUT_PING_SOURCE_ADDRESS                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1283
#define RSBTC_ATTR_MMG_GATEWAY_MODE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1361
#define RSBTC_ATTR_MMG_GATEWAY_STATUS_INFO                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1362
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_USE_CASE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1363
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PROFILES                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1364
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_GROUPS                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1365
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1366
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_VV_STREAM_GENERATION                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1367
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_GROUP_ID                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1368
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MAP_TO_PROFILE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1369
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_INPUT_TYPE                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1370
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_INPUT_SELECT                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1371
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_GENERATOR_DATA_RATE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1372
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_COMMON_SLOT_INTERVAL                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1373
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_NUM_CHAPTERS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1374
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_NUM_EIT_PACKETS                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1375
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_TS_OVER_PLP_ID                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1376
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_TS_OVER_CPLP                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1377
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PRBS_ID                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1378
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_NUMBER                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1379
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_DATA_RATE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1380
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_ID                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1382
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MAP_TO_GROUP                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1383
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_TYPE                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1384
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_DATA                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1385
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_PLP_PAYLOAD_TYPE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1386
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_NEW_TS_ID                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1388
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_BB_MODE                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1390
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_AUTO_ISSY                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1391
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_ISSY                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1392
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_MAX_BUFFER_SIZE          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1393
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_BUFFER_SIZE_UNIT         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1394
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_DESIGN_DELAY             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1395
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_NULL_PACKET_DELETION     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1396
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_INBAND_SIGNALLING        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1397
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_INBAND_SIGNALLING_A      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1398
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_MODE_STREAM_INBAND_SIGNALLING_B      RS_SPECIFIC_PUBLIC_ATTR_BASE + 1399
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_FEC_FRAME                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1400
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_CODE_RATE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1401
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_CONSTELLATION                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1402
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_CONSTELLATION_ROTATION          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1403
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_TIME_INTERLEAVER_TYPE           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1404
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_FRAME_INTERVAL                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1405
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_TIME_INTERLEAVER_LENGTH         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1406
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_FIRST_FRAME_INDEX               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1407
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_AUTO_NUM_BLOCKS_MAX             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1408
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_NUM_BLOCKS_MAX                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1409
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_BICM_MAX_NUM_BLOCKS_MAX              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1410
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_CHANNEL_BANDWIDTH            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1411
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_FFT_SIZE                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1412
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_GUARD_INTERVAL               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1413
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_PILOT_PATTERN                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1414
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_T2_FRAMES_PER_SUPER_FRAME    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1415
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_DATA_SYMBOLS_PER_T2_FRAME    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1416
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_SUBSLICES_PER_T2_FRAME       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1417
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_FRAMING_T2_FRAME_DURATION            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1418
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_SCHEDULING                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1419
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_TRANSMISSION_SYSTEM        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1420
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_PEAK_TO_AVERAGE_POWER_RATIO RS_SPECIFIC_PUBLIC_ATTR_BASE + 1421
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_FUTURE_EXTENSION_FRAMES    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1422
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_FEF_TYPE                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1423
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_FEF_LENGTH                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1424
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_FEF_INTERVAL               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1425
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_TIME_FREQUENCY_SLICING     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1426
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_T2_VERSION              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1427
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_POST_MODULATION         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1428
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_REPETITION              RS_SPECIFIC_PUBLIC_ATTR_BASE + 1429
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_POST_EXTENSION          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1430
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_POST_SCRAMBLED          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1431
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_BIAS_BALANCING_BITS     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1432
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_T2_BASE_LITE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1433
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_L1_RF_FREQUENCY            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1434
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_CELL_ID                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 1435
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_NETWORK_ID                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1436
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_T2_SYSTEM_ID                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1437
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_T2_TIMESTAMP_MODE             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1438
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_TIME_OFFSET                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 1439
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_TOTAL_DATA_RATE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1440
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_RESTART                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1441
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_T2_MI_PID                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1442
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_T2_MI_SID                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1443
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_T2_MI_TS_TYPE                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1444
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_T2_MI_PMT_PID                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1445
#define RSBTC_ATTR_MMG_GATEWAY_DVBT2MI_OUTPUT_T2_MI_DATA_RATE               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1446
#define RSBTC_ATTR_POWER_METER_STATE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1211
#define RSBTC_ATTR_POWER_METER_SENSOR_TYPE                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1212
#define RSBTC_ATTR_POWER_METER_POWER                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1213
#define RSBTC_ATTR_POWER_METER_UNIT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1214
#define RSBTC_ATTR_POWER_METER_ZERO                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1215
#define RSBTC_ATTR_POWER_METER_SOURCE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1216
#define RSBTC_ATTR_POWER_METER_SOURCE_FREQUENCY                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1217
#define RSBTC_ATTR_POWER_METER_SOURCE_REL_POWER_STATE                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1218
#define RSBTC_ATTR_POWER_METER_SOURCE_M2REF                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 1219
#define RSBTC_ATTR_POWER_METER_SOURCE_REF_POWER                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1220
#define RSBTC_ATTR_POWER_METER_SOURCE_REL_POWER                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1221
#define RSBTC_ATTR_POWER_METER_AVERAGE_FILTER_TYPE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1222
#define RSBTC_ATTR_POWER_METER_AVERAGE_FILTER_LENGTH                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1223
#define RSBTC_ATTR_POWER_METER_AVERAGE_AUTO_ONCE                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1224
#define RSBTC_ATTR_TSGEN_STATUS                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 946
#define RSBTC_ATTR_TSGEN_SEEK_POSITION                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 947
#define RSBTC_ATTR_TSGEN_PLAYER_PLAYFILE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 948
#define RSBTC_ATTR_TSGEN_PLAYER_FILE_DATE_AND_SIZE                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 949
#define RSBTC_ATTR_TSGEN_ORIGINAL_TS_DATA_RATE                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 950
#define RSBTC_ATTR_TSGEN_PLAYER_BYTE_RATE                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 951
#define RSBTC_ATTR_TSGEN_TS_DATA_RATE                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 952
#define RSBTC_ATTR_TSGEN_PLAY_WINDOW_START                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 953
#define RSBTC_ATTR_TSGEN_PLAY_WINDOW_STOP                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 954
#define RSBTC_ATTR_TSGEN_PACKET_LENGTH                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 955
#define RSBTC_ATTR_TSGEN_TS_SERIAL_OUTPUT                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 956
#define RSBTC_ATTR_TSGEN_ORIGINAL_LOOP_TIME                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 957
#define RSBTC_ATTR_TSGEN_PLAYER_NULLPACKET_STUFFING                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 958
#define RSBTC_ATTR_TSGEN_SEAMLESS_LOOP_CONTINUITY_COUNTER                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 959
#define RSBTC_ATTR_TSGEN_SEAMLESS_LOOP_PCR_DTS_PTS                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 960
#define RSBTC_ATTR_TSGEN_SEAMLESS_LOOP_TDT_TOT                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 961
#define RSBTC_ATTR_TSGEN_STOP_DATA                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 962
#define RSBTC_ATTR_TSGEN_PLAYER_TEST_TS_PACKET                              RS_SPECIFIC_PUBLIC_ATTR_BASE + 963
#define RSBTC_ATTR_TSGEN_PAYLOAD_TEST_STUFF                                 RS_SPECIFIC_PUBLIC_ATTR_BASE + 964
#define RSBTC_ATTR_TSGEN_PLAYER_PRBS                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 965
#define RSBTC_ATTR_SYSTEM_DISPLAY_UPDATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 966
#define RSBTC_ATTR_SYSTEM_KLOCK                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 967
#define RSBTC_ATTR_SYSTEM_REBOOT                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1284
#define RSBTC_ATTR_SYSTEM_RESTART                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1285
#define RSBTC_ATTR_SYSTEM_SHUTDOWN                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1286
#define RSBTC_ATTR_ERROR_LIST                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 968
#define RSBTC_ATTR_PRESET                                                   RS_SPECIFIC_PUBLIC_ATTR_BASE + 969
#define RSBTC_ATTR_RCL                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 970
#define RSBTC_ATTR_SAV                                                      RS_SPECIFIC_PUBLIC_ATTR_BASE + 971
#define RSBTC_ATTR_TX_ADJUST_ALL                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 972
#define RSBTC_ATTR_TX_ADJUST_IO_MODULATOR_LOCAL_AB                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1196
#define RSBTC_ATTR_TX_LAST_FULL_ADJUSTMENT                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 1197
#define RSBTC_ATTR_BER_MEASUREMENT_STATE                                    RS_SPECIFIC_PUBLIC_ATTR_BASE + 987
#define RSBTC_ATTR_BER_MEASUREMENT_RESTART                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 988
#define RSBTC_ATTR_BER_IP_INPUT                                             RS_SPECIFIC_PUBLIC_ATTR_BASE + 1287
#define RSBTC_ATTR_BER_IP_INPUT_ALIAS                                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1288
#define RSBTC_ATTR_BER_IP_INPUT_TYPE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1289
#define RSBTC_ATTR_BER_IP_INPUT_MULTICAST_ADDRESS                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1290
#define RSBTC_ATTR_BER_IP_INPUT_PORT                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1291
#define RSBTC_ATTR_BER_IP_INPUT_IGMPV3_SOURCE_ADDRESS                       RS_SPECIFIC_PUBLIC_ATTR_BASE + 1292
#define RSBTC_ATTR_BER_IP_INPUT_PING_SOURCE_ADDRESS                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1293
#define RSBTC_ATTR_BER_RESULT                                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 1178
#define RSBTC_ATTR_BER_ALL_RESULTS                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1179
#define RSBTC_ATTR_BER_EVALUATION                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1180
#define RSBTC_ATTR_BER_STATE                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 1181
#define RSBTC_ATTR_BER_ERROR_COUNT                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1182
#define RSBTC_ATTR_BER_GATING_TIME                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1183
#define RSBTC_ATTR_BER_GATING_MODE                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 1184
#define RSBTC_ATTR_BER_GATING_TIME_USER                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1185
#define RSBTC_ATTR_BER_INPUT                                                RS_SPECIFIC_PUBLIC_ATTR_BASE + 989
#define RSBTC_ATTR_BER_INPUT_SIGNAL_IP_TS_CHANNEL                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1186
#define RSBTC_ATTR_BER_INPUT_CLOCK                                          RS_SPECIFIC_PUBLIC_ATTR_BASE + 992
#define RSBTC_ATTR_BER_INPUT_DATA                                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 993
#define RSBTC_ATTR_BER_INPUT_ENABLE                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 994
#define RSBTC_ATTR_BER_INPUT_TRESHOLD_CLK_DATA_EN                           RS_SPECIFIC_PUBLIC_ATTR_BASE + 1580
#define RSBTC_ATTR_BER_INPUT_FORMAT                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 1187
#define RSBTC_ATTR_BER_INPUT_SIGNAL                                         RS_SPECIFIC_PUBLIC_ATTR_BASE + 990
#define RSBTC_ATTR_BER_INPUT_PID_FILTER                                     RS_SPECIFIC_PUBLIC_ATTR_BASE + 1188
#define RSBTC_ATTR_BER_INPUT_PID                                            RS_SPECIFIC_PUBLIC_ATTR_BASE + 1189
#define RSBTC_ATTR_BER_INPUT_SETTINGS_PAYLOAD                               RS_SPECIFIC_PUBLIC_ATTR_BASE + 997
#define RSBTC_ATTR_BER_INPUT_SETTINGS_PRBS                                  RS_SPECIFIC_PUBLIC_ATTR_BASE + 996
#define RSBTC_ATTR_ID_QUERY_RESPONSE                                        RS_SPECIFIC_PUBLIC_ATTR_BASE + 1009

// rsbtc_rngRfLevelAttenMode
#define RSBTC_RF_LEVEL_ATTENUATOR_MODE_AUTO                                 0
#define RSBTC_RF_LEVEL_ATTENUATOR_MODE_FIXED                                1
#define RSBTC_RF_LEVEL_ATTENUATOR_MODE_MANUAL                               2

// rsbtc_rngIqcInput
#define RSBTC_IQC_INPUT_TS1                                                 0
#define RSBTC_IQC_INPUT_TS2                                                 1
#define RSBTC_IQC_INPUT_TS3                                                 2
#define RSBTC_IQC_INPUT_IP                                                  3

// rsbtc_rngDTVPayload
#define RSBTC_VAL_DTV_PAYLOAD_PRBS                                          0
#define RSBTC_VAL_DTV_PAYLOAD_H00                                           1
#define RSBTC_VAL_DTV_PAYLOAD_HFF                                           2

// rsbtc_rngDtvPidTestPacket
#define RSBTC_DTV_PID_TEST_NULL                                             0
#define RSBTC_DTV_PID_TEST_VARIABLE                                         1

// rsbtc_rngDTVPrbsMode
#define RSBTC_VAL_DTV_PRBS_P231                                             0
#define RSBTC_VAL_DTV_PRBS_P151                                             1

// rsbtc_rngIqcDirCodeRate
#define RSBTC_VAL_DIRECTV_CODE_RATE_R12                                     0
#define RSBTC_VAL_DIRECTV_CODE_RATE_R23                                     1
#define RSBTC_VAL_DIRECTV_CODE_RATE_R67                                     2

// rsbtc_rngIqcTestSig
#define RSBTC_VAL_TEST_SIGNAL_TTSP                                          0
#define RSBTC_VAL_TEST_SIGNAL_PBEC                                          1
#define RSBTC_VAL_TEST_SIGNAL_PAFC                                          2
#define RSBTC_VAL_TEST_SIGNAL_PBEM                                          3

// rsbtc_rngDTVTestPacketMode
#define RSBTC_DTV_TS_PACKED_MODE_HEAD                                       0
#define RSBTC_DTV_TS_PACKED_MODE_SYNC                                       1

// rsbtc_rngIqcTimeInter
#define RSBTC_IQC_TINT_OFF                                                  0
#define RSBTC_IQC_TINT_I240                                                 1
#define RSBTC_IQC_TINT_I720                                                 2

// rsbtc_rngIqcDvbcCons
#define RSBTC_IQC_DVBC_CONS_16QAM                                           0
#define RSBTC_IQC_DVBC_CONS_32QAM                                           1
#define RSBTC_IQC_DVBC_CONS_64QAM                                           2
#define RSBTC_IQC_DVBC_CONS_128QAM                                          3
#define RSBTC_IQC_DVBC_CONS_256QAM                                          4

// rsbtc_rngIqcDvbcTest
#define RSBTC_IQC_DVBC_TEST_TTSP                                            0
#define RSBTC_IQC_DVBC_TEST_PBEM                                            1
#define RSBTC_IQC_DVBC_TEST_PBDE                                            2

// rsbtc_rngIqcDvbtChanBw
#define RSBTC_IQC_DVBT_CHBW5                                                1
#define RSBTC_IQC_DVBT_CHBW6                                                2
#define RSBTC_IQC_DVBT_CHBW7                                                3
#define RSBTC_IQC_DVBT_CHBW8                                                4

// rsbtc_rngIqcDvbtCons
#define RSBTC_VAL_DVBT_CONSTELLATION_QPSK                                   0
#define RSBTC_VAL_DVBT_CONSTELLATION_16QAM                                  1
#define RSBTC_VAL_DVBT_CONSTELLATION_64QAM                                  2

// rsbtc_rngIqcDvbtFftMode
#define RSBTC_VAL_DVBT_FFT_MODE_2K                                          0
#define RSBTC_VAL_DVBT_FFT_MODE_4K                                          1
#define RSBTC_VAL_DVBT_FFT_MODE_8K                                          2

// rsbtc_rngIqcDvbtGuarInt
#define RSBTC_VAL_DVBT_GUARD_INTERVAL_14                                    0
#define RSBTC_VAL_DVBT_GUARD_INTERVAL_18                                    1
#define RSBTC_VAL_DVBT_GUARD_INTERVAL_16                                    2
#define RSBTC_VAL_DVBT_GUARD_INTERVAL_32                                    3

// rsbtc_rngIqcDvbtHier
#define RSBTC_IQC_DVBT_HIERARCHY_NON                                        0
#define RSBTC_IQC_DVBT_HIERARCHY_ALPHA1                                     1
#define RSBTC_IQC_DVBT_HIERARCHY_ALPHA2                                     2
#define RSBTC_IQC_DVBT_HIERARCHY_ALPHA4                                     3

// rsbtc_rngIqcCodeRate
#define RSBTC_VAL_CODE_RATE_12                                              0
#define RSBTC_VAL_CODE_RATE_23                                              1
#define RSBTC_VAL_CODE_RATE_34                                              2
#define RSBTC_VAL_CODE_RATE_56                                              3
#define RSBTC_VAL_CODE_RATE_78                                              4
#define RSBTC_VAL_CODE_RATE_89                                              5

// rsbtc_rngIqcDvbsCons
#define RSBTC_VAL_DVBS_CONSTELLATION_QPSK                                   0
#define RSBTC_VAL_DVBS_CONSTELLATION_8PSK                                   1
#define RSBTC_VAL_DVBS_CONSTELLATION_16QAM                                  2

// rsbtc_rngIqcDvbsRolloff
#define RSBTC_VAL_DVBS_ROLL_OFF_0_20                                        0
#define RSBTC_VAL_DVBS_ROLL_OFF_0_25                                        1
#define RSBTC_VAL_DVBS_ROLL_OFF_0_35                                        2

// rsbtc_rngIqcDvbs2TestSig
#define RSBTC_IQC_TEST_TTSP                                                 0

// rsbtc_rngIqcDvbs2Cons
#define RSBTC_IQC_DVBS2_CONS_QPSK                                           0
#define RSBTC_IQC_DVBS2_CONS_8PSK                                           1
#define RSBTC_IQC_DVBS2_CONS_16APSK                                         2
#define RSBTC_IQC_DVBS2_CONS_32APSK                                         3

// rsbtc_rngIqcDvbs2Rolloff
#define RSBTC_VAL_DVBS2_ROLL_OFF_0_05                                       0
#define RSBTC_VAL_DVBS2_ROLL_OFF_0_10                                       1
#define RSBTC_VAL_DVBS2_ROLL_OFF_0_15                                       2
#define RSBTC_VAL_DVBS2_ROLL_OFF_0_20                                       3
#define RSBTC_VAL_DVBS2_ROLL_OFF_0_25                                       4
#define RSBTC_VAL_DVBS2_ROLL_OFF_0_35                                       5

// rsbtc_rngSweepMode
#define RSBTC_VAL_SWEEP_MODE_AUTO                                           0
#define RSBTC_VAL_SWEEP_MODE_MAN                                            1
#define RSBTC_VAL_SWEEP_MODE_STEP                                           2

// rsbtc_rngFreqSweepSpacing
#define RSBTC_VAL_FREQ_SWEEP_SPACING_LIN                                    0
#define RSBTC_VAL_FREQ_SWEEP_SPACING_LOG                                    1

// rsbtc_rngBerSelMeas
#define RSBTC_VAL_BER_INPUT_SERIAL                                          0
#define RSBTC_VAL_BER_INPUT_TS1                                             1
#define RSBTC_VAL_BER_INPUT_TS2                                             2
#define RSBTC_VAL_BER_INPUT_TS3A                                            3
#define RSBTC_VAL_BER_INPUT_TS3B                                            4
#define RSBTC_VAL_BER_INPUT_IPA                                             5
#define RSBTC_VAL_BER_INPUT_IPB                                             6

// rsbtc_rngBerInputPolarity
#define RSBTC_BER_INPUT_POLARITY_NORMAL                                     0
#define RSBTC_BER_INPUT_POLARITY_INVERTED                                   1
#define RSBTC_BER_INPUT_POLARITY_NORMAL1K                                   2
#define RSBTC_BER_INPUT_POLARITY_INVERTED1K                                 3

// rsbtc_rngBerFuncInp
#define RSBTC_BER_FUNC_INPUT_ALWAYS                                         0
#define RSBTC_BER_FUNC_INPUT_HIGH                                           1
#define RSBTC_BER_FUNC_INPUT_LOW                                            2
#define RSBTC_BER_VAL_INPUT_ALWAYS1K                                        3
#define RSBTC_BER_VAL_INPUT_HIGH1K                                          4
#define RSBTC_BER_VAL_INPUT_LOWL1K                                          5

// rsbtc_rngBerDataStream
#define RSBTC_VAL_BER_SIGNAL_STUF                                           0
#define RSBTC_VAL_BER_SIGNAL_S187                                           1
#define RSBTC_VAL_BER_SIGNAL_S203                                           2
#define RSBTC_VAL_BER_SIGNAL_S207                                           3
#define RSBTC_VAL_BER_SIGNAL_H184                                           4
#define RSBTC_VAL_BER_SIGNAL_H200                                           5
#define RSBTC_VAL_BER_SIGNAL_H204                                           6

// rsbtc_rngBerGatMode
#define RSBTC_VAL_BER_GATING_INFINITE                                       0
#define RSBTC_VAL_BER_GATING_AUTO                                           1
#define RSBTC_VAL_BER_GATING_UWINDOW                                        2
#define RSBTC_VAL_BER_GATING_USINGLE                                        3

// rsbtc_rngModGain
#define RSBTC_VAL_MOD_GAIN_DB0                                              0
#define RSBTC_VAL_MOD_GAIN_DB2                                              1
#define RSBTC_VAL_MOD_GAIN_DB4                                              2
#define RSBTC_VAL_MOD_GAIN_DB6                                              3
#define RSBTC_VAL_MOD_GAIN_DB8                                              4
#define RSBTC_VAL_MOD_GAIN_DBM2                                             5
#define RSBTC_VAL_MOD_GAIN_DBM4                                             6

// rsbtc_rngDmSigSource
#define RSBTC_VAL_BB_SIGNAL_TYPE_ARB                                        0
#define RSBTC_VAL_BB_SIGNAL_TYPE_DTV                                        1
#define RSBTC_VAL_BB_SIGNAL_TYPE_ATV                                        2
#define RSBTC_VAL_BB_SIGNAL_TYPE_AUDIO_BC                                   3
#define RSBTC_VAL_BB_SIGNAL_TYPE_DIGITAL                                    4
#define RSBTC_VAL_BB_SIGNAL_TYPE_OFF                                        5

// rsbtc_rngDmInterferSource
#define RSBTC_VAL_DM_INTERFERER_SOURCE_OFF                                  0
#define RSBTC_VAL_DM_INTERFERER_SOURCE_ARB                                  1
#define RSBTC_VAL_DM_INTERFERER_SOURCE_DIGITAL                              2
#define RSBTC_VAL_DM_INTERFERER_SOURCE_SIGGEN                               3

// rsbtc_rngDmInterferAdd
#define RSBTC_VAL_DM_INTERFERER_ADD_BEFF                                    1
#define RSBTC_VAL_DM_INTERFERER_ADD_AFF                                     2
#define RSBTC_VAL_DM_INTERFERER_ADD_TXDIST                                  3
#define RSBTC_VAL_DM_INTERFERER_ADD_MODUL                                   4
#define RSBTC_VAL_DM_INTERFERER_ADD_FADING                                  5
#define RSBTC_VAL_DM_INTERFERER_ADD_NOISE                                   6
#define RSBTC_VAL_DM_INTERFERER_ADD_RXDIST                                  7

// rsbtc_rngDmTransmissionStandard
#define RSBTC_VAL_DM_STANDARD_DVBT                                          0
#define RSBTC_VAL_DM_STANDARD_DVBC                                          1
#define RSBTC_VAL_DM_STANDARD_DVBS                                          2
#define RSBTC_VAL_DM_STANDARD_8VSB                                          3
#define RSBTC_VAL_DM_STANDARD_J83B                                          4
#define RSBTC_VAL_DM_STANDARD_ISDBT                                         5
#define RSBTC_VAL_DM_STANDARD_DVBS2                                         6
#define RSBTC_VAL_DM_STANDARD_DIRECTV                                       7
#define RSBTC_VAL_DM_STANDARD_DTMB                                          8
#define RSBTC_VAL_DM_STANDARD_DVBT2                                         9
#define RSBTC_VAL_DM_STANDARD_ATSM                                          10
#define RSBTC_VAL_DM_STANDARD_DVBC2                                         11
#define RSBTC_VAL_DM_STANDARD_TDMB                                          12
#define RSBTC_VAL_DM_STANDARD_CMMB                                          13
#define RSBTC_VAL_DM_STANDARD_ISDBTMM                                       14
#define RSBTC_VAL_DM_STANDARD_ATSC3                                         15

// rsbtc_rngDmSpectPolar
#define RSBTC_VAL_DM_SPEC_POLAR_NORM                                        0
#define RSBTC_VAL_DM_SPEC_POLAR_INV                                         1

// rsbtc_rngNoiseState
#define RSBTC_VAL_NOISE_OFF                                                 0
#define RSBTC_VAL_NOISE_ADD                                                 1
#define RSBTC_VAL_NOISE_ONLY                                                2

// rsbtc_rngPowAlcState
#define RSBTC_VAL_POWER_ALC_OFF                                             0
#define RSBTC_VAL_POWER_ALC_ON                                              1
#define RSBTC_VAL_POWER_ALC_AUTO                                            2
#define RSBTC_VAL_POWER_ALC_SAMPLE_ON                                       3
#define RSBTC_VAL_POWER_ALC_TABLE_OFF                                       4
#define RSBTC_VAL_POWER_ALC_TABLE_ON                                        5

// rsbtc_rngPowLevUnit
#define RSBTC_VAL_POW_LEV_DBM                                               0
#define RSBTC_VAL_POW_LEV_DBUV                                              1
#define RSBTC_VAL_POW_LEV_MV                                                2
#define RSBTC_VAL_POW_LEV_DBMV                                              3
#define RSBTC_VAL_POW_LEV_UV                                                4

// rsbtc_rngExternalRefFrequency
#define RSBTC_VAL_EXT_REF_FREQ_5M                                           0
#define RSBTC_VAL_EXT_REF_FREQ_10M                                          1
#define RSBTC_VAL_EXT_REF_FREQ_13M                                          2

// rsbtc_rngExtRefFreqSyncBw
#define RSBTC_VAL_EXT_REF_FREQ_SYNC_BW_STANDARD                             0
#define RSBTC_VAL_EXT_REF_FREQ_SYNC_BW_NARROW                               1

// rsbtc_rngRoscSource
#define RSBTC_VAL_ROSC_SOURCE_INT                                           0
#define RSBTC_VAL_ROSC_SOURCE_EXT                                           1

// rsbtc_rngFadingProfile
#define RSBTC_VAL_FADING_PROFILE_PURE_DOPPLER                               0
#define RSBTC_VAL_FADING_PROFILE_STATIC_PATH                                1
#define RSBTC_VAL_FADING_PROFILE_RAYLEIGH                                   2
#define RSBTC_VAL_FADING_PROFILE_RICE                                       3
#define RSBTC_VAL_FADING_PROFILE_CONST_PHASE                                4
#define RSBTC_VAL_FADING_PROFILE_GAUSS1                                     5
#define RSBTC_VAL_FADING_PROFILE_GAUSS2                                     6
#define RSBTC_VAL_FADING_PROFILE_GAUSS_DAB                                  7
#define RSBTC_VAL_FADING_PROFILE_GAUSS_DOPPLER                              8
#define RSBTC_VAL_FADING_PROFILE_GAUSS01                                    9
#define RSBTC_VAL_FADING_PROFILE_GAUSS008                                   10
#define RSBTC_VAL_FADING_PROFILE_GAUSS_WATTERS                              11
#define RSBTC_VAL_FADING_PROFILE_WM_DOPPLER                                 12
#define RSBTC_VAL_FADING_PROFILE_WM_RICE                                    13

// rsbtc_rngFadingParam
#define RSBTC_VAL_FADING_PAR_EASY3                                          0
#define RSBTC_VAL_FADING_PAR_ECHO                                           1
#define RSBTC_VAL_FADING_PAR_FX                                             2
#define RSBTC_VAL_FADING_PAR_PT                                             3
#define RSBTC_VAL_FADING_PAR_SFN                                            4
#define RSBTC_VAL_FADING_PAR_TU6                                            5
#define RSBTC_VAL_FADING_PAR_RA6                                            6
#define RSBTC_VAL_FADING_PAR_RC6                                            7
#define RSBTC_VAL_FADING_PAR_RL6                                            8
#define RSBTC_VAL_FADING_PAR_RED                                            9
#define RSBTC_VAL_FADING_PAR_ET50                                           10
#define RSBTC_VAL_FADING_PAR_VAL                                            11
#define RSBTC_VAL_FADING_PAR_RC12                                           12
#define RSBTC_VAL_FADING_PAR_RL12                                           13
#define RSBTC_VAL_FADING_PAR_TU3                                            14
#define RSBTC_VAL_FADING_PAR_TU50                                           15
#define RSBTC_VAL_FADING_PAR_HT100                                          16
#define RSBTC_VAL_FADING_PAR_AA_S                                           17
#define RSBTC_VAL_FADING_PAR_AB_S                                           18
#define RSBTC_VAL_FADING_PAR_AC_S                                           19
#define RSBTC_VAL_FADING_PAR_AD_S                                           20
#define RSBTC_VAL_FADING_PAR_AE_S                                           21
#define RSBTC_VAL_FADING_PAR_AF_S                                           22
#define RSBTC_VAL_FADING_PAR_AG_S                                           23
#define RSBTC_VAL_FADING_PAR_B_ECHO                                         24
#define RSBTC_VAL_FADING_PAR_C1R                                            25
#define RSBTC_VAL_FADING_PAR_C2R                                            26
#define RSBTC_VAL_FADING_PAR_C3R                                            27
#define RSBTC_VAL_FADING_PAR_RC20                                           28
#define RSBTC_VAL_FADING_PAR_RL20                                           29
#define RSBTC_VAL_FADING_PAR_HIC                                            30
#define RSBTC_VAL_FADING_PAR_HNIC                                           31
#define RSBTC_VAL_FADING_PAR_HPT                                            32
#define RSBTC_VAL_FADING_PAR_HNPT                                           33
#define RSBTC_VAL_FADING_PAR_MBR                                            34
#define RSBTC_VAL_FADING_PAR_A_BR                                           35
#define RSBTC_VAL_FADING_PAR_B_B                                            36
#define RSBTC_VAL_FADING_PAR_C_BR                                           37
#define RSBTC_VAL_FADING_PAR_CS_B                                           38
#define RSBTC_VAL_FADING_PAR_CM_B                                           39
#define RSBTC_VAL_FADING_PAR_D_BR                                           40
#define RSBTC_VAL_FADING_PAR_DM_B                                           41
#define RSBTC_VAL_FADING_PAR_E_BR                                           42
#define RSBTC_VAL_FADING_PAR_D1CR                                           43
#define RSBTC_VAL_FADING_PAR_D2CR                                           44
#define RSBTC_VAL_FADING_PAR_D3CR                                           45
#define RSBTC_VAL_FADING_PAR_D4CR                                           46
#define RSBTC_VAL_FADING_PAR_A_AC                                           47
#define RSBTC_VAL_FADING_PAR_IEC62                                          55
#define RSBTC_VAL_FADING_PAR_RA4                                            56
#define RSBTC_VAL_FADING_PAR_RA6DAB                                         57
#define RSBTC_VAL_FADING_PAR_TU12DAB                                        58
#define RSBTC_VAL_FADING_PAR_TU6DAB                                         59
#define RSBTC_VAL_FADING_PAR_DAB_SFN                                        60
#define RSBTC_VAL_FADING_PAR_PI                                             61
#define RSBTC_VAL_FADING_PAR_PO                                             62
#define RSBTC_VAL_FADING_PAR_VU                                             63
#define RSBTC_VAL_FADING_PAR_MR                                             64

// rsbtc_rngFsimSpeedUnit
#define RSBTC_VAL_FADING_SPEED_MPS                                          0
#define RSBTC_VAL_FADING_SPEED_KMH                                          1
#define RSBTC_VAL_FADING_SPEED_MPH                                          2

// rsbtc_rngFsimDedicTo
#define RSBTC_VAL_FADING_DEDICATED_TO_BB                                    0
#define RSBTC_VAL_FADING_DEDICATED_TO_RF                                    1

// rsbtc_rngFsimInsLossMode
#define RSBTC_VAL_FADING_LOSS_MODE_NORMAL                                   0
#define RSBTC_VAL_FADING_LOSS_MODE_LOW_ACP                                  1
#define RSBTC_VAL_FADING_LOSS_MODE_USER                                     2

// rsbtc_rngFsimRestartMode
#define RSBTC_VAL_FADING_RESTART_MODE_AUTO                                  0
#define RSBTC_VAL_FADING_RESTART_MODE_MAN                                   1
#define RSBTC_VAL_FADING_RESTART_MODE_TRIGGER                               2

// rsbtc_rngIqcDvbs2CodeRate
#define RSBTC_IQC_DVBS2_CRATE_12                                            0
#define RSBTC_IQC_DVBS2_CRATE_35                                            1
#define RSBTC_IQC_DVBS2_CRATE_23                                            2
#define RSBTC_IQC_DVBS2_CRATE_34                                            3
#define RSBTC_IQC_DVBS2_CRATE_45                                            4
#define RSBTC_IQC_DVBS2_CRATE_56                                            5
#define RSBTC_IQC_DVBS2_CRATE_67                                            6
#define RSBTC_IQC_DVBS2_CRATE_78                                            7
#define RSBTC_IQC_DVBS2_CRATE_89                                            8
#define RSBTC_IQC_DVBS2_CRATE_910                                           9
#define RSBTC_IQC_DVBS2_CRATE_13                                            10
#define RSBTC_IQC_DVBS2_CRATE_14                                            11
#define RSBTC_IQC_DVBS2_CRATE_25                                            12

// rsbtc_rngIqcAtvSoundMode
#define RSBTC_VAL_ATV_SOUND_MODE_OFF                                        0
#define RSBTC_VAL_ATV_SOUND_MODE_FM_MONO                                    1
#define RSBTC_VAL_ATV_SOUND_MODE_FM_STEREO                                  2
#define RSBTC_VAL_ATV_SOUND_MODE_FM_DUAL                                    3
#define RSBTC_VAL_ATV_SOUND_MODE_NICAM                                      4
#define RSBTC_VAL_ATV_SOUND_MODE_AM_MONO                                    5
#define RSBTC_VAL_ATV_SOUND_MODE_MONO_NICAM                                 6
#define RSBTC_VAL_ATV_SOUND_MODE_FM_STEREO_KOREA                            7
#define RSBTC_VAL_ATV_SOUND_MODE_FM_DUAL_KOREA                              8
#define RSBTC_VAL_ATV_SOUND_MODE_BTSC                                       9

// rsbtc_rngTsgenPacketLength
#define RSBTC_VAL_MM_GENERATOR_PLENGTH_188                                  0
#define RSBTC_VAL_MM_GENERATOR_PLENGTH_204                                  1
#define RSBTC_VAL_MM_GENERATOR_PLENGTH_208                                  2

// rsbtc_rngMMGeneratorCommand
#define RSBTC_VAL_MM_GENERATOR_STOP                                         0
#define RSBTC_VAL_MM_GENERATOR_PLAY                                         1
#define RSBTC_VAL_MM_GENERATOR_PAUSE                                        2

// rsbtc_rngIqcDirCons
#define RSBTC_VAL_DIRECTV_CONSTELLATION_QPSK                                0

// rsbtc_rngIqcDirTestPacketMode
#define RSBTC_VAL_DIRECTV_TS_PACKED_MODE_H127                               0
#define RSBTC_VAL_DIRECTV_TS_PACKED_MODE_P130                               1

// rsbtc_rngDmAtvTransmStd
#define RSBTC_VAL_DM_ATV_STANDARD_BG                                        0
#define RSBTC_VAL_DM_ATV_STANDARD_DK                                        1
#define RSBTC_VAL_DM_ATV_STANDARD_I                                         2
#define RSBTC_VAL_DM_ATV_STANDARD_MN                                        3
#define RSBTC_VAL_DM_ATV_STANDARD_L                                         4

// rsbtc_rngIqcDirRollOff
#define RSBTC_VAL_DIRECTV_ROLL_OFF_0_15                                     0
#define RSBTC_VAL_DIRECTV_ROLL_OFF_0_20                                     1
#define RSBTC_VAL_DIRECTV_ROLL_OFF_0_25                                     2
#define RSBTC_VAL_DIRECTV_ROLL_OFF_0_35                                     3

// rsbtc_rngIqcJ83BTestSig
#define RSBTC_VAL_IQC_J83B_TEST_TTSP                                        0
#define RSBTC_VAL_IQC_J83B_TEST_PBET                                        1
#define RSBTC_VAL_IQC_J83B_TEST_PBEM                                        2

// rsbtc_rngIqcVsbTestSig
#define RSBTC_VAL_DTV_VSB_TEST_TTSP                                         0
#define RSBTC_VAL_DTV_VSB_TEST_PBIN                                         1
#define RSBTC_VAL_DTV_VSB_TEST_PBET                                         2
#define RSBTC_VAL_DTV_VSB_TEST_PBEM                                         3

// rsbtc_rngIqcAtvVisionPict
#define RSBTC_VAL_ATV_VISION_PICT_C75P                                      0
#define RSBTC_VAL_ATV_VISION_PICT_C75N                                      1
#define RSBTC_VAL_ATV_VISION_PICT_C75S                                      2
#define RSBTC_VAL_ATV_VISION_PICT_FUBP                                      3
#define RSBTC_VAL_ATV_VISION_PICT_CPAM                                      4
#define RSBTC_VAL_ATV_VISION_PICT_CPAN                                      5
#define RSBTC_VAL_ATV_VISION_PICT_LIBR                                      6

// rsbtc_rngIqcJ83BConstellation
#define RSBTC_VAL_IQC_J83B_CONSTEL_J64                                      0
#define RSBTC_VAL_IQC_J83B_CONSTEL_J256                                     1
#define RSBTC_VAL_IQC_J83B_CONSTEL_J1024                                    2

// rsbtc_rngIqcJ83BIntrleaverMode
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_0                                    0
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_1                                    1
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_2                                    2
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_3                                    3
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_4                                    4
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_5                                    5
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_6                                    6
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_7                                    7
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_8                                    8
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_9                                    9
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_10                                   10
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_12                                   12
#define RSBTC_VAL_IQC_J83B_INTERLEAVER_14                                   14

// rsbtc_rngIqcPackLength
#define RSBTC_VAL_IQC_PACKED_LENGTH_INV                                     0
#define RSBTC_VAL_IQC_PACKED_LENGTH_P188                                    1

// rsbtc_rngIqcPackedLength
#define RSBTC_VAL_IQC_PACK_LEN_INV                                          0
#define RSBTC_VAL_IQC_PACK_LEN_P188                                         1
#define RSBTC_VAL_IQC_PACK_LEN_P204                                         2

// rsbtc_rngIqcIsdbtPortion
#define RSBTC_VAL_ISDBT_PORTION_PCC                                         0
#define RSBTC_VAL_ISDBT_PORTION_DCC                                         1
#define RSBTC_VAL_ISDBT_PORTION_CCC                                         2
#define RSBTC_VAL_ISDBT_PORTION_PDD                                         3
#define RSBTC_VAL_ISDBT_PORTION_PDC                                         4
#define RSBTC_VAL_ISDBT_PORTION_DDD                                         5
#define RSBTC_VAL_ISDBT_PORTION_DDC                                         6

// rsbtc_rngIqcIsdbtConstel
#define RSBTC_IQC_VAL_ISDBT_CONSTEL_DQPSK                                   0
#define RSBTC_IQC_VAL_ISDBT_CONSTEL_QPSK                                    1
#define RSBTC_IQC_VAL_ISDBT_CONSTEL_16QAM                                   2
#define RSBTC_IQC_VAL_ISDBT_CONSTEL_64QAM                                   3

// rsbtc_rngIqcIsdbtInterleaving
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_0                                    0
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_1                                    1
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_2                                    2
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_4                                    3
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_8                                    4
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_16                                   5
#define RSBTC_VAL_IQC_ISDBT_TIME_INTER_32                                   6

// rsbtc_rngIqcIsdbtMode
#define RSBTC_VAL_IQC_ISDBT_MODE_2K                                         0
#define RSBTC_VAL_IQC_ISDBT_MODE_4K                                         1
#define RSBTC_VAL_IQC_ISDBT_MODE_8K                                         2

// rsbtc_rngIqcIsdbtGuardInt
#define RSBTC_VAL_IQC_ISDBT_GUARD_1_4                                       0
#define RSBTC_VAL_IQC_ISDBT_GUARD_1_8                                       1
#define RSBTC_VAL_IQC_ISDBT_GUARD_1_16                                      2
#define RSBTC_VAL_IQC_ISDBT_GUARD_1_32                                      3

// rsbtc_rngIqcIsdbtAcData
#define RSBTC_VAL_ISDBT_AC_DATA_PRBS                                        0
#define RSBTC_VAL_ISDBT_AC_DATA_ALL1                                        1

// rsbtc_rngIqcAtvPreemphasis
#define RSBTC_VAL_IQC_ATV_PREEMPHASIS_OFF                                   0
#define RSBTC_VAL_IQC_ATV_PREEMPHASIS_D50                                   1
#define RSBTC_VAL_IQC_ATV_PREEMPHASIS_D75                                   2

// rsbtc_rngFsimp2dynProfile
#define RSBTC_VAL_FADING_P2DYN_PROF_PDOP                                    0
#define RSBTC_VAL_FADING_P2DYN_PROF_SPAT                                    1
#define RSBTC_VAL_FADING_P2DYN_PROF_RAYL                                    2

// rsbtc_rngFsimP2dynMovMode
#define RSBTC_VAL_FADING_P2DYN_MOV_MODE_HOPP                                0
#define RSBTC_VAL_FADING_P2DYN_MOV_MODE_SLID                                1

// rsbtc_rngFreqVsbFreq
#define RSBTC_VAL_FREQ_VSB_FREQ_PILOT                                       0
#define RSBTC_VAL_FREQ_VSB_FREQ_CENTER                                      1

// rsbtc_rngSystemKLock
#define RSBTC_VAL_SYST_KLOCK_OFF                                            0
#define RSBTC_VAL_SYST_KLOCK_ON                                             1

// rsbtc_rngIqcAtvAudioSource
#define RSBTC_VAL_ATV_AUDIO_SOURCE_GENERATOR                                1
#define RSBTC_VAL_ATV_AUDIO_SOURCE_PLAYER                                   2

// rsbtc_rngIqcAtvSpecialVsbChar
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_DKFM                             0
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_DK                               1
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_DKN                              2
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_I                                3
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_I1                               4
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_M                                5
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_N                                6
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_BG                               7
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_L                                8
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_LNIC                             9
#define RSBTC_VAL_IQC_ATV_SPECIAL_VSB_CHAR_BGA                              10

// rsbtc_rngIqcAtvVideoInput
#define RSBTC_VAL_IQC_ATV_VIDEO_SOURCE_GENERATOR                            0

// rsbtc_rngFsimReference
#define RSBTC_VAL_FADING_REFERENCE_SPEED                                    0
#define RSBTC_VAL_FADING_REFERENCE_DOPPLER_FREQ                             1

// rsbtc_rngFsimConfig
#define RSBTC_VAL_FADING_STANDARD_DELAY                                     0
#define RSBTC_VAL_FADING_BIRTH_DEATH_PROP                                   1
#define RSBTC_VAL_FADING_MOVING_PROPAGATION                                 2
#define RSBTC_VAL_FADING_2CH_INTERFERER                                     3

// rsbtc_rngIqcDvbcRollOff
#define RSBTC_VAL_DVBC_ROLL_OFF_0_13                                        1
#define RSBTC_VAL_DVBC_ROLL_OFF_0_15                                        2

// rsbtc_rngIqcSigSource
#define RSBTC_VAL_DTV_SIGNAL_SOURCE_EXTERNAL                                0
#define RSBTC_VAL_DTV_SIGNAL_SOURCE_MMGEN                                   1
#define RSBTC_VAL_DTV_SIGNAL_SOURCE_TEST                                    2

// rsbtc_rngIqcDtmbTestSig
#define RSBTC_DTMB_TEST_TTSP                                                0

// rsbtc_rngIqcDtmbConstel
#define RSBTC_IQC_DTMB_CONST_QAM                                            0
#define RSBTC_IQC_DTMB_CONST_QAM_NR                                         1
#define RSBTC_IQC_DTMB_CONST_16QAM                                          2
#define RSBTC_IQC_DTMB_CONST_32QAM                                          3
#define RSBTC_IQC_DTMB_CONST_64QAM                                          4

// rsbtc_rngIqcDtmbCodeRate
#define RSBTC_IQC_DTMB_CRATE_04                                             0
#define RSBTC_IQC_DTMB_CRATE_06                                             1
#define RSBTC_IQC_DTMB_CRATE_08                                             2

// rsbtc_rngIqcDtmbGuar
#define RSBTC_IQC_DTMB_GUAR_G420                                            0
#define RSBTC_IQC_DTMB_GUAR_G595                                            1
#define RSBTC_IQC_DTMB_GUAR_G945                                            2

// rsbtc_rngIqcDtmbChanBw
#define RSBTC_IQC_DTMB_CHBW6                                                1
#define RSBTC_IQC_DTMB_CHBW7                                                2
#define RSBTC_IQC_DTMB_CHBW8                                                3

// rsbtc_rngIqcDtmbGic
#define RSBTC_VAL_IQC_DTMB_GIC_VAR                                          0
#define RSBTC_VAL_IQC_DTMB_GIC_CONST                                        1

// rsbtc_rngIqcIsdbtSystem
#define RSBTC_VAL_IQC_ISDBT_SYSTEM_T                                        0
#define RSBTC_VAL_IQC_ISDBT_SYSTEM_TSB1                                     1
#define RSBTC_VAL_IQC_ISDBT_SYSTEM_TSB3                                     2

// rsbtc_rngDmInterferRef
#define RSBTC_VAL_DM_INTERFERER_REF_LEVEL                                   0
#define RSBTC_VAL_DM_INTERFERER_REF_ATTEN                                   1

// rsbtc_rngBBImpairOptMode
#define RSBTC_VAL_BB_IMPAIR_MODE_FAST                                       0
#define RSBTC_VAL_BB_IMPAIR_MODE_QHIGH                                      1

// rsbtc_rngIqcDvbs2Fecframe
#define RSBTC_IQC_DVBS2_FEC_FRAME_NORMAL                                    0
#define RSBTC_IQC_DVBS2_FEC_FRAME_SHORT                                     1
#define RSBTC_IQC_DVBS2_FEC_FRAME_MEDIUM                                    2

// rsbtc_rngDmAudioBcStd
#define RSBTC_VAL_AUDIO_BC_STANDARD_AM                                      0
#define RSBTC_VAL_AUDIO_BC_STANDARD_FM                                      1

// rsbtc_rngAudioBcInput
#define RSBTC_VAL_AUDIO_BC_INPUT_AUDIO_GENERATOR                            0
#define RSBTC_VAL_AUDIO_BC_INPUT_AUDIO_PLAYER                               1

// rsbtc_rngAudioBcFmMode
#define RSBTC_VAL_AUDIO_BC_AF_MODE_LEFT                                     0
#define RSBTC_VAL_AUDIO_BC_AF_MODE_RIGH                                     1
#define RSBTC_VAL_AUDIO_BC_AF_MODE_REL                                      2
#define RSBTC_VAL_AUDIO_BC_AF_MODE_REM                                      3
#define RSBTC_VAL_AUDIO_BC_AF_MODE_RNE                                      4

// rsbtc_rngAudioBcFmSignalModulationType
#define RSBTC_VAL_AUDIO_BC_SIGNAL_TYPE_MONO                                 0
#define RSBTC_VAL_AUDIO_BC_SIGNAL_TYPE_STEREO                               1

// rsbtc_rngAudioBcMoPreemphasis
#define RSBTC_VAL_AUDIO_BC_PREEMPHASIS_OFF                                  0
#define RSBTC_VAL_AUDIO_BC_PREEMPHASIS_D50                                  1
#define RSBTC_VAL_AUDIO_BC_PREEMPHASIS_D75                                  2

// rsbtc_rngIcgAudioFmRdsMs
#define RSBTC_VAL_AUDIO_BC_RDS_MS_MUSIC                                     0
#define RSBTC_VAL_AUDIO_BC_RDS_MS_SPEECH                                    1

// rsbtc_rngIcgAudioFmRdsMethod
#define RSBTC_VAL_AUDIO_BC_AF_METHOD_A                                      0
#define RSBTC_VAL_AUDIO_BC_AF_METHOD_B                                      1

// rsbtc_rngIcgAudioFmAfBOrder
#define RSBTC_VAL_AUDIO_BC_AFB_ORDER_ASC                                    0
#define RSBTC_VAL_AUDIO_BC_AFB_ORDER_DESC                                   1

// rsbtc_rngIcgAudioFmEonMethod
#define RSBTC_VAL_AUDIO_BC_EON_METHOD_A                                     0
#define RSBTC_VAL_AUDIO_BC_EON_METHOD_MAPED_FREQ                            1

// rsbtc_rngDvbt2InpFormat
#define RSBTC_VAL_DVBT2_INPUT_FORMAT_TS                                     0
#define RSBTC_VAL_DVBT2_INPUT_FORMAT_GFPS                                   1
#define RSBTC_VAL_DVBT2_INPUT_FORMAT_GCS                                    2
#define RSBTC_VAL_DVBT2_INPUT_FORMAT_GSE                                    3

// rsbtc_rngDvbt2PlpType
#define RSBTC_VAL_DVBT2_PLP_TYPE_DT1                                        0
#define RSBTC_VAL_DVBT2_PLP_TYPE_DT2                                        1
#define RSBTC_VAL_DVBT2_PLP_TYPE_COMM                                       2

// rsbtc_rngDvbt2BbMode
#define RSBTC_VAL_DVBT2_BB_MODE_HEM                                         0
#define RSBTC_VAL_DVBT2_BB_MODE_NM                                          1

// rsbtc_rngDvbt2BicmFec
#define RSBTC_VAL_DVBT2_BICM_FEC_NORMAL                                     0
#define RSBTC_VAL_DVBT2_BICM_FEC_SHORT                                      1

// rsbtc_rngDvbt2CRate
#define RSBTC_VAL_IQC_DVBT2_CRATE_R12                                       0
#define RSBTC_VAL_IQC_DVBT2_CRATE_R23                                       1
#define RSBTC_VAL_IQC_DVBT2_CRATE_R34                                       2
#define RSBTC_VAL_IQC_DVBT2_CRATE_R56                                       3
#define RSBTC_VAL_IQC_DVBT2_CRATE_R35                                       4
#define RSBTC_VAL_IQC_DVBT2_CRATE_R45                                       5
#define RSBTC_VAL_IQC_DVBT2_CRATE_R13                                       6
#define RSBTC_VAL_IQC_DVBT2_CRATE_R25                                       7

// rsbtc_rngDvbt2Constel
#define RSBTC_VAL_IQC_DVBT2_CONSTEL_T4                                      0
#define RSBTC_VAL_IQC_DVBT2_CONSTEL_T16                                     1
#define RSBTC_VAL_IQC_DVBT2_CONSTEL_T64                                     2
#define RSBTC_VAL_IQC_DVBT2_CONSTEL_T256                                    3

// rsbtc_rngDvbt2TimeInt
#define RSBTC_VAL_IQC_DVBT2_TIME_TYPE_T0                                    0
#define RSBTC_VAL_IQC_DVBT2_TIME_TYPE_T1                                    1

// rsbtc_rngDvbt2ChBWidth
#define RSBTC_VAL_IQC_DVBT2_CH_BWIDTH_BW_2                                  0
#define RSBTC_VAL_IQC_DVBT2_CH_BWIDTH_BW_5                                  1
#define RSBTC_VAL_IQC_DVBT2_CH_BWIDTH_BW_6                                  2
#define RSBTC_VAL_IQC_DVBT2_CH_BWIDTH_BW_7                                  3
#define RSBTC_VAL_IQC_DVBT2_CH_BWIDTH_BW_8                                  4

// rsbtc_rngDvbt2FftMode
#define RSBTC_VAL_IQC_DVBT2_FFT_M1K                                         0
#define RSBTC_VAL_IQC_DVBT2_FFT_M2K                                         1
#define RSBTC_VAL_IQC_DVBT2_FFT_M4K                                         2
#define RSBTC_VAL_IQC_DVBT2_FFT_M8K                                         3
#define RSBTC_VAL_IQC_DVBT2_FFT_M8E                                         4
#define RSBTC_VAL_IQC_DVBT2_FFT_M16K                                        5
#define RSBTC_VAL_IQC_DVBT2_FFT_M16E                                        6
#define RSBTC_VAL_IQC_DVBT2_FFT_M32K                                        7
#define RSBTC_VAL_IQC_DVBT2_FFT_M32E                                        8

// rsbtc_rngDvbt2GuardInt
#define RSBTC_VAL_IQC_DVBT2_GUARD_G1_4                                      0
#define RSBTC_VAL_IQC_DVBT2_GUARD_G1_8                                      1
#define RSBTC_VAL_IQC_DVBT2_GUARD_G1_16                                     2
#define RSBTC_VAL_IQC_DVBT2_GUARD_G1_32                                     3
#define RSBTC_VAL_IQC_DVBT2_GUARD_G1128                                     4
#define RSBTC_VAL_IQC_DVBT2_GUARD_G19128                                    5
#define RSBTC_VAL_IQC_DVBT2_GUARD_G19256                                    6

// rsbtc_rngDvbt2PilotPatt
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP1                                       0
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP2                                       1
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP3                                       2
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP4                                       3
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP5                                       4
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP6                                       5
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP7                                       6
#define RSBTC_VAL_IQC_DVBT2_PILOT_PP8                                       7

// rsbtc_rngDvbt2TransSyst
#define RSBTC_VAL_IQC_DVBT2_TRANS_SISO                                      0
#define RSBTC_VAL_IQC_DVBT2_TRANS_MISO                                      1
#define RSBTC_VAL_IQC_DVBT2_TRANS_NONT2                                     2
#define RSBTC_VAL_IQC_DVBT2_TRANS_T2LS                                      3
#define RSBTC_VAL_IQC_DVBT2_TRANS_T2LM                                      4

// rsbtc_rngDvbt2PeakAverPow
#define RSBTC_VAL_IQC_DVBT2_PEAK_AVER_POW_OFF                               0
#define RSBTC_VAL_IQC_DVBT2_PEAK_AVER_POW_TR                                1

// rsbtc_rngDvbt2L1PostModul
#define RSBTC_VAL_IQC_DVBT2_L1_POST_T2                                      0
#define RSBTC_VAL_IQC_DVBT2_L1_POST_T4                                      1
#define RSBTC_VAL_IQC_DVBT2_L1_POST_T16                                     2
#define RSBTC_VAL_IQC_DVBT2_L1_POST_T64                                     3

// rsbtc_rngIqcIsdbtTmccNext
#define RSBTC_VAL_ISDBT_TMCC_UNUSED                                         0
#define RSBTC_VAL_ISDBT_TMCC_CURRENT                                        1

// rsbtc_rngIqcIsdbtTxParam
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_NORM                                    0
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N1                                      1
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N2                                      2
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N3                                      3
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N4                                      4
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N5                                      5
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N6                                      6
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N7                                      7
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N8                                      8
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N9                                      9
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N10                                     10
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N11                                     11
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N12                                     12
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N13                                     13
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N14                                     14
#define RSBTC_IQC_VAL_ISDBT_TXPARAM_N15                                     15

// rsbtc_rngIqcAtsmPackLength
#define RSBTC_VAL_ATSM_PACKED_LENGTH_INV                                    0
#define RSBTC_VAL_ATSM_PACKED_LENGTH_P188                                   1
#define RSBTC_VAL_ATSM_PACKED_LENGTH_P208                                   2

// rsbtc_rngIqcAtsmTestSig
#define RSBTC_VAL_ATSM_TEST_TTSP                                            0
#define RSBTC_VAL_ATSM_TEST_PBIN                                            1
#define RSBTC_VAL_ATSM_TEST_PBET                                            2
#define RSBTC_VAL_ATSM_TEST_PBEM                                            3

// rsbtc_rngIqcAtsmBury
#define RSBTC_VAL_IQC_ATSM_BURY_39DB                                        0
#define RSBTC_VAL_IQC_ATSM_BURY_36DB                                        1
#define RSBTC_VAL_IQC_ATSM_BURY_33DB                                        2
#define RSBTC_VAL_IQC_ATSM_BURY_30DB                                        3
#define RSBTC_VAL_IQC_ATSM_BURY_27DB                                        4
#define RSBTC_VAL_IQC_ATSM_BURY_24DB                                        5
#define RSBTC_VAL_IQC_ATSM_BURY_21DB                                        6

// rsbtc_rngTsgenStopData
#define RSBTC_VAL_MM_GENERATOR_STOP_DATA_NONE                               0
#define RSBTC_VAL_MM_GENERATOR_STOP_DATA_TTSP                               1

// rsbtc_rngTsgenTsPacket
#define RSBTC_VAL_MM_GENERATOR_TS_PACKET_H184                               0
#define RSBTC_VAL_MM_GENERATOR_TS_PACKET_S187                               1
#define RSBTC_VAL_MM_GENERATOR_TS_PACKET_H200                               2
#define RSBTC_VAL_MM_GENERATOR_TS_PACKET_S203                               3
#define RSBTC_VAL_MM_GENERATOR_TS_PACKET_H204                               4
#define RSBTC_VAL_MM_GENERATOR_TS_PACKET_S207                               5

// rsbtc_rngDvbt2L1Version
#define RSBTC_VAL_IQC_DVBT2_L1_VER_V111                                     0
#define RSBTC_VAL_IQC_DVBT2_L1_VER_V121                                     1
#define RSBTC_VAL_IQC_DVBT2_L1_VER_V131                                     2

// rsbtc_rngDvbt2MisoGroup
#define RSBTC_VAL_IQC_DVBT2_MISO_G1                                         1
#define RSBTC_VAL_IQC_DVBT2_MISO_G2                                         2

// rsbtc_rngNetworkMode
#define RSBTC_VAL_NETWORK_MODE_MFN                                          0
#define RSBTC_VAL_NETWORK_MODE_SFN                                          1

// rsbtc_rngDvbt2PlpModulType
#define RSBTC_VAL_DVBT2_PLP_MODUL_TYPE_CCM                                  0
#define RSBTC_VAL_DVBT2_PLP_MODUL_TYPE_ACM                                  1

// rsbtc_rngDvbt2IssyMode
#define RSBTC_VAL_DVBT2_MODE_ISSY_OFF                                       0
#define RSBTC_VAL_DVBT2_MODE_ISSY_SHORT                                     1
#define RSBTC_VAL_DVBT2_MODE_ISSY_LONG                                      2

// rsbtc_rngDvbc2InpFormat
#define RSBTC_VAL_DVBC2_INPUT_FORMAT_TS                                     0

// rsbtc_rngDvbc2PlpType
#define RSBTC_VAL_DVBC2_PLP_TYPE_NORMAL                                     0
#define RSBTC_VAL_DVBC2_PLP_TYPE_COMMON                                     1
#define RSBTC_VAL_DVBC2_PLP_TYPE_GROUPED                                    2

// rsbtc_rngDvbc2BbMode
#define RSBTC_VAL_DVBC2_BB_MODE_HEM                                         0
#define RSBTC_VAL_DVBC2_BB_MODE_NM                                          1

// rsbtc_rngDvbc2IssyMode
#define RSBTC_VAL_DVBC2_MODE_ISSY_OFF                                       0
#define RSBTC_VAL_DVBC2_MODE_ISSY_SHORT                                     1
#define RSBTC_VAL_DVBC2_MODE_ISSY_LONG                                      2

// rsbtc_rngDvbc2BicmFecFrame
#define RSBTC_VAL_DVBC2_BICM_FEC_NORMAL                                     0
#define RSBTC_VAL_DVBC2_BICM_FEC_SHORT                                      1

// rsbtc_rngDvbc2CodeRate
#define RSBTC_VAL_IQC_DVBC2_CRATE_R2_3                                      0
#define RSBTC_VAL_IQC_DVBC2_CRATE_R3_4                                      1
#define RSBTC_VAL_IQC_DVBC2_CRATE_R4_5                                      2
#define RSBTC_VAL_IQC_DVBC2_CRATE_R5_6                                      3
#define RSBTC_VAL_IQC_DVBC2_CRATE_R8_9                                      4
#define RSBTC_VAL_IQC_DVBC2_CRATE_R9_10                                     5

// rsbtc_rngDvbc2Constell
#define RSBTC_VAL_IQC_DVBC2_CONSTEL_Q16                                     0
#define RSBTC_VAL_IQC_DVBC2_CONSTEL_Q64                                     1
#define RSBTC_VAL_IQC_DVBC2_CONSTEL_Q256                                    2
#define RSBTC_VAL_IQC_DVBC2_CONSTEL_Q1024                                   3
#define RSBTC_VAL_IQC_DVBC2_CONSTEL_Q4096                                   4

// rsbtc_rngDvbc2DslTIDepth
#define RSBTC_VAL_IQC_DVBC2_DSLICE_TI_DEPTH_0                               0
#define RSBTC_VAL_IQC_DVBC2_DSLICE_TI_DEPTH_4                               1
#define RSBTC_VAL_IQC_DVBC2_DSLICE_TI_DEPTH_8                               2
#define RSBTC_VAL_IQC_DVBC2_DSLICE_TI_DEPTH_16                              3

// rsbtc_rngDvbc2DslType
#define RSBTC_VAL_IQC_DVBC2_DSLICE_TYPE_1                                   0
#define RSBTC_VAL_IQC_DVBC2_DSLICE_TYPE_2                                   1

// rsbtc_rngDvbc2DslFecHeaderType
#define RSBTC_VAL_DVBC2_DSLICE_FEC_HEADER_ROBUST                            0
#define RSBTC_VAL_DVBC2_DSLICE_FEC_HEADER_HEM                               1

// rsbtc_rngDvbc2ChBWidth
#define RSBTC_VAL_IQC_DVBC2_CHANNEL_BW_6                                    0
#define RSBTC_VAL_IQC_DVBC2_CHANNEL_BW_8                                    1

// rsbtc_rngDvbc2GuardInt
#define RSBTC_VAL_IQC_DVBC2_GUARD_1_64                                      0
#define RSBTC_VAL_IQC_DVBC2_GUARD_1_128                                     1

// rsbtc_rngDvbc2SystemInterleaver
#define RSBTC_VAL_DVBC2_SYSTEM_INTERLEAVER_OFF                              0
#define RSBTC_VAL_DVBC2_SYSTEM_INTERLEAVER_BEST                             1
#define RSBTC_VAL_DVBC2_SYSTEM_INTERLEAVER_FOUR                             2
#define RSBTC_VAL_DVBC2_SYSTEM_INTERLEAVER_EIGHT                            3

// rsbtc_rngIqcIsdbtChanBw
#define RSBTC_VAL_ISDBT_CHAN_BW6                                            0
#define RSBTC_VAL_ISDBT_CHAN_BW7                                            1
#define RSBTC_VAL_ISDBT_CHAN_BW8                                            2

// rsbtc_rngCodingControl
#define RSBTC_VAL_IQC_CONTROL_MODE_AUTO                                     0
#define RSBTC_VAL_IQC_CONTROL_MODE_MANUAL                                   1

// rsbtc_rngDVBT2SFNTimeStampMode
#define RSBTC_VAL_DVBT2_SFN_TIMESTAMP_MODE_REL                              0
#define RSBTC_VAL_DVBT2_SFN_TIMESTAMP_MODE_ABS                              1

// rsbtc_rngDVBT2T2SystemFutureExtensionFramesPayload
#define RSBTC_VAL_IQC_DVBT2_FEF_PAYLOAD_NULL                                0
#define RSBTC_VAL_IQC_DVBT2_FEF_PAYLOAD_NOISE                               1

// rsbtc_rngDVBT2T2SystemProfileMode
#define RSBTC_VAL_IQC_DVBT2_PROFILE_SINGLE                                  0
#define RSBTC_VAL_IQC_DVBT2_PROFILE_MULTI                                   1

// rsbtc_rngDVBC2FramingBasebandFilter
#define RSBTC_VAL_IQC_DVBC2_BB_FILTER_FIR                                   0
#define RSBTC_VAL_IQC_DVBC2_BB_FILTER_G1_4                                  1
#define RSBTC_VAL_IQC_DVBC2_BB_FILTER_G1_2                                  2
#define RSBTC_VAL_IQC_DVBC2_BB_FILTER_G3_4                                  3
#define RSBTC_VAL_IQC_DVBC2_BB_FILTER_G1_1                                  4

// rsbtc_rngDVBC2NotchType
#define RSBTC_VAL_DVBC2_NOTCH_TYPE_NARROW                                   0
#define RSBTC_VAL_DVBC2_NOTCH_TYPE_BROAD                                    1

// rsbtc_rngAnalogTVNICAMSource
#define RSBTC_VAL_ATV_NICAM_SOURCE_GENERATOR                                0
#define RSBTC_VAL_ATV_NICAM_SOURCE_PRBS                                     1
#define RSBTC_VAL_ATV_NICAM_SOURCE_AUDIO_PLAYER                             2

// rsbtc_rngAnalogTVNICAMSoundMode
#define RSBTC_VAL_ATV_NICAM_SOUND_MODE_MONO                                 0
#define RSBTC_VAL_ATV_NICAM_SOUND_MODE_STEREO                               1
#define RSBTC_VAL_ATV_NICAM_SOUND_MODE_DUAL                                 2

// rsbtc_rngAnalogTVSpecialNICAMRollOff
#define RSBTC_VAL_IQC_ATV_NICAM_ROLLOFF_R40                                 0
#define RSBTC_VAL_IQC_ATV_NICAM_ROLLOFF_R100                                1

// rsbtc_rngInputSignalFormat
#define RSBTC_VAL_DTV_SIGNAL_FORMAT_ASI                                     0
#define RSBTC_VAL_DTV_SIGNAL_FORMAT_SMPTE                                   1
#define RSBTC_VAL_DTV_SIGNAL_FORMAT_ETI                                     2

// rsbtc_rngISDBTPacketLength
#define RSBTC_VAL_DTV_PACKED_LENGTH_INV                                     0
#define RSBTC_VAL_DTV_PACKED_LENGTH_188                                     1
#define RSBTC_VAL_DTV_PACKED_LENGTH_204                                     2

// rsbtc_rng8VSBPacketLength
#define RSBTC_VAL_DTV_VSB_PACKED_LENGTH_INV                                 0
#define RSBTC_VAL_DTV_VSB_PACKED_LENGTH_188                                 1
#define RSBTC_VAL_DTV_VSB_PACKED_LENGTH_208                                 2

// rsbtc_rngRFSignalStepState
#define RSBTC_VAL_RF_STEP_INCREMENT                                         0
#define RSBTC_VAL_RF_STEP_DECIMAL                                           1

// rsbtc_rngRFSignalLocalOscillatorMode
#define RSBTC_VAL_LOCAL_OSCILLATOR_MODE_INTERNAL                            0
#define RSBTC_VAL_LOCAL_OSCILLATOR_MODE_EXTERNAL                            1
#define RSBTC_VAL_LOCAL_OSCILLATOR_MODE_COUPLE                              2
#define RSBTC_VAL_LOCAL_OSCILLATOR_MODE_ECOUPLE                             3

// rsbtc_rngRFSignalLevelSettingsMode
#define RSBTC_VAL_RF_LEVEL_MODE_NORMAL                                      0
#define RSBTC_VAL_RF_LEVEL_MODE_LOW_NOISE                                   1
#define RSBTC_VAL_RF_LEVEL_MODE_LOW_DISTORTION                              2

// rsbtc_rngRFSignalLevelSettingCharacteristic
#define RSBTC_VAL_RF_LEVEL_CHARACTERISTIC_AUTO                              0
#define RSBTC_VAL_RF_LEVEL_CHARACTERISTIC_INTERRUPTED                       2
#define RSBTC_VAL_RF_LEVEL_CHARACTERISTIC_STRICTLY_MONO                     3
#define RSBTC_VAL_RF_LEVEL_CHARACTERISTIC_CONST_VSWR                        4
#define RSBTC_VAL_RF_LEVEL_CHARACTERISTIC_CONST_PHASE                       5

// rsbtc_rngRFSignalLevelALCDetectorSensitivity
#define RSBTC_VAL_RF_ALC_SENSITIVITY_AUTO                                   0
#define RSBTC_VAL_RF_ALC_SENSITIVITY_LOW                                    1
#define RSBTC_VAL_RF_ALC_SENSITIVITY_MED                                    2
#define RSBTC_VAL_RF_ALC_SENSITIVITY_HIGH                                   3

// rsbtc_rngRFSignalLevelALCSampleLevel
#define RSBTC_VAL_RF_ALC_SAMPLE_FULL                                        0
#define RSBTC_VAL_RF_ALC_SAMPLE_ATTENUATED                                  1
#define RSBTC_VAL_RF_ALC_SAMPLE_MINIMUM                                     2

// rsbtc_rngRFSignalSweepShape
#define RSBTC_VAL_RF_SWEEP_SHAPE_TRIANGLE                                   0
#define RSBTC_VAL_RF_SWEEP_SHAPE_SAWTOOTH                                   1

// rsbtc_rngRFSignalModulationSignalSource_001
#define RSBTC_VAL_RF_MODULATION_SIGNAL_INTERNAL_BASEBAND                    0
#define RSBTC_VAL_RF_MODULATION_SIGNAL_INTERNAL_ANALOG_WIDEBAND             1
#define RSBTC_VAL_RF_MODULATION_SIGNAL_INTERNAL_DIFF_ANALOG                 2

// rsbtc_rngArbitraryWaveformGenerationUsage
#define RSBTC_VAL_ARB_GENERATOR_USAGE_SIGNAL                                0
#define RSBTC_VAL_ARB_GENERATOR_USAGE_INTERFERER                            1

// rsbtc_rngDVBT2InputSignalT2MIPacketTimingMeasMode
#define RSBTC_VAL_DVBT2_TIMING_MODE_ABSOLUTE                                0
#define RSBTC_VAL_DVBT2_TIMING_MODE_DELTA                                   1

// rsbtc_rngDVBTCodingSymbolInterleaver
#define RSBTC_VAL_DVBT_INTERLEAVER_NATIVE                                   0
#define RSBTC_VAL_DVBT_INTERLEAVER_INDEPTH                                  1

// rsbtc_rngFadingSimulatorSignalRouting
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_A                                   0
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_AD                                  1
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_AB                                  2
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_ABX                                 3
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_AXB                                 4
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_ADBD                                5
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_ADBN                                6
#define RSBTC_VAL_FADING_SIGNAL_ROUTING_ANBD                                7

// rsbtc_rngFadingFrequencyHoppingMode
#define RSBTC_VAL_FADING_FREQ_HOPPING_OFF                                   0
#define RSBTC_VAL_FADING_FREQ_HOPPING_IN_BAND                               1
#define RSBTC_VAL_FADING_FREQ_HOPPING_OUT_BAND                              2

// rsbtc_rngNoiseImpulsiveFrameDuration
#define RSBTC_VAL_NOISE_FRAME_DURATION_10                                   0
#define RSBTC_VAL_NOISE_FRAME_DURATION_100                                  1
#define RSBTC_VAL_NOISE_FRAME_DURATION_1000                                 2

// rsbtc_rngTSGENPlayerTSSerialOutput
#define RSBTC_VAL_TSGEN_ASIBURST                                            0
#define RSBTC_VAL_TSGEN_ASICONT                                             1

// rsbtc_rngISDBTmmSuperSegmentsType
#define RSBTC_VAL_TMM_SUPER_SEGMENT_A                                       0
#define RSBTC_VAL_TMM_SUPER_SEGMENT_B                                       1

// rsbtc_rngISDBTmmCodingISDBTMode
#define RSBTC_VAL_TMM_ISDBT_MODE_1                                          0
#define RSBTC_VAL_TMM_ISDBT_MODE_2                                          1
#define RSBTC_VAL_TMM_ISDBT_MODE_3                                          2

// rsbtc_rngIqcCmmbBw
#define RSBTC_VAL_CMMB_BW_2                                                 0
#define RSBTC_VAL_CMMB_BW_8                                                 1

// rsbtc_rngIqcCmmbOdfm
#define RSBTC_VAL_CMMB_OFDM_M1K                                             0
#define RSBTC_VAL_CMMB_OFDM_M4K                                             1

// rsbtc_rngCMMBInputSignalTestSignal
#define RSBTC_VAL_IQC_CMMB_TST_SIG_CLCHT                                    0
#define RSBTC_VAL_IQC_CMMB_TST_SIG_TONLY                                    1

// rsbtc_rngTDMBETIInput
#define RSBTC_VAL_TDMB_ETI_INPUT_INV                                        0
#define RSBTC_VAL_TDMB_ETI_INPUT_ENI                                        1
#define RSBTC_VAL_TDMB_ETI_INPUT_E559                                       2
#define RSBTC_VAL_TDMB_ETI_INPUT_E537                                       3

// rsbtc_rngTDMBProtProfile
#define RSBTC_VAL_TDMB_PROFILE_UEP                                          0
#define RSBTC_VAL_TDMB_PROFILE_EEP                                          1

// rsbtc_rngTDMBProtLevel
#define RSBTC_VAL_TDMB_LEVEL_UP1                                            0
#define RSBTC_VAL_TDMB_LEVEL_UP2                                            1
#define RSBTC_VAL_TDMB_LEVEL_UP3                                            2
#define RSBTC_VAL_TDMB_LEVEL_UP4                                            3
#define RSBTC_VAL_TDMB_LEVEL_UP5                                            4
#define RSBTC_VAL_TDMB_LEVEL_EP1A                                           5
#define RSBTC_VAL_TDMB_LEVEL_EP2A                                           6
#define RSBTC_VAL_TDMB_LEVEL_EP3A                                           7
#define RSBTC_VAL_TDMB_LEVEL_EP4A                                           8
#define RSBTC_VAL_TDMB_LEVEL_EP1B                                           9
#define RSBTC_VAL_TDMB_LEVEL_EP2B                                           10
#define RSBTC_VAL_TDMB_LEVEL_EP3B                                           11
#define RSBTC_VAL_TDMB_LEVEL_EP4B                                           12

// rsbtc_rngTDMBTransmissionMode
#define RSBTC_VAL_TDMB_TRAN_MODE_MID                                        0
#define RSBTC_VAL_TDMB_TRAN_MODE_MANUAL                                     1

// rsbtc_rngTDMBPRBS
#define RSBTC_VAL_TDMB_PRBS_P23_1                                           0
#define RSBTC_VAL_TDMB_PRBS_P15_1                                           1
#define RSBTC_VAL_TDMB_PRBS_P20_1                                           2
#define RSBTC_VAL_TDMB_PRBS_ZERO                                            3

// rsbtc_rngDVBS2InputSignalCMMode
#define RSBTC_VAL_DVBS2_CM_MODE_ACM                                         0
#define RSBTC_VAL_DVBS2_CM_MODE_CCM                                         1
#define RSBTC_VAL_DVBS2_CM_MODE_VCM                                         2

// rsbtc_rngDVBT2SFN1PPSRouting
#define RSBTC_VAL_DVBT2_SFN_ROUTING_IN_EXT                                  0
#define RSBTC_VAL_DVBT2_SFN_ROUTING_IN_INT                                  1
#define RSBTC_VAL_DVBT2_SFN_ROUTING_OUT_INT                                 2

// rsbtc_rngBERInputFormat
#define RSBTC_VAL_BER_INPUT_FORMAT_ASI                                      0
#define RSBTC_VAL_BER_INPUT_FORMAT_SMPTE                                    1

// rsbtc_rngRFSignalIQDigitalOutputSampleRateSource
#define RSBTC_VAL_RF_MOD_IQ_DIGITAL_USER                                    0
#define RSBTC_VAL_RF_MOD_IQ_DIGITAL_OUT                                     1

// rsbtc_rngRFSignalIQDigitalOutputSetLevelVia
#define RSBTC_VAL_RF_MOD_IQ_DIGITAL_PEAK_LEVEL                              0
#define RSBTC_VAL_RF_MOD_IQ_DIGITAL_LEVEL                                   1

// rsbtc_rngRFSignalIQDigitalOutputUserVariationActive
#define RSBTC_VAL_RF_MOD_IQ_DIGITAL_USER_VARIATION_OFF                      0
#define RSBTC_VAL_RF_MOD_IQ_DIGITAL_USER_VARIATION_ON                       1

// rsbtc_rngRFSignalIQAnalogOutputType
#define RSBTC_VAL_RF_MOD_IQ_ANALOG_OUT_TYPE_SINGLE                          0
#define RSBTC_VAL_RF_MOD_IQ_ANALOG_OUT_TYPE_DIFF                            1

// rsbtc_rngRFSignalIQAnalogOutputMode
#define RSBTC_VAL_RF_MOD_IQ_ANALOG_OUT_MODE_FIXED                           0
#define RSBTC_VAL_RF_MOD_IQ_ANALOG_OUT_MODE_VARIABLE                        1

// rsbtc_rngAudioBCDigitalIQInputSampleRateSource
#define RSBTC_VAL_DIG_IQ_SAMPLE_RATE_USER                                   0
#define RSBTC_VAL_DIG_IQ_SAMPLE_RATE_DIGITAL_IN                             1

// rsbtc_rngISDBTmmInputSignalTestSignal
#define RSBTC_VAL_TMM_TEST_TTSP                                             0
#define RSBTC_VAL_TMM_TEST_PBEC                                             1
#define RSBTC_VAL_TMM_TEST_PAFC                                             2

// rsbtc_rngPowerMeterUnit
#define RSBTC_VAL_POWER_METER_UNIT_DBM                                      0
#define RSBTC_VAL_POWER_METER_UNIT_DBUV                                     1
#define RSBTC_VAL_POWER_METER_UNIT_WATT                                     2

// rsbtc_rngPowerMeterSource
#define RSBTC_VAL_PWR_METER_SOURCE_TX                                       0
#define RSBTC_VAL_PWR_METER_SOURCE_TXA                                      1
#define RSBTC_VAL_PWR_METER_SOURCE_TXB                                      2
#define RSBTC_VAL_PWR_METER_SOURCE_USER                                     3

// rsbtc_rngPowerMeterRelPowerState
#define RSBTC_VAL_PWR_METER_REL_STAT_OFF                                    0
#define RSBTC_VAL_PWR_METER_REL_STAT_DB                                     1
#define RSBTC_VAL_PWR_METER_REL_STAT_PCT                                    2

// rsbtc_rngPowerMeterAverageFilterType
#define RSBTC_VAL_PWR_METER_FILTER_AUTO                                     0
#define RSBTC_VAL_PWR_METER_FILTER_USER                                     1

// rsbtc_rngPowerMeterFilterLength
#define RSBTC_VAL_PWR_METER_FILTER_LEN_1                                    0
#define RSBTC_VAL_PWR_METER_FILTER_LEN_2                                    1
#define RSBTC_VAL_PWR_METER_FILTER_LEN_4                                    2
#define RSBTC_VAL_PWR_METER_FILTER_LEN_8                                    3
#define RSBTC_VAL_PWR_METER_FILTER_LEN_16                                   4
#define RSBTC_VAL_PWR_METER_FILTER_LEN_32                                   5
#define RSBTC_VAL_PWR_METER_FILTER_LEN_64                                   6
#define RSBTC_VAL_PWR_METER_FILTER_LEN_128                                  7

// rsbtc_rngReadCMMBStreamFormat
#define RSBTC_VAL_CMMB_STREAM_FORMAT_PMS                                    0
#define RSBTC_VAL_CMMB_STREAM_FORMAT_MFS                                    1

// rsbtc_rngSFN1PPSRouting
#define RSBTC_VAL_SFN_ROUTING_IN_EXT                                        0
#define RSBTC_VAL_SFN_ROUTING_IN_INT                                        1
#define RSBTC_VAL_SFN_ROUTING_OUT_INT                                       2

// rsbtc_rngIPInputType
#define RSBTC_VAL_IP_INPUT_TYPE_UNICAST                                     0
#define RSBTC_VAL_IP_INPUT_TYPE_MULTICAST                                   1

// rsbtc_rngISDBTEEWSignalType
#define RSBTC_VAL_DTV_EEW_WWA                                               0
#define RSBTC_VAL_DTV_EEW_WWOA                                              1
#define RSBTC_VAL_DTV_EEW_TWA                                               2
#define RSBTC_VAL_DTV_EEW_TWOA                                              3

// rsbtc_rngISDBTEEWEpicenterInformationType
#define RSBTC_VAL_DTV_EEW_EPICENTER_ISSUED                                  0
#define RSBTC_VAL_DTV_EEW_EPICENTER_CANCELED                                1

// rsbtc_rngMMGeneratorTransportStreamPlayerType
#define RSBTC_VAL_MM_GENERATOR_TSTREAM_TYPE_ES                              0
#define RSBTC_VAL_MM_GENERATOR_TSTREAM_TYPE_TS                              1

// rsbtc_rngBasebandIQDigitalOutputSampleRateSource
#define RSBTC_VAL_BB_IQ_DIGITAL_USER                                        0
#define RSBTC_VAL_BB_IQ_DIGITAL_OUT                                         1
#define RSBTC_VAL_BB_IQ_DIGITAL_SIGN                                        2

// rsbtc_rngBasebandDigitalIQOutputSetLevelVia
#define RSBTC_VAL_BASEBAND_IQ_DIGITAL_PEAK_LEVEL                            0
#define RSBTC_VAL_BASEBAND_IQ_DIGITAL_LEVEL                                 1

// rsbtc_rngMMGMode
#define RSBTC_VAL_MMG_GATEWAY_MODE_OFF                                      0
#define RSBTC_VAL_MMG_GATEWAY_MODE_T2MI                                     1

// rsbtc_rngMMGDVBT2MIUseCase
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_SPNORMAL                             0
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_SPEXPERT                             1
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_MPSNORMAL                            2
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_MPSEXPERT                            3
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_MPDNORMAL                            4
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_MPDEXPERT                            5
#define RSBTC_VAL_MMG_DVBT2MI_USE_CASE_MPGW                                 6

// rsbtc_rngMMGDVBT2MIInputType
#define RSBTC_VAL_MMG_DVBT2MI_INPUT_TS                                      0
#define RSBTC_VAL_MMG_DVBT2MI_INPUT_PRBS                                    1
#define RSBTC_VAL_MMG_DVBT2MI_INPUT_PRTS                                    2

// rsbtc_rngMMGDVBT2MIInputSelect
#define RSBTC_VAL_MMG_DVBT2MI_INPUT_SELECT_TSM1                             0
#define RSBTC_VAL_MMG_DVBT2MI_INPUT_SELECT_P1                               1
#define RSBTC_VAL_MMG_DVBT2MI_INPUT_SELECT_P2                               2

// rsbtc_rngMMGDVBT2MIPLPPayloadType
#define RSBTC_VAL_MMG_DVBT2MI_PLP_PAYLOAD_TS                                0

// rsbtc_rngMMGDVBT2MIPLPData
#define RSBTC_VAL_MMG_DVBT2MI_PLP_DATA_COMMON                               0
#define RSBTC_VAL_MMG_DVBT2MI_PLP_DATA_SERVICE                              1
#define RSBTC_VAL_MMG_DVBT2MI_PLP_DATA_FULL                                 2

// rsbtc_rngMMGDVBT2MIBufferSizeUnit
#define RSBTC_VAL_MMG_DVBT2MI_BSIZE_UNIT_BITS                               0
#define RSBTC_VAL_MMG_DVBT2MI_BSIZE_UNIT_KB                                 1
#define RSBTC_VAL_MMG_DVBT2MI_BSIZE_UNIT_MB                                 2
#define RSBTC_VAL_MMG_DVBT2MI_BSIZE_UNIT_8KB                                3

// rsbtc_rngMMGDVBT2MIT2TransSyst
#define RSBTC_VAL_MMG_DVBT2MI_TRANS_SISO                                    0
#define RSBTC_VAL_MMG_DVBT2MI_TRANS_MISO                                    1
#define RSBTC_VAL_MMG_DVBT2MI_TRANS_T2LS                                    2
#define RSBTC_VAL_MMG_DVBT2MI_TRANS_T2LM                                    3

// rsbtc_rngMMGDVBT2MIT2SystemScheduling
#define RSBTC_VAL_MMG_DVBT2MI_SHCEDULING_STATIC                             0
#define RSBTC_VAL_MMG_DVBT2MI_SHCEDULING_DYNAMIC                            1

// rsbtc_rngMMGDVBT2MIOutputT2TimestampMode
#define RSBTC_VAL_MMG_DVBT2MI_TSTAMP_MODE_NULL                              0
#define RSBTC_VAL_MMG_DVBT2MI_TSTAMP_MODE_REL                               1

// rsbtc_rngMMGDVBT2MIOutputT2MITSType
#define RSBTC_VAL_MMG_DVBT2MI_OUT_TS_TYPE_SINGLE                            0
#define RSBTC_VAL_MMG_DVBT2MI_OUT_TS_TYPE_FULL                              1

// rsbtc_rngPhaseNoiseShape
#define RSBTC_VAL_PHASE_NOISE_SHAPE_SHA1                                    1
#define RSBTC_VAL_PHASE_NOISE_SHAPE_SHA2                                    2
#define RSBTC_VAL_PHASE_NOISE_SHAPE_SHA3                                    3

// rsbtc_rngDmSigPolarity
#define RSBTC_VAL_BB_POLARITY_NORM                                          0
#define RSBTC_VAL_BB_POLARITY_INV                                           1

// rsbtc_rngDvbt2MIChBWidth
#define RSBTC_VAL_IQC_DVBT2MI_CH_BWIDTH_BW_2                                0
#define RSBTC_VAL_IQC_DVBT2MI_CH_BWIDTH_BW_5                                1
#define RSBTC_VAL_IQC_DVBT2MI_CH_BWIDTH_BW_6                                2
#define RSBTC_VAL_IQC_DVBT2MI_CH_BWIDTH_BW_7                                3
#define RSBTC_VAL_IQC_DVBT2MI_CH_BWIDTH_BW_8                                4
#define RSBTC_VAL_IQC_DVBT2MI_CH_BWIDTH_BW_10                               5

// rsbtc_rngATSC3InputSignalPLPTestSignal
#define RSBTC_VAL_ATSC3_TEST_TTSP                                           0
#define RSBTC_VAL_ATSC3_TEST_TIPP                                           1

// rsbtc_rngATSC3InputSignalInputType
#define RSBTC_VAL_ATSC3_INPUT_TS                                            0
#define RSBTC_VAL_ATSC3_INPUT_IP                                            1

// rsbtc_rngATSC3InputSignalInputProtocol
#define RSBTC_VAL_ATSC3_INPUT_PROTOCOL_AUTO                                 0
#define RSBTC_VAL_ATSC3_INPUT_PROTOCOL_UDP                                  1
#define RSBTC_VAL_ATSC3_INPUT_PROTOCOL_RTP                                  2

// rsbtc_rngATSC3ConfiguringTestIPPacket
#define RSBTC_VAL_ATSC3_INPUT_TEST_IP_HUDP                                  0

// rsbtc_rngATSC3BICMLowLevelSignaling
#define RSBTC_VAL_ATSC3_BICM_LLS_ABSENT                                     0
#define RSBTC_VAL_ATSC3_BICM_LLS_PRESENT                                    1

// rsbtc_rngATSC3BICMLayer
#define RSBTC_VAL_ATSC3_BICM_LAYER_CORE                                     0
#define RSBTC_VAL_ATSC3_BICM_LAYER_ENHANCED                                 1

// rsbtc_rngATSC3BICMLDMInjectionLevel
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L00                                      0
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L05                                      1
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L10                                      2
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L15                                      3
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L20                                      4
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L25                                      5
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L30                                      6
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L35                                      7
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L40                                      8
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L45                                      9
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L50                                      10
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L60                                      11
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L70                                      12
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L80                                      13
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L90                                      14
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L100                                     15
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L110                                     16
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L120                                     17
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L130                                     18
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L140                                     19
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L150                                     20
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L160                                     21
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L170                                     22
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L180                                     23
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L190                                     24
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L200                                     25
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L210                                     26
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L220                                     27
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L230                                     28
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L240                                     29
#define RSBTC_VAL_ATSC3_BICM_LEVEL_L250                                     30

// rsbtc_rngATSC3BICMFECType
#define RSBTC_VAL_ATSC3_BICM_FEC_TYPE_B16K                                  0
#define RSBTC_VAL_ATSC3_BICM_FEC_TYPE_B64K                                  1
#define RSBTC_VAL_ATSC3_BICM_FEC_TYPE_C16K                                  2
#define RSBTC_VAL_ATSC3_BICM_FEC_TYPE_C64K                                  3
#define RSBTC_VAL_ATSC3_BICM_FEC_TYPE_O16K                                  4
#define RSBTC_VAL_ATSC3_BICM_FEC_TYPE_O64K                                  5

// rsbtc_rngATSC3BICMCodeRate
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R2_15                                0
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R3_15                                1
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R4_15                                2
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R5_15                                3
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R6_15                                4
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R7_15                                5
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R8_15                                6
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R9_15                                7
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R10_15                               8
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R11_15                               9
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R12_15                               10
#define RSBTC_VAL_ATSC3_BICM_CODE_RATE_R13_15                               11

// rsbtc_rngDTVConstellation
#define RSBTC_VAL_DTV_CONSTELLATION_T4                                      0
#define RSBTC_VAL_DTV_CONSTELLATION_T16                                     1
#define RSBTC_VAL_DTV_CONSTELLATION_T64                                     2
#define RSBTC_VAL_DTV_CONSTELLATION_T256                                    3
#define RSBTC_VAL_DTV_CONSTELLATION_T1024                                   4
#define RSBTC_VAL_DTV_CONSTELLATION_T4096                                   5

// rsbtc_rngATSC3BICMType
#define RSBTC_VAL_ATSC3_BICM_TYPE_NONDISPERSED                              0
#define RSBTC_VAL_ATSC3_BICM_TYPE_DISPERSED                                 1

// rsbtc_rngATSC3BICMTimeInterleaverMode
#define RSBTC_VAL_ATSC3_BICM_TIL_OFF                                        0
#define RSBTC_VAL_ATSC3_BICM_TIL_CTI                                        1
#define RSBTC_VAL_ATSC3_BICM_TIL_HTI                                        2

// rsbtc_rngATSC3BICMTimeInterleaverDepth
#define RSBTC_VAL_ATSC3_BICM_TIL_DEPTH_D512                                 0
#define RSBTC_VAL_ATSC3_BICM_TIL_DEPTH_D724                                 1
#define RSBTC_VAL_ATSC3_BICM_TIL_DEPTH_D887                                 2
#define RSBTC_VAL_ATSC3_BICM_TIL_DEPTH_D1024                                3
#define RSBTC_VAL_ATSC3_BICM_TIL_DEPTH_D1254                                4
#define RSBTC_VAL_ATSC3_BICM_TIL_DEPTH_D1448                                5

// rsbtc_rngATSC3SChannelBandwidth
#define RSBTC_VAL_ATSC3_CHANNEL_BW_6                                        0
#define RSBTC_VAL_ATSC3_CHANNEL_BW_7                                        1
#define RSBTC_VAL_ATSC3_CHANNEL_BW_8                                        2

// rsbtc_rngATSC3SubframingOFDMMISO
#define RSBTC_VAL_ATSC3_SUBFRAME_MISO_OFF                                   0
#define RSBTC_VAL_ATSC3_SUBFRAME_MISO_C64                                   1
#define RSBTC_VAL_ATSC3_SUBFRAME_MISO_C256                                  2

// rsbtc_rngATSC3SubframingOFDMFFTSize
#define RSBTC_VAL_ATSC3_SUBFRAME_FFT_M8K                                    0
#define RSBTC_VAL_ATSC3_SUBFRAME_FFT_M16K                                   1
#define RSBTC_VAL_ATSC3_SUBFRAME_FFT_M32K                                   2

// rsbtc_rngATSC3GuardInterval
#define RSBTC_VAL_ATSC3_GUARD_INT_192                                       0
#define RSBTC_VAL_ATSC3_GUARD_INT_384                                       1
#define RSBTC_VAL_ATSC3_GUARD_INT_512                                       2
#define RSBTC_VAL_ATSC3_GUARD_INT_768                                       3
#define RSBTC_VAL_ATSC3_GUARD_INT_1024                                      4
#define RSBTC_VAL_ATSC3_GUARD_INT_1536                                      5
#define RSBTC_VAL_ATSC3_GUARD_INT_2048                                      6
#define RSBTC_VAL_ATSC3_GUARD_INT_2432                                      7
#define RSBTC_VAL_ATSC3_GUARD_INT_3072                                      8
#define RSBTC_VAL_ATSC3_GUARD_INT_3648                                      9
#define RSBTC_VAL_ATSC3_GUARD_INT_4096                                      10
#define RSBTC_VAL_ATSC3_GUARD_INT_4864                                      11

// rsbtc_rngATSC3SubframingOFDMSISOPilotPattern
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_3_2                                   0
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_3_4                                   1
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_4_2                                   2
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_4_4                                   3
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_6_2                                   4
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_6_4                                   5
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_8_2                                   6
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_8_4                                   7
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_12_2                                  8
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_12_4                                  9
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_16_2                                  10
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_16_4                                  11
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_24_2                                  12
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_24_4                                  13
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_32_2                                  14
#define RSBTC_VAL_ATSC3_PILOT_PATTERN_32_4                                  15

// rsbtc_rngATSC3SystemNetworkMode
#define RSBTC_VAL_ATSC3_SYSTEM_NETWORK_MODE_MFN                             0

// rsbtc_rngATSC3AbsentPresent
#define RSBTC_VAL_ATSC3_ABSENT                                              0
#define RSBTC_VAL_ATSC3_PRESENT                                             1

// rsbtc_rngATSC3SystemNetworkTimeInfo
#define RSBTC_VAL_ATSC3_SYSTEM_TIMING_OFF                                   0
#define RSBTC_VAL_ATSC3_SYSTEM_TIMING_MSEC                                  1
#define RSBTC_VAL_ATSC3_SYSTEM_TIMING_USEC                                  2
#define RSBTC_VAL_ATSC3_SYSTEM_TIMING_NSEC                                  3

// rsbtc_rngATSC3SystemNetworkTxIDMode
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_OFF                                     0
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_MANUAL                                  1
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_AUTO                                    2

// rsbtc_rngATSC3SystemNetworkTxIDInjectionLevel
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_OFF                               0
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L450                              1
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L420                              2
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L390                              3
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L360                              4
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L330                              5
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L300                              6
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L270                              7
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L240                              8
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L210                              9
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L180                              10
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L150                              11
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L120                              12
#define RSBTC_VAL_ATSC3_SYSTEM_TXID_LEVEL_L90                               13

// rsbtc_rngATSC3SystemFramePeakToAveragePowerRatio
#define RSBTC_VAL_ATSC3_SYSTEM_PAPR_OFF                                     0
#define RSBTC_VAL_ATSC3_SYSTEM_PAPR_TR                                      1

// rsbtc_rngATSC3SystemFrameMode
#define RSBTC_VAL_ATSC3_SYSTEM_FRAME_MODE_TIME                              0
#define RSBTC_VAL_ATSC3_SYSTEM_FRAME_MODE_SYMBOL                            1

// rsbtc_rngATSC3SystemL1FECType
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD1                             1
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD2                             2
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD3                             3
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD4                             4
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD5                             5
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD6                             6
#define RSBTC_VAL_ATSC3_SYSTEM_L1_FEC_TYPE_MOD7                             7

// rsbtc_rngATSC3SystemReducedCarrierModePreamble
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D3                            0
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D4                            1
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D6                            2
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D8                            3
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D12                           4
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D16                           5
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D24                           6
#define RSBTC_VAL_ATSC3_SYSTEM_PILOT_PREAMBLE_D32                           7

// rsbtc_rngATSC3SystemL1DetailAdditionalParityMode
#define RSBTC_VAL_ATSC3_SYSTEM_L1_PARITY_OFF                                0
#define RSBTC_VAL_ATSC3_SYSTEM_L1_PARITY_K1                                 1
#define RSBTC_VAL_ATSC3_SYSTEM_L1_PARITY_K2                                 2

// rsbtc_rngATSC3BootstrapEmergencyAlertSignaling
#define RSBTC_VAL_ATSC3_BOOTSTRAP_ALERT_NOEMERGENCY                         0
#define RSBTC_VAL_ATSC3_BOOTSTRAP_ALERT_SET1                                1
#define RSBTC_VAL_ATSC3_BOOTSTRAP_ALERT_SET2                                2
#define RSBTC_VAL_ATSC3_BOOTSTRAP_ALERT_SET3                                3

// rsbtc_rngATSC3FrameInfoBootstrapMinTimetoNext
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_NOTAPPL                          0
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N50                              1
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N100                             2
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N150                             3
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N200                             4
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N250                             5
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N300                             6
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N350                             7
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N400                             8
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N500                             9
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N600                             10
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N700                             11
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N800                             12
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N900                             13
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1000                            14
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1100                            15
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1200                            16
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1300                            17
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1400                            18
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1500                            19
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1700                            20
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N1900                            21
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N2100                            22
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N2300                            23
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N2500                            24
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N2700                            25
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N2900                            26
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N3300                            27
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N3700                            28
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N4100                            29
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N4500                            30
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N4900                            31
#define RSBTC_VAL_ATSC3_BOOTSTRAP_MIN_TIME_N5300                            32

// rsbtc_rngATSC3FrameInfoBootstrapSystemBandwidth
#define RSBTC_VAL_ATSC3_BOOTSTRAP_BW_BW_6                                   0
#define RSBTC_VAL_ATSC3_BOOTSTRAP_BW_BW_7                                   1
#define RSBTC_VAL_ATSC3_BOOTSTRAP_BW_BW_8                                   2
#define RSBTC_VAL_ATSC3_BOOTSTRAP_BW_BW8G                                   3

// rsbtc_rngATSC3FrameInfoBootstrapFFTSize
#define RSBTC_VAL_ATSC3_BOOTSTRAP_FFT_M8K                                   0
#define RSBTC_VAL_ATSC3_BOOTSTRAP_FFT_M16K                                  1
#define RSBTC_VAL_ATSC3_BOOTSTRAP_FFT_M32K                                  2

// rsbtc_rngConnectionState
#define RSBTC_VAL_DISCONNECT                                                0
#define RSBTC_VAL_CONNECT                                                   1
/* END GENERATE */



/*****************************************************************************
 *- Hidden Attribute Declarations -------------------------------------------*
 *****************************************************************************/

#define RSBTC_ATTR_OPC_TIMEOUT               (RS_ATTR_OPC_TIMEOUT)           /* ViInt32 */
#define RSBTC_ATTR_IO_SESSION                (RS_ATTR_IO_SESSION)            /* ViSession */
#define RSBTC_ATTR_OPC_CALLBACK              (RS_ATTR_OPC_CALLBACK)          /* ViAddr */
#define RSBTC_ATTR_CHECK_STATUS_CALLBACK     (RS_ATTR_CHECK_STATUS_CALLBACK) /* ViAddr */

/****************************************************************************
 *------------------------ Attribute Value Defines -------------------------*
 ****************************************************************************/

    /*- Instrument specific attribute value definitions --------------------*/

// BTC:

#define RSBTC_VAL_AUDIO_BC_SOURCE_MONO   0
#define RSBTC_VAL_AUDIO_BC_SOURCE_LEFT   1
#define RSBTC_VAL_AUDIO_BC_SOURCE_RIGHT  2



#define RSBTC_VAL_FADING_PATH_LOSS        0
#define RSBTC_VAL_FADING_BASIC_DELAY      1
#define RSBTC_VAL_FADING_ADD_DELAY        2
#define RSBTC_VAL_FADING_POWER_RATIO      3
#define RSBTC_VAL_FADING_CONST_PHASE      4
#define RSBTC_VAL_FADING_SPEED            5
#define RSBTC_VAL_FADING_FREQ_RATIO       6
#define RSBTC_VAL_FADING_RES_DOPPLER_SHIFT 7
#define RSBTC_VAL_FADING_CORRELATION_COEFFICIENT  8
#define RSBTC_VAL_FADING_CORRELATION_PHASE        9
#define RSBTC_VAL_FADING_LOGNORMAL_LOCAL_CONST    10
#define RSBTC_VAL_FADING_LOGNORMAL_STD_DEVIATION  11
#define RSBTC_VAL_FADING_FREQ_SHIFT               12
#define RSBTC_VAL_FADING_FREQ_SPREAD              13



#define RSBTC_VAL_FADING_RES_DELAY 0
#define RSBTC_VAL_FADING_RES_DOPPLER_SHIFT 7

#define RSBTC_VAL_FADING_CORRELATION_PATH 0
#define RSBTC_VAL_FADING_LOGNORMAL_STATE  1

#define RSBTC_VAL_SAVE                          0
#define RSBTC_VAL_RECALL                        1
#define RSBTC_VAL_DELETE                        2

#define RSBTC_PATH_A 	1
#define RSBTC_PATH_B 	2

#define RSBTC_VAL_REQ_SOURCE_DRATE  0
#define RSBTC_VAL_MAX_USEFUL_DRATE 	1
#define RSBTC_VAL_MEAS_DRATE       	2
#define RSBTC_VAL_USEFUL_DRATE    	3

#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMIN1 	0
#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMIN2 	1
#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMIN3 	2
#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMAX1 	3
#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMAX2 	4
#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMAX3 	5
#define RSBTC_VAL_DVBT2_MI_PACKET_TIMING_TMAX4 	6

#define RSBTC_VAL_DVBT2_SFN_MAX_DEVIATION_TIME 0
#define RSBTC_VAL_DVBT2_SFN_STATIC_DELAY       1
#define RSBTC_VAL_DVBT2_SFN_DYNAMIC_DELAY      2
#define RSBTC_VAL_DVBT2_SFN_TOTAL_DELAY        3
#define RSBTC_VAL_DVBT2_SFN_PROCESS_DELAY      4

#define RSBTC_VAL_PRIORITY_HIGH     0
#define RSBTC_VAL_PRIORITY_LOW      1

#define RSBTC_VAL_ATV_AUDIO_GEN_AF1 0
#define RSBTC_VAL_ATV_AUDIO_GEN_AF2 1

#define RSBTC_VAL_ATV_AUDIO_SOUND_FREQ_MONO  0
#define RSBTC_VAL_ATV_AUDIO_SOUND_FREQ_LEFT  1
#define RSBTC_VAL_ATV_AUDIO_SOUND_FREQ_RIGHT 2
#define RSBTC_VAL_ATV_AUDIO_SOUND_FREQ_CH1   3
#define RSBTC_VAL_ATV_AUDIO_SOUND_FREQ_CH2   4

#define RSBTC_VAL_ATV_AUDIO_SOUND_LEVEL_MONO  0
#define RSBTC_VAL_ATV_AUDIO_SOUND_LEVEL_LEFT  1
#define RSBTC_VAL_ATV_AUDIO_SOUND_LEVEL_RIGHT 2
#define RSBTC_VAL_ATV_AUDIO_SOUND_LEVEL_CH1   3
#define RSBTC_VAL_ATV_AUDIO_SOUND_LEVEL_CH2   4

#define RSBTC_VAL_ATV_SOUND_FREQ_DEV_MONO    0
#define RSBTC_VAL_ATV_SOUND_FREQ_DEV_SUBCAR1 1
#define RSBTC_VAL_ATV_SOUND_FREQ_DEV_SUBCAR2 2
#define RSBTC_VAL_ATV_SOUND_FREQ_DEV_PILOT   3

#define RSBTC_VAL_ATV_SOUND_SPEC_SUBCAR_MONO 0
#define RSBTC_VAL_ATV_SOUND_SPEC_SUBCAR1     1
#define RSBTC_VAL_ATV_SOUND_SPEC_SUBCAR2     2

#define RSBTC_VAL_FREQ_CHAN_TAB_SELECT 0
#define RSBTC_VAL_FREQ_CHAN_TAB_STORE  1
#define RSBTC_VAL_FREQ_CHAN_TAB_DELETE 2

#define RSBTC_SWEEP_START  0
#define RSBTC_SWEEP_CENTER 1



#define RSBTC_VAL_ARB_WF_TAG_TYPE        0
#define RSBTC_VAL_ARB_WF_TAG_CRESTFACTOR 1
#define RSBTC_VAL_ARB_WF_TAG_LEVEL_OFFS  2
#define RSBTC_VAL_ARB_WF_TAG_CLOCK       3
#define RSBTC_VAL_ARB_WF_TAG_SAMPLES     4
#define RSBTC_VAL_ARB_WF_TAG_IQ_SWAP     5
#define RSBTC_VAL_ARB_WF_TAG_DATE        6
#define RSBTC_VAL_ARB_WF_TAG_COMMENT     7
						   
#define RSBTC_VAL_PRESET_STAT_REG 0
#define RSBTC_VAL_QUEST_STAT_REG  1
#define RSBTC_VAL_OPER_STAT_REG   2

#define RSBTC_VAL_REG_CONDITION   0
#define RSBTC_VAL_REG_EVENT       1
#define RSBTC_VAL_REG_ENABLE      2
#define RSBTC_VAL_REG_PTRANSITION 3
#define RSBTC_VAL_REG_NTRANSITION 4

#define RSBTC_VAL_FILE_STOP   0
#define RSBTC_VAL_FILE_PLAY   1
#define RSBTC_VAL_FILE_PAUSE  2

#define RSBTC_VAL_IQC_ISDBT_LAYER_A 0
#define RSBTC_VAL_IQC_ISDBT_LAYER_B 1
#define RSBTC_VAL_IQC_ISDBT_LAYER_C 2


#define RSBTC_VAL_IQC_TMM_RECEPTION_CODER_A1 	0
#define RSBTC_VAL_IQC_TMM_RECEPTION_CODER_A2 	1

#define RSBTC_VAL_IQC_ISDBTMM_LAYER_A1A 		0
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_A1B 		1
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_A1C 		2
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_A2A 		3
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_A2B 		4
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_A2C 		5
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B1  		6
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B2  		7
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B3  		8
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B4  		9
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B5  		10
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B6  		11
#define RSBTC_VAL_IQC_ISDBTMM_LAYER_B7  		12

#define RSBTC_VAL_IQC_CMMB_RS_0   0
#define RSBTC_VAL_IQC_CMMB_RS_1   1
#define RSBTC_VAL_IQC_CMMB_RS_2   2
#define RSBTC_VAL_IQC_CMMB_RS_3   3
	
#define RSBTC_VAL_IQC_CMMB_LDPC_R1_2  0
#define RSBTC_VAL_IQC_CMMB_LDPC_R3_4  1

#define RSBTC_VAL_IQC_CMMB_CONST_C2  0
#define RSBTC_VAL_IQC_CMMB_CONST_C4  1
#define RSBTC_VAL_IQC_CMMB_CONST_C16 2

#define RSBTC_VAL_BER_BER               0
#define RSBTC_VAL_BER_EVALUATION        1
#define RSBTC_VAL_BER_STATE             2
#define RSBTC_VAL_BER_ERROR_COUNT       3
#define RSBTC_VAL_BER_GATING_TIME       4
#define RSBTC_VAL_BER_ALL               5


#define RSBTC_VAL_MMEM_COPY             0
#define RSBTC_VAL_MMEM_DELETE           1
#define RSBTC_VAL_MMEM_RENAME           2
#define RSBTC_VAL_MMEM_MDIR             3
#define RSBTC_VAL_MMEM_CDIR             4
#define RSBTC_VAL_MMEM_RDIR             5
#define RSBTC_VAL_MMEM_MSIS             6


#define RSBTC_VAL_MMEM_CATALOG_FILES    0
#define RSBTC_VAL_MMEM_CATALOG_SUBDIR   1

#define RSBTC_IQC_DVBT_CHBW_VAR         0

#define RSBTC_VAL_DISTORTION_FREQ_RESPONSE 	0
#define RSBTC_VAL_DISTORTION_NONLINEARITY  	1

/* Distortion File Operations */
#define RSBTC_VAL_DISTORTION_FREQ_RESPONSE   		0
#define RSBTC_VAL_DISTORTION_NONLINEARITY    		1
#define RSBTC_VAL_DISTORTION_LOAD_RX_NONLINEARITY	2
#define RSBTC_VAL_DISTORTION_SAVE_FREQ_RESPONSE   	3
#define RSBTC_VAL_DISTORTION_SAVE_TX_NONLINEARITY 	4
#define RSBTC_VAL_DISTORTION_SAVE_RX_NONLINEARITY 	5


/* DTMB SFN Delay Parameters */
#define RSBTC_VAL_TDMB_SFN_MAX_DEVIATION_TIME   	0
#define RSBTC_VAL_TDMB_SFN_STATIC_DELAY         	1
#define RSBTC_VAL_TDMB_SFN_TX_COMPENSATION_DELAY	2
#define RSBTC_VAL_TDMB_SFN_TOTAL_DELAY          	3
#define RSBTC_VAL_TDMB_SFN_TX_PROCESSING_DELAY  	4
#define RSBTC_VAL_TDMB_SFN_NETWORK_COMPENSATION 	5
#define RSBTC_VAL_TDMB_SFN_TX_DELAY             	6


#define RSBTC_VAL_DTMB_SFN_DELAY_NETWORK    0
#define RSBTC_VAL_DTMB_SFN_DELAY_PROCESSING 1
#define RSBTC_VAL_DTMB_SFN_DELAY_DYNAMIC    2
#define RSBTC_VAL_DTMB_SFN_DELAY_TOTAL      3
#define RSBTC_VAL_DTMB_SFN_DELAY_MAXDEV     4
#define RSBTC_VAL_DTMB_SFN_DELAY_STATIC     5
#define RSBTC_VAL_DTMB_SFN_DELAY_MAXIMUM    6
#define RSBTC_VAL_DTMB_SFN_DELAY_DTIME      7

/* DVBT2 MI constants */
#define RSBTC_VAL_DVBT2MI_IN_BAND_SIGN 		0
#define RSBTC_VAL_DVBT2MI_IN_BAND_SIGN_A 	1
#define RSBTC_VAL_DVBT2MI_IN_BAND_SIGN_B	2

/* ISDBT - AC Data1|2 */
#define RSBTC_VAL_ISDBT_AC_DATA_UNUSED 2

/* ATSC 3.0 SFN Delay Parameters */
#define RSBTC_VAL_ATSC3_SFN_MAX_DEVIATION_TIME 0
#define RSBTC_VAL_ATSC3_SFN_STATIC_DELAY       1
#define RSBTC_VAL_ATSC3_SFN_DYNAMIC_DELAY      2
#define RSBTC_VAL_ATSC3_SFN_TOTAL_DELAY        3
#define RSBTC_VAL_ATSC3_SFN_PROCESS_DELAY      4
#define RSBTC_VAL_ATSC3_SFN_NETWORK_DELAY      5
#define RSBTC_VAL_ATSC3_SFN_JITTER_DELAY       6
	
/* External RF Stream */
#define RSBTC_STREAM_C 		1
#define RSBTC_STREAM_D 		2


/****************************************************************************
 *---------------- Instrument Driver Function Declarations -----------------*
 ****************************************************************************/
ViStatus _VI_FUNC rsbtc_init (ViRsrc resourceName, ViBoolean IDQuery,
                              ViBoolean resetDevice, ViPSession instrumentHandle);
ViStatus _VI_FUNC rsbtc_InitWithOptions (ViRsrc resourceName, ViBoolean IDQuery,
                                         ViBoolean resetDevice,
                                         ViString optionString,
                                         ViPSession instrumentHandle);
ViStatus _VI_FUNC rsbtc_ConfigureSignalType (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 signalType,
                                             ViReal64 frequencyOffset);
ViStatus _VI_FUNC rsbtc_ConfigureDTVTransmissionStandard (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 DTVTransmissionStandard);
ViStatus _VI_FUNC rsbtc_ConfigureATVTransmissionStandard (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 ATVTransmissionStandard);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCTransmissionStandard
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 audioBCTransmissionStandard);
ViStatus _VI_FUNC rsbtc_ConfigureSignalPolarity (ViSession instrumentHandle,
                                                 ViInt32 path, ViInt32 polarity);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2InputSignal (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean t2MIInterface,
                                                   ViInt32 source, ViInt32 input,
                                                   ViInt32 IPTSChannel,
                                                   ViInt32 t2MIPID, ViInt32 t2MISID,
                                                   ViInt32 numberOfPLP);
ViStatus _VI_FUNC rsbtc_ReadDVBT2InputSignalDataRate (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 PLP,
                                                      ViInt32 dataRate,
                                                      ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ReadDVBT2InputSignalT2MIPacketTiming
             (ViSession instrumentHandle, ViInt32 path, ViInt32 measurementMode,
              ViInt32 t2MITime, ViPReal64 currentTimeValue);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2Settings (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 testTSPacket,
                                                ViInt32 PIDTestPacket, ViInt32 PID,
                                                ViInt32 payloadTestTSPacket,
                                                ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2ModeStreamAdaptation
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP, ViInt32 PLPID,
              ViInt32 PLPIDGroup, ViInt32 PLPType, ViInt32 BBMode);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2BICMParameters (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 PLP,
                                                      ViInt32 FECFrame,
                                                      ViInt32 codeRate,
                                                      ViInt32 constellation,
                                                      ViBoolean constellRotation);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2BICMTimeInterleaverParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 timeInterleaverType, ViInt32 frameIntervalIJump,
              ViInt32 timeInterleaverLength, ViInt32 firstFrameIndex);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2FramingOFDMParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 channelBandwidth,
              ViInt32 FFTSize, ViInt32 guardInterval, ViInt32 pilotPattern);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2FramingOFDM (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViInt32 t2FramesPerSuperFrame,
                                                   ViInt32 dataSymbolsPerT2Frame,
                                                   ViInt32 subslicesPerT2Frame);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2TransmissionSystem (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 transmissionSystem,
                                                          ViInt32 peakToAvragePowerRatio,
                                                          ViBoolean futureExtensionFrames,
                                                          ViBoolean timeFrequencySlicing,
                                                          ViInt32 MISOGroup,
                                                          ViInt32 MISOGroupMode);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2TransmissionSystemNetwork
             (ViSession instrumentHandle, ViInt32 path, ViInt32 networkMode,
              ViInt32 profileMode);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2TransmissionFutureExtensionFrames
             (ViSession instrumentHandle, ViInt32 path, ViInt32 FEFLength,
              ViInt32 FEFType, ViInt32 FEFInterval, ViInt32 FEFPayload);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2TransmissionL1Parameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 l1T2Version,
              ViInt32 l1PostModulation, ViBoolean l1Repetition,
              ViBoolean l1PostScrambled);
ViStatus _VI_FUNC rsbtc_ConfigureDVBT2TransmissionSystemIdentification
             (ViSession instrumentHandle, ViInt32 path, ViInt32 cellID,
              ViInt32 networkID, ViInt32 t2SystemID, ViInt32 TXIDAvailability);
ViStatus _VI_FUNC rsbtc_ReadDVBT2SFNDelayParameters (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViInt32 delayParameter,
                                                     ViPReal64 timeValue);
ViStatus _VI_FUNC rsbtc_ReadDVBT2SFNTimestampParameters (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViChar _VI_FAR UTCDate[],
                                                         ViChar _VI_FAR UTCTime[],
                                                         ViPInt32 secondsSince2000);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2NumberOfPLP (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViInt32 numberOfPLP);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2InputSignal (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 PLP,
                                                   ViInt32 inputFormat,
                                                   ViInt32 source, ViInt32 input,
                                                   ViInt32 IPTSChannel,
                                                   ViInt32 testSignal);
ViStatus _VI_FUNC rsbtc_ReadDVBC2InputSignalDataRate (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 PLP,
                                                      ViInt32 dataRate,
                                                      ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2Settings (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 testTSPacket,
                                                ViInt32 PIDTestPacket, ViInt32 PID,
                                                ViInt32 payloadTestTSPacket,
                                                ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2ModeStreamAdaptation
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP, ViInt32 PLPID,
              ViInt32 BBMode, ViInt32 ISSY, ViBoolean nullPacketDeletion);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2BICMPLPMappingParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 FECFrame, ViInt32 codeRate, ViInt32 constellation,
              ViInt32 mappedToDataSlice);
ViStatus _VI_FUNC rsbtc_ReadDVBC2DataSliceMinMax (ViSession instrumentHandle,
                                                  ViInt32 path,
                                                  ViPInt32 min_REL_DxTu,
                                                  ViPInt32 max_REL_DxTu,
                                                  ViPInt32 min_ABS_DxTu,
                                                  ViPInt32 max_ABS_DxTu);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2NumberOfDataSlices (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 numberOfDataSlices);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2DataSliceParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 dataSlice,
              ViInt32 dataSliceID, ViInt32 dataSliceTIDepth, ViInt32 dataSliceType,
              ViInt32 FECHeaderType);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2CurrentDataSliceParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 dataSlice,
              ViInt32 offsetLeftDxTu, ViInt32 tunePosition_REL_DxTu,
              ViInt32 offsetRightDxTu);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2NumberOfNotches (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViInt32 numberOfNotches);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2NotchValues (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 notch,
                                                   ViInt32 start_REL_DxTu,
                                                   ViInt32 stop_REL_DxTu,
                                                   ViInt32 widthDxTu);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2FramingAndOFDMParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 channelBandwidth,
              ViInt32 c2Bandwidth, ViInt32 guardInterval, ViInt32 basebandFilter);
ViStatus _VI_FUNC rsbtc_ConfigureDVBC2C2SystemParameters (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 networkID,
                                                          ViInt32 c2SystemID,
                                                          ViInt32 l1TimeInterleaver);
ViStatus _VI_FUNC rsbtc_ConfigureDVBTInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path,
                                                  ViInt32 priorityPath,
                                                  ViInt32 source, ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadDVBTInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViInt32 priorityPath,
                                                     ViInt32 dataRate,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDVBTCodingSettings (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViInt32 testTSPacket,
                                                     ViInt32 PIDTestPacket,
                                                     ViInt32 PID,
                                                     ViInt32 payloadTestStuff,
                                                     ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDVBTTPS (ViSession instrumentHandle, ViInt32 path,
                                          ViInt32 TPSCellID,
                                          ViBoolean TPSResevedState,
                                          ViInt32 TPSReservedBits);
ViStatus _VI_FUNC rsbtc_ConfigureDVBTTimeslicingAndMPE (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 priorityPath,
                                                        ViBoolean timeSlicing,
                                                        ViBoolean MPEFEC);
ViStatus _VI_FUNC rsbtc_ConfigureDVBTCoding (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 channelBandwidth,
                                             ViReal64 usedBandwidth,
                                             ViInt32 FFTMode, ViInt32 guardInterval,
                                             ViInt32 hierarchy,
                                             ViInt32 constellation);
ViStatus _VI_FUNC rsbtc_ConfigureDVBTHCoding (ViSession instrumentHandle,
                                              ViInt32 path, ViBoolean state,
                                              ViInt32 symbolInterleaverMode);
ViStatus _VI_FUNC rsbtc_EnableDVBTSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_StoreAndLoadDVBTCarriers (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 operation,
                                                  ViString source);
ViStatus _VI_FUNC rsbtc_ConfigureDVBCInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadDVBCInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 dataRate,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDVBCSettings (ViSession instrumentHandle,
                                               ViInt32 path, ViInt32 testTSPacket,
                                               ViInt32 PIDTestPacket, ViInt32 PID,
                                               ViInt32 payloadTestStuff,
                                               ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDVBCCoding (ViSession instrumentHandle,
                                             ViInt32 path, ViReal64 symbolRate,
                                             ViInt32 constellation,
                                             ViInt32 rollOff);
ViStatus _VI_FUNC rsbtc_EnableDVBCSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadDVBSInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 dataRate,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSSettings (ViSession instrumentHandle,
                                               ViInt32 path, ViInt32 testTSPacket,
                                               ViInt32 PIDTestPacket, ViInt32 PID,
                                               ViInt32 payloadTestStuff,
                                               ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSCoding (ViSession instrumentHandle,
                                             ViInt32 path, ViReal64 symbolRate,
                                             ViInt32 constellation, ViInt32 rollOff,
                                             ViInt32 codeRate);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSCIDSignal (ViSession instrumentHandle,
                                                ViInt32 path, ViBoolean CID,
                                                ViBoolean hostSignal,
                                                ViInt32 frequencyOffset);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSCIDCoordinates (ViSession instrumentHandle,
                                                     ViInt32 path, ViBoolean state,
                                                     ViReal64 latitude,
                                                     ViReal64 longitude);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSCIDTelephone (ViSession instrumentHandle,
                                                   ViInt32 path, ViBoolean state,
                                                   ViString number,
                                                   ViBoolean extensionState,
                                                   ViString extensionNumber);
ViStatus _VI_FUNC rsbtc_EnableDVBSPhaseNoise (ViSession instrumentHandle,
                                              ViInt32 path, ViBoolean phaseNoise);
ViStatus _VI_FUNC rsbtc_ConfigureDVBSPhaseNoise (ViSession instrumentHandle,
                                                 ViInt32 path, ViBoolean CWMode,
                                                 ViInt32 magnitude, ViInt32 shape);
ViStatus _VI_FUNC rsbtc_EnableDVBSSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_Configure8VSBInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_Read8VSBInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 dataRate,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_Configure8VSBSettings (ViSession instrumentHandle,
                                               ViInt32 path, ViInt32 testTSPacket,
                                               ViInt32 PIDTestPacket, ViInt32 PID,
                                               ViInt32 payloadTestStuff,
                                               ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_Enable8VSBSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_ConfigureJ83BInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadJ83BInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 dataRate,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureJ83BSettings (ViSession instrumentHandle,
                                               ViInt32 path, ViInt32 testTSPacket,
                                               ViInt32 PIDTestPacket, ViInt32 PID,
                                               ViInt32 payloadTestStuff,
                                               ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureJ83BCoding (ViSession instrumentHandle,
                                             ViInt32 path, ViReal64 symbolRate,
                                             ViInt32 constellation,
                                             ViInt32 interleaver);
ViStatus _VI_FUNC rsbtc_EnableJ83BSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTInputSignal (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 layer,
                                                   ViInt32 source,
                                                   ViInt32 testSignal);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTExternalInputParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 input,
              ViInt32 IPTSChannel);
ViStatus _VI_FUNC rsbtc_ReadISDBTInputSignalDataRate (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 layer,
                                                      ViInt32 dataRate,
                                                      ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTSettings (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 testTSPacket,
                                                ViInt32 PIDTestPacket, ViInt32 PID,
                                                ViInt32 payloadTestStuff,
                                                ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTLayerCoding (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 layer,
                                                   ViInt32 portion,
                                                   ViInt32 constellation,
                                                   ViInt32 segments,
                                                   ViInt32 codeRate,
                                                   ViInt32 timeInterleaving);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTCoding (ViSession instrumentHandle,
                                              ViInt32 path, ViInt32 ISDBTMode,
                                              ViInt32 guardInterval,
                                              ViInt32 channelBandwidth,
                                              ViInt32 bandwidthVariation);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTAllSegments (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViInt32 segmentsLayerA,
                                                   ViInt32 segmentsLayerB,
                                                   ViInt32 segmentsLayerC);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTEEWSystem (ViSession instrumentHandle,
                                                 ViInt32 path, ViBoolean EEW,
                                                 ViInt32 signalType,
                                                 ViInt32 numberOfEpicenters,
                                                 ViString areaInformation);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTEEWParameters (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViInt32 epicenter,
                                                     ViInt32 informationType,
                                                     ViInt32 warningID,
                                                     ViReal64 latitude,
                                                     ViReal64 longitude,
                                                     ViInt32 depth,
                                                     ViInt32 occurrenceTime);
ViStatus _VI_FUNC rsbtc_EnableISDBTSpecialSettings (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTSpecialSegmentSettings
             (ViSession instrumentHandle, ViInt32 path, ViInt32 segment,
              ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTSpecialSettings (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViBoolean guardInsertion,
                                                       ViInt32 TXParamSwitchingIndicator,
                                                       ViInt32 TMCCNext,
                                                       ViInt32 ACData_AC1,
                                                       ViInt32 ACData_AC2);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadDTMBInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 dataRate,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBSettings (ViSession instrumentHandle,
                                               ViInt32 path, ViInt32 testTSPacket,
                                               ViInt32 PIDTestPacket, ViInt32 PID,
                                               ViInt32 payloadTestStuff,
                                               ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBControl (ViSession instrumentHandle,
                                              ViInt32 path, ViInt32 control);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBNetworkMode (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 networkMode,
                                                  ViBoolean singleCarrierMode,
                                                  ViBoolean controlFrames);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBCoding (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 constellation,
                                             ViInt32 codeRate,
                                             ViInt32 guardInterval,
                                             ViInt32 guardIntervalPN,
                                             ViInt32 timeInterleaver,
                                             ViInt32 channelBandwidth);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBSFN (ViSession instrumentHandle, ViInt32 path,
                                          ViBoolean SIPRestamping,
                                          ViBoolean mute1stSignalFrame,
                                          ViInt32 _1PPSRouting, ViReal64 offset);
ViStatus _VI_FUNC rsbtc_ConfigureDTMBSFNDelay (ViSession instrumentHandle,
                                               ViInt32 path,
                                               ViReal64 maxDeviationTime,
                                               ViReal64 staticDelay);
ViStatus _VI_FUNC rsbtc_ReadDTMBSFNDelay (ViSession instrumentHandle, ViInt32 path,
                                          ViInt32 type, ViPReal64 delay);
ViStatus _VI_FUNC rsbtc_EnableDTMBSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2CodingAndModulationMode
             (ViSession instrumentHandle, ViInt32 path, ViInt32 CMMode,
              ViInt32 numInputStreams);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2InputSignal (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViInt32 inputStream,
                                                   ViInt32 source, ViInt32 input,
                                                   ViInt32 IPTSChannel,
                                                   ViInt32 testSignal,
                                                   ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadDVBS2InputSignalDataRate (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViInt32 inputStream,
                                                      ViInt32 dataRate,
                                                      ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2Settings (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 testTSPacket,
                                                ViInt32 PIDTestPacket, ViInt32 PID,
                                                ViInt32 payloadTestStuff,
                                                ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2Coding (ViSession instrumentHandle,
                                              ViInt32 path, ViReal64 symbolRate,
                                              ViInt32 rollOff, ViBoolean annexM,
                                              ViInt32 numTimeSlice, ViInt32 TSN);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2TimeSliceCoding (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViInt32 inputStream,
                                                       ViInt32 TSL,
                                                       ViInt32 modulationCoding,
                                                       ViInt32 FECFrame,
                                                       ViBoolean pilots);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2S2X (ViSession instrumentHandle, ViInt32 path,
                                           ViBoolean s2X, ViBoolean superFrame,
                                           ViBoolean channelBonding);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2CIDSignal (ViSession instrumentHandle,
                                                 ViInt32 path, ViBoolean CID,
                                                 ViBoolean hostSignal,
                                                 ViInt32 frequencyOffset);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2CIDCoordinates (ViSession instrumentHandle,
                                                      ViInt32 path, ViBoolean state,
                                                      ViReal64 latitude,
                                                      ViReal64 longitude);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2CIDTelephone (ViSession instrumentHandle,
                                                    ViInt32 path, ViBoolean state,
                                                    ViString number,
                                                    ViBoolean extensionState,
                                                    ViString extensionNumber);
ViStatus _VI_FUNC rsbtc_EnableDVBS2PhaseNoise (ViSession instrumentHandle,
                                               ViInt32 path, ViBoolean phaseNoise);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2PhaseNoise (ViSession instrumentHandle,
                                                  ViInt32 path, ViBoolean CWMode,
                                                  ViInt32 magnitude, ViInt32 shape);
ViStatus _VI_FUNC rsbtc_EnableDVBS2SpecialSettings (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2SpecialPLScrambling
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 PLScramblingSequence, ViInt32 PLGoldCodeIndex);
ViStatus _VI_FUNC rsbtc_ConfigureDVBS2SpecialPLScramblingState
             (ViSession instrumentHandle, ViInt32 path, ViBoolean PLScrambler,
              ViBoolean PLDataslicesPRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDIRECTVInputSignal (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 source,
                                                     ViInt32 input,
                                                     ViInt32 IPTSChannel,
                                                     ViInt32 testSignal,
                                                     ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadDIRECTVInputSignalDataRate (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 dataRate,
                                                        ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureDIRECTVSettings (ViSession instrumentHandle,
                                                  ViInt32 path,
                                                  ViInt32 testTSPacket,
                                                  ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureDIRECTVCoding (ViSession instrumentHandle,
                                                ViInt32 path, ViReal64 symbolRate,
                                                ViInt32 constellation,
                                                ViInt32 rollOff, ViInt32 codeRate);
ViStatus _VI_FUNC rsbtc_EnableDIRECTVSpecialSettings (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViBoolean specialSettings);
ViStatus _VI_FUNC rsbtc_EnableDIRECTVPhaseNoise (ViSession instrumentHandle,
                                                 ViInt32 path,
                                                 ViBoolean phaseNoise);
ViStatus _VI_FUNC rsbtc_ConfigureDIRECTVPhaseNoise (ViSession instrumentHandle,
                                                    ViInt32 path, ViBoolean CWMode,
                                                    ViInt32 magnitude,
                                                    ViInt32 shape);
ViStatus _VI_FUNC rsbtc_ConfigureTDMBInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel);
ViStatus _VI_FUNC rsbtc_ReadTDMBInfoSTCSSTC (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 subchannel,
                                             ViChar _VI_FAR subchannelID_Hex[],
                                             ViPInt32 protectionProfile,
                                             ViPInt32 protectionLevel,
                                             ViPInt32 dataRate);
ViStatus _VI_FUNC rsbtc_TDMBNetworkConfiguration (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 networkMode,
                                                  ViBoolean TTI, ViInt32 TIIMainID,
                                                  ViInt32 TIISubID);
ViStatus _VI_FUNC rsbtc_EnableTDMBSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path, ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureTDMBSpecialSettings (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViInt32 transmissionMode,
                                                      ViInt32 selectTransmissionMode,
                                                      ViBoolean PRBSTestSignal,
                                                      ViInt32 PRBSTestSignalSettings,
                                                      ViInt32 testSignalSCID);
ViStatus _VI_FUNC rsbtc_ConfigureATSCMHInputSignal (ViSession instrumentHandle,
                                                    ViInt32 path, ViInt32 source,
                                                    ViInt32 input,
                                                    ViInt32 IPTSChannel,
                                                    ViInt32 testSignal,
                                                    ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadATSCMHInputSignalDataRate (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViInt32 dataRate,
                                                       ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureATSCMHSettings (ViSession instrumentHandle,
                                                 ViInt32 path, ViInt32 testTSPacket,
                                                 ViInt32 PIDTestPacket, ViInt32 PID,
                                                 ViInt32 payloadTestStuff,
                                                 ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureATSCMHNetworkSettings (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 MHEPID,
                                                        ViBoolean RFWatermark,
                                                        ViInt32 txAddress,
                                                        ViInt32 networkID,
                                                        ViInt32 buryRatio);
ViStatus _VI_FUNC rsbtc_ConfigureATSCMHMTXIDTransmission (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViBoolean MTXIDTransmission,
                                                          ViInt32 transmitterID,
                                                          ViInt32 marketID);
ViStatus _VI_FUNC rsbtc_ConfigureATSCMHCodingParameters (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViReal64 symbolRate,
                                                         ViBoolean MHState);
ViStatus _VI_FUNC rsbtc_EnableATSCMHSpecialSettings (ViSession instrumentHandle,
                                                     ViInt32 path, ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3STLInterface (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViBoolean STLInterface);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3ChannelBandwidth (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 channelBandwidth);
ViStatus _VI_FUNC rsbtc_GetATSC3StatusInfo (ViSession instrumentHandle,
                                            ViInt32 path, ViInt32 bufferSize,
                                            ViChar _VI_FAR statusInfo[]);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3PLPID (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 PLP,
                                             ViInt32 PLPID);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3InputSignal (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViBoolean conformanceCheck,
                                                   ViInt32 source,
                                                   ViString destinationIPAddress,
                                                   ViInt32 destinationIPPort,
                                                   ViInt32 inputType);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3InputTestSignal (ViSession instrumentHandle,
                                                       ViInt32 path, ViInt32 PLP,
                                                       ViInt32 testSignal);
ViStatus _VI_FUNC rsbtc_ResetATSC3InputSignalLogFile (ViSession instrumentHandle,
                                                      ViInt32 path);
ViStatus _VI_FUNC rsbtc_ReadATSC3InputSignalNumberOfPLP (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViPInt32 numberOfPLPs);
ViStatus _VI_FUNC rsbtc_ReadATSC3InputSignalPacketLength (ViSession instrumentHandle,
                                                          ViInt32 path, ViInt32 PLP,
                                                          ViPInt32 packetLength);
ViStatus _VI_FUNC rsbtc_ReadATSC3InputSignalProtocol (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViPInt32 protocol);
ViStatus _VI_FUNC rsbtc_ReadATSC3InputSignalDataRate (ViSession instrumentHandle,
                                                      ViInt32 path,ViInt32 PLP,
                                                      ViInt32 dataRate,
                                                      ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3Settings (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 testTSPacket,
                                                ViInt32 PIDTestPacket, ViInt32 PID,
                                                ViInt32 testIPPacket,
                                                ViInt32 payloadTestStuff,
                                                ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3EncapsAndFormat (ViSession instrumentHandle,
                                                       ViInt32 path, ViInt32 PLP,
                                                       ViBoolean BBFrameCounter);
ViStatus _VI_FUNC rsbtc_ReadATSC3EncapsAndFormat (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 PLP,
                                                  ViPInt32 PLPID,
                                                  ViPInt32 ALPEncapsulationType,
                                                  ViPBoolean BBFramePadding);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3BICMParameters (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 PLP,
                                                      ViInt32 FECType,
													   ViInt32 codeRate,
                                                      ViInt32 constellation);
ViStatus _VI_FUNC rsbtc_ReadATSC3BICMParameters (ViSession instrumentHandle,
                                                 ViInt32 path, ViInt32 PLP,
                                                 ViPInt32 lowLevelSignaling,
                                                 ViPInt32 layer,
                                                 ViPInt32 LDMInjectionLevel,
                                                 ViPInt32 size,
                                                 ViPInt32 scramblerType);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3BICMSubslicing (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 PLP,
                                                      ViInt32 type,
                                                      ViInt32 numberOfSubslices,
													  ViInt32 subsliceInterval);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3BICMTimeInterleaver
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 timeInterleaverMode, ViInt32 depth, ViInt32 numberOfTIBlocks,
              ViBoolean cellInterleaver);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3BICMInterSubframe (ViSession instrumentHandle,
                                                         ViInt32 path, ViInt32 PLP,
                                                         ViBoolean interSubframe);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3BICMExtendedInterleaving
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViBoolean extendedInterleaving);
ViStatus _VI_FUNC rsbtc_ReadATSC3BICMNumberOfFECBlocks (ViSession instrumentHandle,
                                                        ViInt32 path, ViInt32 PLP,
                                                        ViPInt32 maxNumberOfFECBlocks,
														ViPInt32 numberOfFECBlocks);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SubframingOFDMParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 subframe,
              ViInt32 MISO, ViInt32 FFTSize, ViInt32 reducedCarrierMode,
              ViInt32 guardInterval, ViInt32 totalNumbOfDataOFDMSymbos,
              ViInt32 firstSubframeBoundarySymbol, ViBoolean frequencyInterleaver);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SubframingOFDMPilot
             (ViSession instrumentHandle, ViInt32 path, ViInt32 subframe,
              ViInt32 SISOPilotPattern, ViInt32 pilotBoostMode);
ViStatus _VI_FUNC rsbtc_ReadATSC3SubframingOFDMMIMO (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 subframe,
                                                     ViPBoolean MIMO);
ViStatus _VI_FUNC rsbtc_ReadATSC3SubframingOFDMBoundarySymbol
             (ViSession instrumentHandle, ViInt32 path, ViInt32 subframe,
              ViPInt32 subframeBoundarySymbol_Last,
              ViPInt32 boundarySymbol_NullCells);
ViStatus _VI_FUNC rsbtc_ReadATSC3SubframingOFDMUsedBandwidth
             (ViSession instrumentHandle, ViInt32 path, ViInt32 subframe,
              ViPReal64 duration, ViPReal64 usedBandwidth);
ViStatus _VI_FUNC rsbtc_ReadATSC3SubframingOFDMPLPs (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 subframe,
                                                     ViPInt32 numberOfPLPsInSubframe,
                                                     ViInt32 bufferSize,
                                                     ViChar _VI_FAR PLPIDsInSubframe[]);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SystemNetworkParameters
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 networkMode,ViInt32 timeInfo);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SystemNetworkMISOParameters
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 numberOfTransmitters_NTX, ViInt32 transmitterIndex_x);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SystemNetworkTxID (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViInt32 txIDMode,
                                                         ViInt32 txIDAddress,
                                                         ViInt32 txIDInjectionLevel);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SystemNetworkBroadcastStreamID
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 broadcastStreamID_BSID);
ViStatus _VI_FUNC rsbtc_ReadATSC3SystemNetwork (ViSession instrumentHandle,
                                                ViInt32 path,
                                                ViPInt32 lowLevelSignaling,
                                                ViPInt32 returnChannel);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SystemFrameParameters
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 peakToAveragePowerRatio, ViInt32 frameMode,
              ViInt32 frameLength);
ViStatus _VI_FUNC rsbtc_ReadATSC3SystemFrameTimeParameters
             (ViSession instrumentHandle, ViInt32 path,
              ViPInt32 excessSamplesPerSymbol, ViPInt32 finalExcessSamples);
ViStatus _VI_FUNC rsbtc_ReadATSC3SystemFrameSymbolParameters
             (ViSession instrumentHandle, ViInt32 path, ViPInt32 timeOffset,
              ViPInt32 additionalSamples);
ViStatus _VI_FUNC rsbtc_ReadATSC3SystemNumberOfSubframes (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViPInt32 numberOfSubframes);
ViStatus _VI_FUNC rsbtc_ReadATSC3SystemL1BasicVersion (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViPInt32 l1BasicVersion);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SystemL1Parameters (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 l1BasicFECType,
                                                          ViInt32 reducedCarrierModePreamble,
                                                          ViInt32 pilotPattern_Dx_Preamble,
                                                          ViInt32 l1DetailVersion,
                                                          ViInt32 l1DetailFECType);
ViStatus _VI_FUNC rsbtc_ReadATSC3SystemL1Parameters (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViPInt32 numberOfPreambleSymbols,
                                                     ViPInt32 l1DetailAdditParityMode,
                                                     ViPInt32 numberOfRFs);
ViStatus _VI_FUNC rsbtc_ReadATSC3FrameInfoFrameDuration (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViPReal64 frameDuration);
ViStatus _VI_FUNC rsbtc_ReadATSC3FrameInfoL1SignalingParameters
             (ViSession instrumentHandle, ViInt32 path, ViPInt32 l1BasicBytes,
              ViPInt32 l1BasicCells, ViPInt32 l1DetailBytes,
              ViPInt32 l1DetailCells);
ViStatus _VI_FUNC rsbtc_ReadATSC3FrameInfoBootstrapParameters
             (ViSession instrumentHandle, ViInt32 path, ViPReal64 bootstrapDuration,
              ViPInt32 emergencyAlertSignaling, ViPInt32 minTimeToNext,
              ViPInt32 systemBandwidth, ViPInt32 BSRCoefficient,
              ViPInt32 preambleStructure);
ViStatus _VI_FUNC rsbtc_ReadATSC3FrameInfoBootstrapVersion
             (ViSession instrumentHandle, ViInt32 path, ViPInt32 majorVersion,
              ViPInt32 minorVersion);
ViStatus _VI_FUNC rsbtc_ReadATSC3FrameInfoBootstrapPreambleParameters
             (ViSession instrumentHandle, ViInt32 path, ViPInt32 FFTSize,
              ViPInt32 guardInterval, ViPInt32 pilotPattern_Dx,
              ViPInt32 l1BasicFECType);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SFNMute (ViSession instrumentHandle,
                                               ViInt32 path, ViBoolean mute);
ViStatus _VI_FUNC rsbtc_ReadATSC3SFNDelayParameters (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViInt32 delayParameter,
                                                     ViPReal64 timeValue);
ViStatus _VI_FUNC rsbtc_EnableATSC3SpecialSettings (ViSession instrumentHandle,
                                                    ViInt32 path, ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SpecialSTLInterface
             (ViSession instrumentHandle, ViInt32 path,
              ViBoolean preambleCompatibilityMode, ViBoolean TMPCompatibilityMode);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SpecialBootstrap (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 minorVersion,
                                                        ViInt32 emergencyAlertSignaling);
ViStatus _VI_FUNC rsbtc_ConfigureATSC3SpecialALPEncapsulation
             (ViSession instrumentHandle, ViInt32 path,
              ViBoolean LMTCompatibilityMode);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTmmInputSignal (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 layer,
                                                     ViInt32 source, ViInt32 input,
                                                     ViInt32 IPTSChannel,
                                                     ViInt32 testSignal,
                                                     ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadISDBTmmInputSignalDataRate (ViSession instrumentHandle,
                                                        ViInt32 path, ViInt32 layer,
                                                        ViInt32 value,
                                                        ViPReal64 result);
ViStatus _VI_FUNC rsbtc_confISDBTmmSettings (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 testTSPacket,
                                             ViInt32 PIDTestPacket, ViInt32 PID,
                                             ViInt32 payloadTestStuff,
                                             ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTmmCodingParameters (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 control,
                                                          ViInt32 coder,
                                                          ViBoolean partialReception,
                                                          ViInt32 numberOfSuperSegments);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTmmCodingSegmentParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 superSegments,
              ViInt32 superSegmentsType, ViInt32 numSegments);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTmmCodingLayerParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 layer,
              ViInt32 OFDMSegments, ViInt32 centerSubchannel, ViInt32 constellation,
              ViInt32 codeRate, ViInt32 timeInterleaving);
ViStatus _VI_FUNC rsbtc_ConfigureISDBTmmCoding (ViSession instrumentHandle,
                                                ViInt32 path,
                                                ViBoolean emergencyAlarm,
                                                ViInt32 ISDBTMode,
                                                ViInt32 guardInterval,
                                                ViInt32 referenceChannel);
ViStatus _VI_FUNC rsbtc_EnableISDBTmmSpecialSettings (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureCMMBInputSignal (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 source,
                                                  ViInt32 input,
                                                  ViInt32 IPTSChannel,
                                                  ViInt32 testSignal,
                                                  ViBoolean stuffing);
ViStatus _VI_FUNC rsbtc_ReadCMMBInputSignalDataRate (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 value,
                                                     ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureCMMBSettings (ViSession instrumentHandle,
                                               ViInt32 path, ViInt32 transmitterID,
                                               ViInt32 areaID,
                                               ViInt32 payloadTestSignal,
                                               ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigureCMMBCoding (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 control,
                                             ViInt32 bandwidth);
ViStatus _VI_FUNC rsbtc_ConfigureCMMBPLCHTable (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 PLCH,
                                                ViInt32 RS, ViInt32 byteIL,
                                                ViInt32 LDPC, ViInt32 constel,
                                                ViInt32 SCRMode, ViInt32 startTS,
                                                ViInt32 stopTS);
ViStatus _VI_FUNC rsbtc_EnableCMMBSpecialSettings (ViSession instrumentHandle,
                                                   ViInt32 path, ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvVideo (ViSession instrumentHandle,
                                                ViInt32 path,
                                                ViInt32 videoInputSignal,
                                                ViInt32 visionPicture);
ViStatus _VI_FUNC rsbtc_LoadAnalogTvVideoLibrary (ViSession instrumentHandle,
                                                  ViInt32 path,
                                                  ViString loadLibrary);
ViStatus _VI_FUNC rsbtc_AnalogTvVideoLibraryCatalog (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViString directory,
                                                     ViInt32 bufferSize,
                                                     ViChar _VI_FAR output[],
                                                     ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_AnalogTvVideoLibraryCatalogLength
             (ViSession instrumentHandle, ViInt32 path, ViString directory,
              ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvAudioInputSignal
             (ViSession instrumentHandle, ViInt32 path, ViInt32 audioInputSignal);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvAudioAF (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 audioSource,
                                                  ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvAudio (ViSession instrumentHandle,
                                                ViInt32 path,
                                                ViInt32 frequencyParameter,
                                                ViReal64 frequencyValue,
                                                ViInt32 levelParameter,
                                                ViReal64 levelValue);
ViStatus _VI_FUNC rsbtc_LoadAnalogTvAudioLibrary (ViSession instrumentHandle,
                                                  ViInt32 path,
                                                  ViString loadLibrary);
ViStatus _VI_FUNC rsbtc_AnalogTvAudioLibraryCatalog (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViString directory,
                                                     ViInt32 bufferSize,
                                                     ViChar _VI_FAR output[],
                                                     ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_AnalogTvAudioLibraryCatalogLength
             (ViSession instrumentHandle, ViInt32 path, ViString directory,
              ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_ReadAnalogTvAudioPlayerValues (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViInt32 levelParameter,
                                                       ViPReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvNICAMSource (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 source,
                                                      ViInt32 NICAMSoundMode);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvNICAMAF (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 audioSource,
                                                  ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvNICAM (ViSession instrumentHandle,
                                                ViInt32 path,
                                                ViInt32 frequencyParameter,
                                                ViReal64 frequencyValue,
                                                ViInt32 levelParameter,
                                                ViReal64 levelValue);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSoundModulation (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 mode);
ViStatus _VI_FUNC rsbtc_ReadAnalogTvSoundModulationValues
             (ViSession instrumentHandle, ViInt32 path, ViInt32 parameter,
              ViPReal64 result);
ViStatus _VI_FUNC rsbtc_LoadAnalogTVBTSCLibrary (ViSession instrumentHandle,
                                                 ViInt32 path,
                                                 ViString loadLibrary);
ViStatus _VI_FUNC rsbtc_AnalogTVBTSCLibraryCatalog (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViString directory,
                                                    ViInt32 bufferSize,
                                                    ViChar _VI_FAR output[],
                                                    ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_AnalogTVBTSCLibraryCatalogLength (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViString directory,
                                                          ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_EnableAnalogTvSpecialSettings (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSpecialVideoCarrier
             (ViSession instrumentHandle, ViInt32 path, ViBoolean video,
              ViBoolean carrier);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSpecialSettings (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViBoolean VSBFilter,
                                                          ViBoolean GDPrecorrection,
                                                          ViReal64 residualCarrier);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSpecialVsbCharacteristic
             (ViSession instrumentHandle, ViInt32 path, ViInt32 VSBCharacteristic);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvPreemphasis (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViInt32 preemphasis);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSubcarrierSettings
             (ViSession instrumentHandle, ViInt32 path, ViInt32 subcarrier,
              ViReal64 nomFreqDev, ViBoolean subcarrierState,
              ViReal64 frequencySubcarrier, ViReal64 levelSubcarrier);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvPilot (ViSession instrumentHandle,
                                                ViInt32 path, ViBoolean pilot,
                                                ViReal64 nomFreqDevPilot);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSpecialAMModulationDepth
             (ViSession instrumentHandle, ViInt32 path, ViInt32 AMModulationDepth);
ViStatus _VI_FUNC rsbtc_ConfigureAnalogTvSpecialNICAM (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViBoolean preemphasis,
                                                       ViInt32 rollOff,
                                                       ViReal64 gainAdjustAF1,
                                                       ViReal64 gainAdjustAF2);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMInputSignal (ViSession instrumentHandle,
                                                       ViInt32 path, ViInt32 source,
                                                       ViInt32 inputSignal,
                                                       ViInt32 AFMode);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMGeneratorSource
             (ViSession instrumentHandle, ViInt32 path, ViInt32 source,
              ViBoolean audio, ViReal64 frequency, ViReal64 level);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMPlayerSource (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 source,
                                                        ViBoolean audio,
                                                        ViReal64 attenuation);
ViStatus _VI_FUNC rsbtc_LoadAudioBCFMPlayerFile (ViSession instrumentHandle,
                                                 ViInt32 path,
                                                 ViString loadLibrary);
ViStatus _VI_FUNC rsbtc_AudioBCFMLibraryCatalog (ViSession instrumentHandle,
                                                 ViInt32 path, ViString directory,
                                                 ViInt32 bufferSize,
                                                 ViChar _VI_FAR output[],
                                                 ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_AudioBCFMCatalogLength (ViSession instrumentHandle,
                                                ViInt32 path, ViString directory,
                                                ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAudioModulator (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 mode,
                                                          ViInt32 preemphasis,
                                                          ViReal64 nomFreqDevAudio);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMModulationRDS (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViBoolean RDS,
                                                         ViReal64 freqDevRDS,
                                                         ViReal64 freqDevPilot);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMRDSRBDS (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViString groupSequence,
                                                   ViInt32 RDSPI, ViString PS,
                                                   ViBoolean TP, ViBoolean TA,
                                                   ViInt32 PTY, ViString PTYN,
                                                   ViString RT, ViInt32 MS);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMRDSRBDSDecoderIdentification
             (ViSession instrumentHandle, ViInt32 path, ViBoolean DIDynamicPTY,
              ViBoolean DICompressed, ViBoolean DIArtHead, ViBoolean DIStereo,
              ViBoolean CT, ViString CTOffset);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAFMethod (ViSession instrumentHandle,
                                                    ViInt32 path, ViInt32 AFMethod);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAFANumberOfFrequencies
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 numberOfFrequencies);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAFAFrequency (ViSession instrumentHandle,
                                                        ViInt32 path, ViInt32 AF,
                                                        ViReal64 frequency);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAFBNumberOfLists
             (ViSession instrumentHandle, ViInt32 path, ViInt32 numberOfLists);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAFBList (ViSession instrumentHandle,
                                                   ViInt32 path,
                                                   ViInt32 selectEditList,
                                                   ViInt32 numberOfFrequencies,
                                                   ViReal64 tuningFrequency);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMAFBFrequency (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 selectEditList,
                                                        ViInt32 AF,
                                                        ViReal64 frequency,
                                                        ViInt32 order);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMEONParameters (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViInt32 EONPI,
                                                         ViString EONPS,
                                                         ViBoolean EONTP,
                                                         ViBoolean EONTA,
                                                         ViBoolean EONLA,
                                                         ViBoolean EONEG,
                                                         ViBoolean EONILS,
                                                         ViInt32 EONLSN,
                                                         ViInt32 EONPTY,
                                                         ViInt32 EONPIN,
                                                         ViInt32 EONAFMethod);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMEONAFMappedNumberOfFrequency
             (ViSession instrumentHandle, ViInt32 path, ViInt32 numberOfFrequencies,
              ViReal64 tuningFrequency);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMEONAFMappedFrequency
             (ViSession instrumentHandle, ViInt32 path, ViInt32 AF,
              ViReal64 frequency);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMEONAFANumberOfFrequencies
             (ViSession instrumentHandle, ViInt32 path,
              ViInt32 numberOfFrequencies);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMEONAFAFrequency
             (ViSession instrumentHandle, ViInt32 path, ViInt32 AF,
              ViReal64 frequency);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCTMCGroup3AVARParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 group3AVAR00,
              ViInt32 group3AVAR01);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCTMCNumberOf8AGroups
             (ViSession instrumentHandle, ViInt32 path, ViInt32 numberOf8AGroups);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMTMC8ABlock (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 group8A,
                                                      ViInt32 block,
                                                      ViInt32 setValue);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCOpenFormatGroupA (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 groupA,
                                                          ViInt32 block,
                                                          ViInt32 setValue);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCOpenFormatGroupB (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 groupA,
                                                          ViInt32 block,
                                                          ViInt32 setValue);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCFMSpecialSettings
             (ViSession instrumentHandle, ViInt32 path, ViBoolean specialSettings,
              ViBoolean pilot, ViReal64 phaseOffsetPilot19kHz,
              ViReal64 phaseOffsetRDS);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCAMInputSignal (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViInt32 inputSignal);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCAMGeneratorSource
             (ViSession instrumentHandle, ViInt32 path, ViBoolean audio,
              ViReal64 frequency, ViReal64 level);
ViStatus _VI_FUNC rsbtc_ConfigureAudioBCAMPlayerSource (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViBoolean audio,
                                                        ViReal64 attenuation);
ViStatus _VI_FUNC rsbtc_LoadAudioBCAMPlayerFile (ViSession instrumentHandle,
                                                 ViInt32 path,
                                                 ViString loadLibrary);
ViStatus _VI_FUNC rsbtc_AudioBCAMLibraryCatalog (ViSession instrumentHandle,
                                                 ViInt32 path, ViString directory,
                                                 ViInt32 bufferSize,
                                                 ViChar _VI_FAR output[],
                                                 ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_AudioBCAMCatalogLength (ViSession instrumentHandle,
                                                ViInt32 path, ViString directory,
                                                ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_confAudioBCAMModulationDepth (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViReal64 nominalModulationDepth);
ViStatus _VI_FUNC rsbtc_LoadARBWaveform (ViSession instrumentHandle, ViInt32 path,
                                         ViInt32 ARB, ViString loadWaveform);
ViStatus _VI_FUNC rsbtc_DeleteARBWaveform (ViSession instrumentHandle, ViInt32 path,
                                           ViInt32 ARB, ViString deleteWaveform);
ViStatus _VI_FUNC rsbtc_ARBWaveformFileTag (ViSession instrumentHandle,
                                            ViInt32 path, ViInt32 ARB,
                                            ViString waveformFileName, ViInt32 tag,
                                            ViInt32 bufferSize,
                                            ViChar _VI_FAR output[],
                                            ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_ARBWaveformCatalog (ViSession instrumentHandle,
                                            ViInt32 path, ViInt32 ARB,
                                            ViString directory, ViInt32 bufferSize,
                                            ViChar _VI_FAR output[],
                                            ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_ARBWaveformCatalogLength (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 ARB,
                                                  ViString directory,
                                                  ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_ARBWaveformPoints (ViSession instrumentHandle, ViInt32 path,
                                           ViInt32 ARB, ViPInt32 points);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalInputLevel (ViSession instrumentHandle,
                                                      ViInt32 path, ViBoolean state,
                                                      ViBoolean IQSwap,
                                                      ViBoolean digitalIQAutoSettings,
                                                      ViReal64 measurementPeriod,
                                                      ViReal64 crestFactor,
                                                      ViReal64 peakLevel);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalInSampleRate (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 source,
                                                        ViReal64 sampleRateValue);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalOutState (ViSession instrumentHandle,
                                                    ViInt32 path, ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalOutIQSwap (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViBoolean IQSwap);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalOutLevel (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViInt32 setLevelVia,
                                                    ViReal64 level);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalOutSampleRate (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViInt32 source,
                                                         ViReal64 sampleRateValue);
ViStatus _VI_FUNC rsbtc_ConfigureInterferer (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 source,
                                             ViInt32 routing,
                                             ViReal64 frequencyOffset,
                                             ViInt32 reference);
ViStatus _VI_FUNC rsbtc_DistortionFileOperations (ViSession instrumentHandle,
                                                  ViInt32 path, ViInt32 operation,
                                                  ViString filename);
ViStatus _VI_FUNC rsbtc_ConfigureDistortionLimiter (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViBoolean txDistortionLimiter,
                                                    ViReal64 limiterPosition);
ViStatus _VI_FUNC rsbtc_FadingConfiguration (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 configuration,
                                             ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureFadingMode (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 mode,
                                             ViInt32 signalRouting);
ViStatus _VI_FUNC rsbtc_ConfigureFadingPathProfile (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViInt32 configuration,
                                                    ViInt32 group,
                                                    ViInt32 fadingPath,
                                                    ViInt32 profile);
ViStatus _VI_FUNC rsbtc_FadingCatalog (ViSession instrumentHandle, ViInt32 path,
                                       ViString directory, ViInt32 bufferSize,
                                       ViChar _VI_FAR output[],
                                       ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_ConfigureFadingStandardSet (ViSession instrumentHandle,
                                                    ViInt32 path,
                                                    ViInt32 standardParameterSet);
ViStatus _VI_FUNC rsbtc_EnableFadingPath (ViSession instrumentHandle, ViInt32 path,
                                          ViInt32 configuration, ViInt32 group,
                                          ViInt32 fadingPath, ViBoolean state);
ViStatus _VI_FUNC rsbtc_StoreAndLoadFadingProfile (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 operation,
                                                   ViString source);
ViStatus _VI_FUNC rsbtc_ConfigureFadingPathProfileTableNumeric
             (ViSession instrumentHandle, ViInt32 path, ViInt32 configuration,
              ViInt32 group, ViInt32 fadingPath, ViInt32 parameterName,
              ViReal64 parameterValue);
ViStatus _VI_FUNC rsbtc_ConfigureFadingPathProfileTableBoolean
             (ViSession instrumentHandle, ViInt32 path, ViInt32 configuration,
              ViInt32 group, ViInt32 fadingPath, ViInt32 parameterName,
              ViBoolean parameterValue);
ViStatus _VI_FUNC rsbtc_ReadFadingPathProfileTableResults
             (ViSession instrumentHandle, ViInt32 path, ViInt32 configuration,
              ViInt32 group, ViInt32 fadingPath, ViInt32 parameterName,
              ViPReal64 parameterValue);
ViStatus _VI_FUNC rsbtc_ConfigureFadingMIMOCorrelationMatrixTable
             (ViSession instrumentHandle, ViInt32 path, ViInt32 fadingPath,
              ViInt32 direction, ViInt32 steering, ViReal64 magnitudeReal,
              ViReal64 phaseImaginary);
ViStatus _VI_FUNC rsbtc_ConfigureFadingMIMORelativeTapGainVectorTable
             (ViSession instrumentHandle, ViInt32 path, ViInt32 fadingPath,
              ViInt32 direction, ViReal64 gain, ViReal64 phase);
ViStatus _VI_FUNC rsbtc_ConfigureFadingMIMORelativeTapGainVectorSetToUnity
             (ViSession instrumentHandle, ViInt32 path, ViInt32 fadingPath);
ViStatus _VI_FUNC rsbtc_ConfigureFadingCommonSettings (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViInt32 speedUnit,
                                                       ViInt32 signalDedicatedTo,
                                                       ViReal64 virtualRF);
ViStatus _VI_FUNC rsbtc_ConfigureFadingInsertionLoss (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 mode,
                                                      ViReal64 insertionLoss);
ViStatus _VI_FUNC rsbtc_ConfigureFadingReferenceSettings (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 reference,
                                                          ViBoolean commonSpeedAllPath);
ViStatus _VI_FUNC rsbtc_ConfigureFadingExternalRFRemoteConnection
             (ViSession instrumentHandle, ViInt32 path, ViInt32 stream,
              ViString targetName, ViInt32 remoteConnection);
ViStatus _VI_FUNC rsbtc_ConfigureFadingExternalRFCoupling
             (ViSession instrumentHandle, ViInt32 path, ViInt32 stream,
              ViBoolean RFCoupling);
ViStatus _VI_FUNC rsbtc_ConfigureFadingExternalRFParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 stream,
              ViReal64 RFFrequency, ViReal64 RFLevel, ViBoolean RFState);
ViStatus _VI_FUNC rsbtc_ReadFadingExternalRFConnectionState
             (ViSession instrumentHandle, ViInt32 path, ViInt32 stream,
              ViPInt32 remoteConnectionState,ViPInt32 digitalIQConnectionState,
              ViInt32 bufferSize, ViChar _VI_FAR digitalIQConnectedInstrument[]);
ViStatus _VI_FUNC rsbtc_ConfigureAWGNNoise (ViSession instrumentHandle,
                                            ViInt32 path, ViInt32 noise);
ViStatus _VI_FUNC rsbtc_ConfigureAWGNGenerator (ViSession instrumentHandle,
                                                ViInt32 path, ViBoolean AWGN,
                                                ViReal64 cN);
ViStatus _VI_FUNC rsbtc_ReadAWGNNoiseEbN0 (ViSession instrumentHandle, ViInt32 path,
                                           ViPReal64 ebN0);
ViStatus _VI_FUNC rsbtc_ConfigureAWGNNoiseBandwidth (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViBoolean bandwidthCoupling,
                                                     ViReal64 receiverBandwidth,
                                                     ViReal64 generatedNoiseBandwidth);
ViStatus _VI_FUNC rsbtc_ConfigureImpulsiveNoise (ViSession instrumentHandle,
                                                 ViInt32 path,
                                                 ViBoolean noiseImpulsive,
                                                 ViReal64 cI, ViInt32 frameDuration,
                                                 ViInt32 pulsesPerBurst,
                                                 ViReal64 pulseSpacingMin,
                                                 ViReal64 pulseSpacingMax);
ViStatus _VI_FUNC rsbtc_ConfigurePhaseNoiseState (ViSession instrumentHandle,
                                                  ViInt32 path,
                                                  ViBoolean phaseNoise);
ViStatus _VI_FUNC rsbtc_ConfigurePhaseNoiseSSB (ViSession instrumentHandle,
                                                ViInt32 path,
                                                ViString SSBShapeProfile);
ViStatus _VI_FUNC rsbtc_ConfigurePhaseNoiseLevel (ViSession instrumentHandle,
                                                  ViInt32 path, ViReal64 result);
ViStatus _VI_FUNC rsbtc_ConfigurePhaseNoiseFrequency (ViSession instrumentHandle,
                                                      ViInt32 path, ViInt32 pointNo,
                                                      ViReal64 frequency);
ViStatus _VI_FUNC rsbtc_NoiseFileOperations (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 operation,
                                             ViString source);
ViStatus _VI_FUNC rsbtc_ConfigureFrequency (ViSession instrumentHandle,
                                            ViInt32 path, ViReal64 frequency,
                                            ViReal64 frequencyOffset);
ViStatus _VI_FUNC rsbtc_ConfigureLocalOscillatorCoupling (ViSession instrumentHandle,
                                                          ViInt32 path,
                                                          ViInt32 mode,
                                                          ViBoolean outState);
ViStatus _VI_FUNC rsbtc_FrequencyChannelTableOperations (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViInt32 operation,
                                                         ViString channelName);
ViStatus _VI_FUNC rsbtc_ConfigureChannelTableFrequency (ViSession instrumentHandle,
                                                        ViInt32 path,
                                                        ViInt32 channel,
                                                        ViString channelName,
                                                        ViReal64 channelFrequency);
ViStatus _VI_FUNC rsbtc_ConfigureExternalReferenceFrequency
             (ViSession instrumentHandle, ViInt32 path, ViInt32 source,
              ViBoolean adjustmentState, ViInt32 adjustmentDACValue,
              ViInt32 externalReferenceFrequency, ViInt32 synchronizationBandwidth);
ViStatus _VI_FUNC rsbtc_ConfigureLevel (ViSession instrumentHandle, ViInt32 path,
                                        ViInt32 mode, ViInt32 settingCharacteristic,
                                        ViReal64 level, ViReal64 userLimit,
                                        ViReal64 offset);
ViStatus _VI_FUNC rsbtc_ConfigureLevelWithUnits (ViSession instrumentHandle,
                                                 ViInt32 path, ViReal64 level,
                                                 ViInt32 levelUnit, ViBoolean RF);
ViStatus _VI_FUNC rsbtc_ConfigureAttenuator (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 attenuatorMode,
                                             ViReal64 attenuation);
ViStatus _VI_FUNC rsbtc_ConfigureAutomaticLevelControl (ViSession instrumentHandle,
                                                        ViInt32 path, ViInt32 state,
                                                        ViInt32 detectorSensitivity,
                                                        ViInt32 sampleLevel);
ViStatus _VI_FUNC rsbtc_ConfigureFrequencySweep (ViSession instrumentHandle,
                                                 ViInt32 path, ViBoolean state,
                                                 ViInt32 sweepMode);
ViStatus _VI_FUNC rsbtc_ConfigureFrequencySweepRange (ViSession instrumentHandle,
                                                      ViInt32 path,
                                                      ViInt32 settingType,
                                                      ViReal64 startFrequency,
                                                      ViReal64 stopFrequency,
                                                      ViReal64 centerFrequency,
                                                      ViReal64 frequencySpan);
ViStatus _VI_FUNC rsbtc_ConfigureFrequencySweepStep (ViSession instrumentHandle,
                                                     ViInt32 path,
                                                     ViInt32 sweepSpacing,
                                                     ViReal64 sweepStep,
                                                     ViReal64 sweepDwell);
ViStatus _VI_FUNC rsbtc_ConfigureLevelSweep (ViSession instrumentHandle,
                                             ViInt32 path, ViBoolean state,
                                             ViInt32 mode);
ViStatus _VI_FUNC rsbtc_ConfigureLevelSweepRange (ViSession instrumentHandle,
                                                  ViInt32 path, ViReal64 startLevel,
                                                  ViReal64 stopLevel);
ViStatus _VI_FUNC rsbtc_ConfigureLevelSweepStep (ViSession instrumentHandle,
                                                 ViInt32 path, ViInt32 shape,
                                                 ViReal64 step, ViReal64 dwell);
ViStatus _VI_FUNC rsbtc_ConfigureModulation (ViSession instrumentHandle,
                                             ViInt32 path, ViInt32 signalSource,
                                             ViBoolean iQWideband,
                                             ViInt32 optimizationMode, ViInt32 gain,
                                             ViReal64 crestFactor);
ViStatus _VI_FUNC rsbtc_ConfigureBasebandImpairments (ViSession instrumentHandle,
                                                      ViInt32 path, ViBoolean state,
                                                      ViReal64 iOffset,
                                                      ViReal64 qOffset,
                                                      ViReal64 quadratureError,
                                                      ViReal64 amplitudeImbalance,
                                                      ViReal64 iQDelay);
ViStatus _VI_FUNC rsbtc_ConfigureModulatorImpairments (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViBoolean state,
                                                       ViReal64 iOffset,
                                                       ViReal64 qOffset,
                                                       ViReal64 quadratureError,
                                                       ViReal64 amplitudeImbalance);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalOutput (ViSession instrumentHandle,
                                                  ViInt32 path, ViBoolean state,
                                                  ViInt32 sampleRateSource,
                                                  ViReal64 sampleRateValue,
                                                  ViInt32 setLevelVia,
                                                  ViReal64 levelValue);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalOutputVariation
             (ViSession instrumentHandle, ViInt32 path, ViInt32 variationActive,
              ViReal64 userVariationStep);
ViStatus _VI_FUNC rsbtc_ConfigureIQDigitalImpairments (ViSession instrumentHandle,
                                                       ViInt32 path,
                                                       ViBoolean state,
                                                       ViReal64 iOffset,
                                                       ViReal64 qOffset,
                                                       ViReal64 quadratureError,
                                                       ViReal64 amplitudeImbalance);
ViStatus _VI_FUNC rsbtc_ConfigureIQAnalogOutput (ViSession instrumentHandle,
                                                 ViInt32 path, ViBoolean state,
                                                 ViInt32 IQOutputType, ViInt32 mode,
                                                 ViReal64 IQLevelVp,
                                                 ViBoolean coupleIQBias);
ViStatus _VI_FUNC rsbtc_ConfigureIQAnalogSettings (ViSession instrumentHandle,
                                                   ViInt32 path, ViReal64 iBias_EMF,
                                                   ViReal64 iOffset_EMF,
                                                   ViReal64 qBias_EMF,
                                                   ViReal64 qOffset_EMF);
ViStatus _VI_FUNC rsbtc_ConfigureIQAnalogImpairments (ViSession instrumentHandle,
                                                      ViInt32 path, ViBoolean state,
                                                      ViReal64 iOffset,
                                                      ViReal64 qOffset,
                                                      ViReal64 quadratureError,
                                                      ViReal64 amplitudeImbalance);
ViStatus _VI_FUNC rsbtc_ConfigureBERDataInputInterface (ViSession instrumentHandle,
                                                        ViInt32 input,
                                                        ViInt32 clock, ViInt32 data,
                                                        ViInt32 enable,
                                                        ViInt32 inputFormat,
                                                        ViInt32 signal,
                                                        ViBoolean PIDFilter,
                                                        ViInt32 PID_Hex);
ViStatus _VI_FUNC rsbtc_ConfigureBERTresholdClkDataEn (ViSession instrumentHandle,
                                                       ViReal64 tresholdClkDataEn);
ViStatus _VI_FUNC rsbtc_ConfigureBERGatingTime (ViSession instrumentHandle,
                                                ViString gatingTime,
                                                ViInt32 gatingMode);
ViStatus _VI_FUNC rsbtc_ConfigureBERInputSettings (ViSession instrumentHandle,
                                                   ViInt32 payload, ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_BERMeasurement (ViSession instrumentHandle,
                                        ViInt32 measurement, ViInt32 bufferSize,
                                        ViChar _VI_FAR result[]);
ViStatus _VI_FUNC rsbtc_SelectPlayFile (ViSession instrumentHandle, ViInt32 path,
                                        ViInt32 player, ViString fileName);
ViStatus _VI_FUNC rsbtc_SelectFileCommandsControl (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 player,
                                                   ViInt32 command);
ViStatus _VI_FUNC rsbtc_ReadPlayerParameters (ViSession instrumentHandle,
                                              ViInt32 path, ViInt32 player,
                                              ViChar _VI_FAR fileDateTime[],
                                              ViChar _VI_FAR originalLoopTime[],
                                              ViPReal64 originalTSDataRate);
ViStatus _VI_FUNC rsbtc_ConfigurePlayerOutputParameters (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViInt32 player,
                                                         ViReal64 dataRate,
                                                         ViBoolean nullpacketStuffing,
                                                         ViInt32 stopData);
ViStatus _VI_FUNC rsbtc_ConfigurePlayerOutputSettings (ViSession instrumentHandle,
                                                       ViInt32 path, ViInt32 player,
                                                       ViInt32 testTSPacket,
                                                       ViInt32 PIDTestPacket,
                                                       ViInt32 PID, ViInt32 payload,
                                                       ViInt32 PRBS);
ViStatus _VI_FUNC rsbtc_ConfigurePlayFilePosition (ViSession instrumentHandle,
                                                   ViInt32 path, ViInt32 player,
                                                   ViReal64 start, ViReal64 stop);
ViStatus _VI_FUNC rsbtc_ConfigurePlayFileReset (ViSession instrumentHandle,
                                                ViInt32 path, ViInt32 player);
ViStatus _VI_FUNC rsbtc_ConfigurePlayerSeamlessLoop (ViSession instrumentHandle,
                                                     ViInt32 path, ViInt32 player,
                                                     ViBoolean continuityCounter,
                                                     ViBoolean PCRDTSPTS,
                                                     ViBoolean TDTTOT);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayMode (ViSession instrumentHandle,
                                                 ViInt32 path, ViInt32 mode);
ViStatus _VI_FUNC rsbtc_GetMMGGatewayStatusInfo (ViSession instrumentHandle,
                                                 ViInt32 path, ViInt32 bufferSize,
                                                 ViChar _VI_FAR statusInfo[]);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIInput (ViSession instrumentHandle,
                                                         ViInt32 path,
                                                         ViInt32 useCase,
                                                         ViInt32 profiles,
                                                         ViInt32 PLPGroups,
                                                         ViInt32 PLPs,
                                                         ViBoolean vVStreamGeneration);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPGroup
             (ViSession instrumentHandle, ViInt32 path, ViInt32 group,
              ViInt32 PLPGroupID, ViInt32 mapToProfile, ViInt32 inputType);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPGroupInput
             (ViSession instrumentHandle, ViInt32 path, ViInt32 group,
              ViInt32 selectInput);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPGroupPRBS
             (ViSession instrumentHandle, ViInt32 path, ViInt32 group,
              ViReal64 genDataRate, ViInt32 commonSlotInterval, ViInt32 numChapters,
              ViInt32 numEITPackets, ViInt32 TSOverPLPID, ViBoolean TSOverCPLP,
              ViInt32 PRBSPID);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPGroupPRBSNumRep
             (ViSession instrumentHandle, ViInt32 path, ViInt32 group,
              ViInt32 arraySize, ViInt32 _VI_FAR repetitions[]);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLP (ViSession instrumentHandle,
                                                       ViInt32 path, ViInt32 PLP,
                                                       ViInt32 PLPID,
                                                       ViInt32 PLPType);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPMapToGroup
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP, ViInt32 group);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPData
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 PLPData, ViInt32 arraySize, ViInt32 _VI_FAR serviceIDList[],
              ViInt32 newTSID);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIPLPRunLength
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 arraySize, ViInt32 _VI_FAR numberPackets[]);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIModeStreamPLP
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP, ViInt32 BBMode,
              ViBoolean nullPacketDeletion, ViBoolean autoISSY);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIModeStreamPLPISSY
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP, ViInt32 ISSY,
              ViInt32 maxBufferSize, ViInt32 bufferSizeUnit, ViInt32 designDelay);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIModeStreamPLPInBandSign
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 inBandSignalling, ViBoolean state);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIBICMParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViInt32 FECFrame, ViInt32 codeRate, ViInt32 constellation,
              ViBoolean constellRotation);
ViStatus 
    _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIBICMTimeInterleaverParameters
        (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
         ViInt32 timeInterleaverType, ViInt32 frameIntervalIJump,
         ViInt32 timeInterleaverLength, ViInt32 firstFrameIndex);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIBICMMaxBlocks
             (ViSession instrumentHandle, ViInt32 path, ViInt32 PLP,
              ViBoolean autoNumBlockMax, ViInt32 numBlocksMax);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIFramingOFDMParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
              ViInt32 channelBandwidth, ViInt32 FFTSize, ViInt32 guardInterval,
              ViInt32 pilotPattern);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIFramingOFDM
             (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
              ViInt32 t2FramesPerSuperFrame, ViInt32 dataSymbolsPerT2Frame,
              ViInt32 subslicesPerT2Frame);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MITransmissionSystem
             (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
              ViInt32 scheduling, ViInt32 transmissionSystem,
              ViInt32 peakToAveragePowerRatio, ViBoolean futureExtensionFrames,
              ViBoolean timeFrequencySlicing);
ViStatus 
    _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MITransmissionSystemIdentification
        (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
         ViInt32 networkID, ViInt32 cellID, ViInt32 t2SystemID);
ViStatus 
    _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MITransmissionFutureExtensionFrames
        (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
         ViInt32 FEFLength, ViInt32 FEFType, ViInt32 FEFInterval);
ViStatus _VI_FUNC rsbtc_ConfigureMMGDVBT2MITransmissionL1Parameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
              ViInt32 l1T2Version, ViInt32 l1PostModulation, ViBoolean l1Repetition,
              ViBoolean l1PostExtension, ViBoolean l1PostScrambled,
              ViBoolean l1BiasBalancingBits, ViBoolean t2BaseLite,
              ViReal64 l1RFFrequency);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIOutputParameters
             (ViSession instrumentHandle, ViInt32 path, ViInt32 t2TimestampMode,
              ViInt32 timeOffset);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIOutputProfile
             (ViSession instrumentHandle, ViInt32 path, ViInt32 profile,
              ViInt32 PID, ViInt32 SID, ViInt32 TSType, ViInt32 PMTPID,
              ViReal64 outputDataRate);
ViStatus _VI_FUNC rsbtc_ConfigureMMGGatewayDVBT2MIOutputRestart
             (ViSession instrumentHandle, ViInt32 path);
ViStatus _VI_FUNC rsbtc_PowerMeterConfiguration (ViSession instrumentHandle,
                                                 ViInt32 unit,
                                                 ViReal64 referencePower);
ViStatus _VI_FUNC rsbtc_ConfigurePowerMeterFrequency (ViSession instrumentHandle,
                                                      ViInt32 source,
                                                      ViReal64 frequency);
ViStatus _VI_FUNC rsbtc_ConfigurePowerMeterAverageFilter (ViSession instrumentHandle,
                                                          ViInt32 filterType,
                                                          ViInt32 filterLength);
ViStatus _VI_FUNC rsbtc_SystemDisplayUpdate (ViSession instrumentHandle,
                                             ViBoolean displayUpdate);
ViStatus _VI_FUNC rsbtc_SystemKeypadLock (ViSession instrumentHandle,
                                          ViInt32 keypadLock);
ViStatus _VI_FUNC rsbtc_SetStatusRegister (ViSession instrumentHandle,
                                           ViInt32 registerOperation,
                                           ViInt32 enable, ViInt32 PTransition,
                                           ViInt32 NTransition);
ViStatus _VI_FUNC rsbtc_GetStatusRegister (ViSession instrumentHandle,
                                           ViInt32 registerOperation,
                                           ViInt32 statusRegistersQuery,
                                           ViPInt32 registerValue);
ViStatus _VI_FUNC rsbtc_SetUserDefStatusRegister (ViSession instrumentHandle,
                                                  ViInt32 registerOperation,
                                                  ViInt32 userdefinableBit,
                                                  ViInt32 enable,
                                                  ViInt32 PTransition,
                                                  ViInt32 NTransition);
ViStatus _VI_FUNC rsbtc_GetUserDefStatusRegister (ViSession instrumentHandle,
                                                  ViInt32 registerOperation,
                                                  ViInt32 userdefinableBit,
                                                  ViInt32 statusRegistersQuery,
                                                  ViPInt32 registerValue);
ViStatus _VI_FUNC rsbtc_FileOperations (ViSession instrumentHandle,
                                        ViInt32 operationToBePerformed,
                                        ViString source, ViString destination);
ViStatus _VI_FUNC rsbtc_DirectoryPath (ViSession instrumentHandle, ViInt32 read,
                                       ViString directory, ViInt32 bufferSize,
                                       ViChar _VI_FAR output[],
                                       ViPInt32 returnedValues);
ViStatus _VI_FUNC rsbtc_CatalogLength (ViSession instrumentHandle, ViInt32 read,
                                       ViPInt32 catalogLength);
ViStatus _VI_FUNC rsbtc_SetAttributeViInt32 (ViSession instrumentHandle,
                                             ViString repeatedCapabilityName,
                                             ViUInt32 attributeID,
                                             ViInt32 attributeValue);
ViStatus _VI_FUNC rsbtc_SetAttributeViReal64 (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViReal64 attributeValue);
ViStatus _VI_FUNC rsbtc_SetAttributeViString (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViString attributeValue);
ViStatus _VI_FUNC rsbtc_SetAttributeViBoolean (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViBoolean attributeValue);
ViStatus _VI_FUNC rsbtc_SetAttributeViSession (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViSession attributeValue);
ViStatus _VI_FUNC rsbtc_GetAttributeViInt32 (ViSession instrumentHandle,
                                             ViString repeatedCapabilityName,
                                             ViUInt32 attributeID,
                                             ViPInt32 attributeValue);
ViStatus _VI_FUNC rsbtc_GetAttributeViReal64 (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViPReal64 attributeValue);
ViStatus _VI_FUNC rsbtc_GetAttributeViString (ViSession instrumentHandle,
                                              ViString repeatedCapabilityName,
                                              ViUInt32 attributeID,
                                              ViInt32 bufferSize,
                                              ViChar _VI_FAR attributeValue[]);
ViStatus _VI_FUNC rsbtc_GetAttributeViBoolean (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViPBoolean attributeValue);
ViStatus _VI_FUNC rsbtc_GetAttributeViSession (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViPSession attributeValue);
ViStatus _VI_FUNC rsbtc_CheckAttributeViInt32 (ViSession instrumentHandle,
                                               ViString repeatedCapabilityName,
                                               ViUInt32 attributeID,
                                               ViInt32 attributeValue);
ViStatus _VI_FUNC rsbtc_CheckAttributeViReal64 (ViSession instrumentHandle,
                                                ViString repeatedCapabilityName,
                                                ViUInt32 attributeID,
                                                ViReal64 attributeValue);
ViStatus _VI_FUNC rsbtc_CheckAttributeViString (ViSession instrumentHandle,
                                                ViString repeatedCapabilityName,
                                                ViUInt32 attributeID,
                                                ViString attributeValue);
ViStatus _VI_FUNC rsbtc_CheckAttributeViBoolean (ViSession instrumentHandle,
                                                 ViString repeatedCapabilityName,
                                                 ViUInt32 attributeID,
                                                 ViBoolean attributeValue);
ViStatus _VI_FUNC rsbtc_CheckAttributeViSession (ViSession instrumentHandle,
                                                 ViString repeatedCapabilityName,
                                                 ViUInt32 attributeID,
                                                 ViSession attributeValue);
ViStatus _VI_FUNC rsbtc_GetAttributeRepeatedCapabilityIds
             (ViSession instrumentHandle, ViUInt32 attributeID, ViInt32 bufferSize,
              ViChar _VI_FAR repeatedCapabilityId_s[]);
ViStatus _VI_FUNC rsbtc_GetAttributeRepeatedCapabilityIdNames
             (ViSession instrumentHandle, ViUInt32 attributeID,
              ViString repeatedCapabilityId, ViInt32 bufferSize,
              ViChar _VI_FAR repeatedCapabilityIdName_s[]);
ViStatus _VI_FUNC rsbtc_StoreAndLoadApplication (ViSession instrumentHandle,
                                                 ViInt32 operation,
                                                 ViString source);
ViStatus _VI_FUNC rsbtc_StoreAndLoadInstrumentData (ViSession instrumentHandle,
                                                    ViInt32 operation,
                                                    ViInt32 stateNumber);
ViStatus _VI_FUNC rsbtc_reset (ViSession instrumentHandle);
ViStatus _VI_FUNC rsbtc_self_test (ViSession instrumentHandle,
                                   ViPInt16 selfTestResult,
                                   ViChar _VI_FAR selfTestMessage[]);
ViStatus _VI_FUNC rsbtc_revision_query (ViSession instrumentHandle,
                                        ViChar _VI_FAR instrumentDriverRevision[],
                                        ViChar _VI_FAR firmwareRevision[]);
ViStatus _VI_FUNC rsbtc_error_query (ViSession instrumentHandle, ViPInt32 errorCode,
                                     ViChar _VI_FAR errorMessage[]);
ViStatus _VI_FUNC rsbtc_error_message (ViSession instrumentHandle,
                                       ViStatus errorCode,
                                       ViChar _VI_FAR errorMessage[]);
ViStatus _VI_FUNC rsbtc_SetOPCTimeout (ViSession instrumentHandle, ViInt32 timeout);
ViStatus _VI_FUNC rsbtc_GetOPCTimeout (ViSession instrumentHandle,
                                       ViPInt32 timeout);
ViStatus _VI_FUNC rsbtc_GetError (ViSession instrumentHandle, ViPStatus code,
                                  ViInt32 bufferSize, ViChar _VI_FAR description[]);
ViStatus _VI_FUNC rsbtc_ClearError (ViSession instrumentHandle);
ViStatus _VI_FUNC rsbtc_LockSession (ViSession instrumentHandle,
                                     ViPBoolean callerHasLock);
ViStatus _VI_FUNC rsbtc_UnlockSession (ViSession instrumentHandle,
                                       ViPBoolean callerHasLock);
ViStatus _VI_FUNC rsbtc_WriteInstrData (ViSession instrumentHandle,
                                        ViString writeBuffer);
ViStatus _VI_FUNC rsbtc_ReadInstrData (ViSession instrumentHandle,
                                       ViInt32 numberOfBytesToRead,
                                       ViChar _VI_FAR readBuffer[],
                                       ViPInt32 numBytesRead);
ViStatus _VI_FUNC rsbtc_close (ViSession instrumentHandle);


    /*********************************************************
        Functions reserved for class driver use only.
        End-users should not call these functions.
     *********************************************************/

ViStatus _VI_FUNC  rsbtc_RsInit (ViRsrc resourceName, ViBoolean IDQuery,
                                  ViBoolean reset, ViSession* vi);
ViStatus _VI_FUNC  rsbtc_RsClose (ViSession vi);


ViStatus _VI_FUNC rsbtc_SetErrorInfo (ViSession session,
                          ViBoolean overWrite,
                          ViStatus primaryError,
                          ViStatus secondaryError,
                          ViConstString errorElaboration);

ViStatus _VI_FUNC rsbtc_GetErrorInfo (ViSession session,
                          ViStatus *primaryError,
                          ViStatus *secondaryError,
                          ViChar errorElaboration[]);
                 
                 

/****************************************************************************
 *------------------------ Error And Completion Codes ----------------------*
 ****************************************************************************/

#define RSBTC_WARNMSG_INSTR_SPECIFIC "Instrument warning"

#define RSBTC_ERROR_CODES_AND_MSGS \
    {RSBTC_WARN_INSTR_SPECIFIC,  RSBTC_WARNMSG_INSTR_SPECIFIC}

#define VI_ERROR_INSTR_FILE_OPEN                (_VI_ERROR+0x3FFC0800L)
#define VI_ERROR_INSTR_FILE_WRITE               (_VI_ERROR+0x3FFC0801L)
#define VI_ERROR_INSTR_INTERPRETING_RESPONSE    (_VI_ERROR+0x3FFC0803L)

#define RSBTC_WARN_INSTR_SPECIFIC       (RS_WARN_BASE + 0x106)

/****************************************************************************
 *---------------------------- End Include File ----------------------------*
 ****************************************************************************/


#if defined(__cplusplus) || defined(__cplusplus__)
}
#endif
#endif /* __RSBTC_HEADER */
